unit DataThread;

interface

uses
  Classes, SysUtils, Windows, WinSock;

type
   TDataThread = class(TThread)
   protected
      ErrMsg      : string;
      LogMsg      : string;
      LogQuery    : string;
      LogHdrs     : string;
      DLLName     : string;
      ECB_Result  : string;
      procedure   Execute; override;
      function    ProcessRequest(var R: string):boolean;
      procedure   RunDLL(var Result: string);
      function    FmtHTMLResponse(const HTMLText: string): string;
      procedure   ShowError;
      procedure   UpdateLog;
      procedure   SendFile(const FileName, FileType: string; var LogData:string);
   public
      DataSock    : TSocket;
      ServerName  : AnsiString;
      RemoteHost  : ansistring;
      QueryString : ansistring;
      FileName    : ansistring;
      Method      : ansistring;
      ContentLength:integer;
      PathInfo    : ansistring;

      Data        : ansistring;
   end;

implementation

uses Main, SockUtil, Isapi2, StrUtil, System.StrUtils;

{ TDataThread }

function TDataThread.FmtHTMLResponse(const HTMLText: string): string;
var
   ResStr : string;
begin
   ResStr := Format(
      'HTTP/1.0 200 OK'+CR+
      'Content-Type: text/html'+CR+
      'Content-Length: %d'+CR+
      'Content:'+CR+CR, [Length(HTMLText)]) + HTMLText;
   Result := ResStr;
end;

function Get_HTTP_Response(Socket: TSocket;
          var FileName, Method, Query, Data: ansistring;  // these variables must be ansistring to use in ECB structure
          var Hdrs: string;
          var ContentLength:integer):boolean;
const
   CONLENSTR = 'CONTENT-LENGTH:';
var
   s           : string;
   tmp         : string;
   Lines       : TStringList;
   Line1       : string;
   p, p1       : integer;
   l           : integer;
begin
   tmp := GetData(Socket);
   if length(tmp)=0 then
   begin
        Result:=false;
        exit;
   end;
   Result:=true;
   Lines := TSTringList.Create;
   try
      Lines.Text := tmp;

      // Get the FileName
      Line1 := Lines[0];
      p := Pos(' ', Line1);
      Method := UpperCase(Trim(Copy(Line1, 1, p-1)));
      if p<>0 then Line1 := Copy(Line1, p+1, Length(Line1));   // Trim After the first space
      p := Pos(' ', Line1);
      if p<>0 then Line1 := Copy(Line1, 1, p-1);               // Trim Before the next space

      FindRepAll(Line1, '\', '/');
      while Copy(Line1, 1, 1)='/' do
         Line1 := Copy(Line1, 2, Length(Line1));   // Get rid of the first "/"
      FileName := Line1;

      // Check if the filename has query string
      Query := '';
      p := Pos('?', FileName);
      if p<>0 then begin
         Query    := Copy(FileName, p+1, Length(FileName));
         FileName := Copy(FileName, 1, p-1);
      end;

      // Get the Headers
      p:=Pos(CR+CR, tmp);
      if p=0 then
         p:=Length(tmp);
      p1:=Length(Lines[0]) + 3;
      Hdrs:=Copy(tmp, p1, p-p1);

      // Look for content-length response.
      ContentLength:=0;
      for l:=0 to Lines.Count-1 do
      begin
           s:=trim(Lines[l]);
           if UpperCase(copy(s,1,length(CONLENSTR))) = CONLENSTR then
           begin
                ContentLength:=strtoint(copy(s,length(CONLENSTR)+1,10));
                break;
           end;
      end;

      // Return rest as data.
      l:=length(tmp);
      if (l>1) and (copy(tmp,l-1,2)=CR) then dec(l,2);
      inc(p,length(CR+CR));
      if p<length(tmp) then
         Data:=copy(tmp,p,l-p+1)
      else
         Data:='';

      // Check if there should be some data to return, then read again.
      if (ContentLength>0) and (Data='') then
         Data:=GetData(Socket);

   finally
      Lines.Free;
   end;
end;

procedure TDataThread.Execute;
var
   ResStr   : string;
   DoSend   : boolean;
   dev,fn,path  : string;
   i        : integer;
begin
   DoSend:=true;
   try
      if not Get_HTTP_Response(DataSock, FileName, Method, QueryString, Data, LogHdrs, ContentLength) then exit;
      LogMsg := FileName;
      LogQuery := QueryString+Data;

      // Perform aliasing
      if UpperCase(FileName) = '' then FileName := 'default.htm';

      // Extract file path.
      path:=FileName;

      fn:=ExtractFileName(path);
      dev:=ExtractFileDrive(path);
      path:=ExtractFilePath(path);
      if dev='' then
      begin
           // Make sure nobody requests an absolute filepath.
           for i:=1 to length(path) do
               if not (path[i] in ['\','/','.',' ']) then break;
           path:=Copy(path,i,length(path));
           FileName:=path+fn;

           ResStr := '';
           DoSend:=ProcessRequest(ResStr);
      end
      else ResStr:='Illegal path';

      // Send the Response
      Synchronize(UpdateLog);
   except
      on e: Exception do begin
         ErrMsg := e.Message + ' (Data Thread)';
         Synchronize(ShowError);
         ResStr := FmtHTMLResponse(ErrMsg);
      end;
   end;

   try
      if DoSend then PutDataNoCR(DataSock,  ResStr);
      if MainForm.miLogdatasend.Checked then
          MainForm.Log(MainForm.Hits, RemoteHost, 'Sending:', ResStr, '','');
   except
   end;

   closesocket(DataSock);
end;


procedure TDataThread.ShowError;
begin
     MainForm.ErrorLab.Caption := ErrMsg;
     MainForm.Log(MainForm.Hits, RemoteHost, LogMsg, LogQuery, LogHdrs, ErrMsg);
end;


procedure TDataThread.UpdateLog;
begin
   MainForm.Log(MainForm.Hits, RemoteHost, LogMsg, LogQuery, LogHdrs, '');
end;

// Return true if the caller should send data returned in R.
function TDataThread.ProcessRequest(var R: string):boolean;
var
   ext      : string;
   i        : integer;
   tmp      : string;
   sname    : string;
   sport    : Integer;
   pb       : Integer;
begin
   ext := UpperCase(ExtractFileExt(FileName));

   Result:=false;
   if (ext='.HTM') or (ext='.HTML') then begin
      SendFile(FileName, 'text/html',R);
      exit;
   end;
//gds
   if (ext='.CSS')  then begin
      SendFile(FileName, 'text/css',R);
      exit;
   end;



   if (ext='.JPG') or (ext='.JPEG') then begin
      SendFile(FileName, 'image/jpeg',R);
      exit;
   end;

   if (ext='.GIF') then begin
      SendFile(FileName, 'image/gif',R);
      exit;
   end;

   if (ext='.BMP') then begin
      SendFile(FileName, 'image/bmp',R);
      exit;
   end;

   if (ext='.WBMP') then begin
      SendFile(FileName, 'image/vnd.wap.wbmp',R);
      exit;
   end;

   if (ext='.WMLC') then begin
      SendFile(FileName, 'application/vnd.wap.wmlc',R);
      exit;
   end;

   if (ext='.WMLS') then begin
      SendFile(FileName, 'text/vnd.wap.wmlscript',R);
      exit;
   end;

   if (ext='.WMLSC') then begin
      SendFile(FileName, 'application/vnd.wap.wmlscriptc',R);
      exit;
   end;

   // Check for ISAPI DLL.
   i:=pos('.DLL',UpperCase(FileName));
   if i>0 then
   begin
      PathInfo:=copy(FileName,i+4,length(FileName));
      FileName:=copy(FileName,1,i+3);
      i:=pos('?',PathInfo);
      if i=1 then PathInfo:=''
      else if i>1 then PathInfo:=copy(PathInfo,1,i-1);
      RunDLL(R);
      Result:=true;
      exit;
   end;

   if ext<>'.EXE' then
      SendFile(FileName, '',R);
   exit;
end;

function GetHeaderLine(Text:string;Name:string):string;
var
   i,j:integer;
begin
     Result:='';
     i:=pos(Name,UpperCase(Text));
     if i<=0 then exit;

     inc(i,length(Name));
     j:=i;
     while j<=length(text) do
     begin
          if (ord(Text[j])=$D) or (ord(Text[j])=$A) then
          begin
               dec(j);
               break;
          end;
          inc(j);
     end;
     Result:=trim(copy(Text,i,j-i+1));
end;

function GetServerVariable(hConn: HCONN; VariableName: PAnsiChar; Buffer: Pointer;
   var Size: DWORD ): BOOL stdcall;
var
   dt : TDataThread;
   s : string;
   v : string;
begin
   dt := TDataThread(hConn);
   v := VariableName;
   v:=UpperCase(v);

   PAnsiChar(Buffer)[0]:=#0;
   Size:=0;
   Result:=false;

   if v = 'SERVER_NAME' then
   begin
        s:=GetHeaderLine(dt.LogHdrs,'HOST:');
        if s<>'' then
        begin
             StrPCopy(PAnsiChar(Buffer),s);
             Size:=length(s)+1;
             Result:=true;
        end;
        Exit;
   end;

   if v = 'SCRIPT_NAME' then
   begin
        s:='/'+ExtractFileName(dt.DLLName);
        StrPCopy(PAnsiChar(Buffer), s);
        Size := Length(s)+1;
        Result := True;
        exit;
   end;

   if (v = 'REMOTE_HOST') or
      (v = 'REMOTE_USER') or
      (v = 'REMOTE_ADDR') then
   begin
        StrPCopy(PAnsiChar(Buffer), dt.RemoteHost);
        Size := Length(dt.RemoteHost)+1;
        Result := True;
        exit;
   end;

   if v = 'ALL_HTTP' then
   begin
        StrPCopy(PAnsiChar(Buffer),dt.LogHdrs);
        Size:=length(dt.LogHdrs)+1;
        Result:=true;
        exit;
   end;

   if v = 'HTTP_AUTHORIZATION' then
   begin
        s:=GetHeaderLine(dt.LogHdrs,'AUTHORIZATION:');
        if s<>'' then
        begin
             StrPCopy(PAnsiChar(Buffer),s);
             Size:=length(s)+1;
             Result:=true;
        end;
        exit;
   end;

   if v = 'HTTP_COOKIE' then
   begin
        s:=GetHeaderLine(dt.LogHdrs,'COOKIE:');
        if s<>'' then
        begin
             StrPCopy(PAnsiChar(Buffer),s);
             Size:=length(s)+1;
             Result:=true;
        end;
        exit;
   end;

   if v = 'CONTENT_LENGTH' then
   begin
       s := GetHeaderLine (dt.LogHdrs, 'CONTENT-LENGTH:');
       if s <> '' then
       begin
            StrPCopy (PAnsiChar (Buffer), s);
            Size := Length (s) + 1;
            Result := true;
       end;
       exit;
   end;

   if v = 'CONTENT_TYPE' then
   begin
        s := GetHeaderLine (dt.LogHdrs, 'CONTENT-TYPE:');
        if s <> '' then
        begin
             StrPCopy (PAnsiChar (Buffer), s);
             Size := Length (s) + 1;
             Result := true;
        end;
        exit;
   end;

   if v = 'SERVER_PROTOCOL' then
   begin
        s := 'HTTP/1.0';
        StrPCopy (PAnsiChar (Buffer), s);
        Size := Length (s) + 1;
        Result := true;
        Exit;
   end;

   if v = 'QUERY_STRING' then
   begin
        StrPCopy(PAnsiChar(Buffer), dt.QueryString);
        Size := Length(dt.QueryString)+1;
        Result := true;
        Exit;
   end;
end;

function WriteClient(ConnID: HCONN; Buffer: Pointer; var Bytes: DWORD;
   dwReserved: DWORD ): BOOL stdcall;
var
   dt : TDataThread;
   s:string;
begin
   dt := TDataThread(ConnID);
   SetString(s,PAnsiChar(Buffer),Bytes);
   dt.ECB_Result := dt.ECB_Result + s;
   Result := True;
end;

function ReadClient(ConnID: HCONN; Buffer: Pointer; var Size: DWORD): BOOL stdcall;
var
   dt : TDataThread;
   n  : integer;
begin
     dt:=TDataThread(ConnID);
     n:=Size;
     Result:=true;
     with dt do
     begin
          n:=Recv(dt.DataSock,Buffer^,n,0);
          if n<0 then
          begin
               ErrMsg:=Format('GetData. Receive WSAError = %d', [WSAGetLastError]);
               Synchronize(ShowError);
               Result:=false;
          end;
     end;
     Size:=n;
end;

function ServerSupport(hConn: HCONN; HSERRequest: DWORD; Buffer: Pointer;
   Size: LPDWORD; DataType: LPDWORD ): BOOL stdcall;
const
   RESPREDIR : pansichar = 'HTTP/1.0 302 Moved temporarily'+CR+'Location: %s'+CR;
var
   Len : Cardinal;
   s:string;
begin
   case HSERRequest of
     HSE_REQ_SEND_RESPONSE_HEADER:
       begin
            Result := True;
            s := 'HTTP/1.0 '+PAnsiChar(Buffer)+CR+PChar(DataType);
            Len := Length(s);
            WriteClient(hConn, PAnsiChar (s), Len, 0);
       end;
     HSE_REQ_SEND_URL_REDIRECT_RESP:
       begin
            Result := True;
            s:=Format(RESPREDIR,[PAnsiChar(Buffer)]);
            Len := Length(s);
            WriteClient(hConn, PAnsiChar(s), Len, 0);
       end;

     else
       begin
            Result := False;
       end;
   end;
end;


procedure TDataThread.RunDLL(var Result: string);
type
   THttpExtensionProc = function(var ECB: TEXTENSION_CONTROL_BLOCK ):DWORD; stdcall;
   THttpGetExtensionVersion = function(var Ver: THSE_VERSION_INFO):DWORD; stdcall;
var
   hi       : THandle;
   Verfn    : THttpGetExtensionVersion;
   fn       : THttpExtensionProc;
   ecb      : TEXTENSION_CONTROL_BLOCK;
   Ver      : THSE_VERSION_INFO;
   tmp      : string;
begin
   DLLName := FileName; // for the GetServerVariable Function
   Result := '';


   // Change path to something understandable for windows environment.
   tmp:=FileName;
   FindRepAll(tmp,'/','\');



   hi:=Main.DLLStorage.LoadDLL(tmp);
   if hi<=0 then raise Exception.CreateFmt('There is currently no access to %s. Please try later.',[FileName]);
   Verfn := GetProcAddress(hi, 'GetExtensionVersion');
   if Assigned(Verfn) then
   begin
        Verfn(Ver);
        MainForm.Log(MainForm.Hits, RemoteHost,
           'Requesting ISAPI version:', inttostr(Ver.dwExtensionVersion shr 16)+','+inttostr(Ver.dwExtensionVersion and $FFFF),
           StrPas(Ver.lpszExtensionDesc),'');
   end;
   fn := GetProcAddress(hi, 'HttpExtensionProc');
   Assert(Assigned(fn), 'Unable to find HttpExtensionProc in DLL: ' + FileName);

   FillChar(ecb, sizeof(ecb), 0);
   ecb.cbSize                := sizeof(ecb);
   ecb.dwVersion             := HSE_VERSION_MAJOR shl 16 + HSE_VERSION_MINOR;
   ecb.ConnID                := THandle(Self);
   ecb.dwHttpStatusCode      := 200;
   ecb.lpszMethod            := PAnsiChar(Method);
   ecb.lpszQueryString       := PAnsiChar(QueryString);
   ecb.lpszPathInfo          := PAnsiChar(PathInfo);
   ecb.lpszPathTranslated    := PAnsiChar(PathInfo);
   ecb.cbAvailable           := Length(Data);
   ecb.lpbData               := PAnsiChar(Data);
   if ContentLength<=0 then
      ecb.cbTotalBytes       := ecb.cbAvailable
   else
      ecb.cbTotalBytes       := ContentLength;

   ecb.GetServerVariable     := GetServerVariable;
   ecb.WriteClient           := WriteClient;
   ecb.ReadClient            := ReadClient;
   ecb.ServerSupportFunction := ServerSupport;

   ECB_Result := '';
   fn(ecb);
   Result := ECB_Result;
end;

procedure TDataThread.SendFile(const FileName, FileType: string; var LogData:string);
var
   ms    : TMemoryStream;
   tmp   : string;
   hdr   : string;
   ct    : string;
   rc    : integer;
begin
   ms := TMemoryStream.Create;
   try
      // Change path to something understandable for windows environment.
      tmp:=FileName;
      FindRepAll(tmp,'/','\');

      if Copy(tmp, 1, 1)='\' then tmp := Copy(tmp, 2, Length(tmp));  // trim leading "\"

      // Load file into stream, and return the data.
      ms.LoadFromFile(tmp);
      ms.Position := 0;

      ct := '';
      if FileType<>'' then ct := 'Content-Type: ' + FileType + CR;

      // Send the Header
      hdr := Format(
         'HTTP/1.0 200 OK'+CR+
         ct+
         'Content-Length: %d'+CR+
         'Content:'+CR, [ms.Size]);
      LogData:=hdr+CR+'FileName: '+tmp;
      PutData(DataSock, hdr);

      // Send the file
      rc := send(DataSock, ms.Memory^, ms.Size, 0);
      Assert(rc>=0, Format('Unable to Send (WSAError = %x)', [WSAGetLastError]));

   finally
      ms.Free;
   end;
end;


end.
