unit ListenThread;

interface

uses
  Classes, SysUtils, WinSock;

type
   TListenThread = class(TThread)
   private
      { Private declarations }
   protected
      RemUser     : TSockAddr;
      len         : integer;
      datathreads : TList;
      ErrMsg      : string;
      procedure   Execute; override;
      procedure   UpdateHits;
      procedure   ShowError;
   public
      LisSock     : TSocket;
      constructor Create(CreateSuspended:boolean);
      destructor  Destroy; override;
   end;

implementation

uses Main, SockUtil, DataThread;

{ TListenThread }

constructor TListenThread.Create(CreateSuspended:boolean);
begin
     inherited;
     datathreads:=TList.create;
end;

destructor TListenThread.Destroy;
var
   i:integer;
begin
     with datathreads do
     begin
          for i:=0 to count-1 do
          begin
               TDataThread(items[i]).free;
          end;
          free;
     end;
     inherited;
end;

procedure TListenThread.Execute;
var
   DataSock : TSocket;
   dt       : TDataThread;
   he       : PHostEnt;
   ps       : PAnsiChar;
begin
   repeat
      try
         repeat
            len := sizeof(RemUser);
            DataSock := accept(LisSock, @RemUser, @len);
            if DataSock=INVALID_SOCKET then break;
            he := gethostbyaddr(@RemUser.sin_addr.S_addr, 4, RemUser.sin_family);

            dt := TDataThread.Create(True);
            datathreads.add(dt);

            dt.DataSock := DataSock;
            if he<>nil then
            begin
               dt.RemoteHost := he.h_name;
               dt.ServerName := string(ps);
            end
            else begin
            dt.RemoteHost := 'Unknown';
            dt.ServerName := '';
            end;
            dt.Resume;

            Synchronize(UpdateHits);
         until Terminated;
      except
         on e: Exception do begin
            ErrMsg := e.Message + ' (Listen Thread)';
            Synchronize(ShowError);
         end;
      end;
   until Terminated;
end;

procedure TListenThread.UpdateHits;
begin
   MainForm.UpdateHits;
end;

procedure TListenThread.ShowError;
begin
   MainForm.ErrorLab.Caption := ErrMsg;
end;

end.
