unit Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, Menus, ClipBrd, ComCtrls, ListenThread,

{$IF CompilerVersion>=23} // Delhi XE2
  Winapi.WinSock,
{$ELSE}
  WinSock,
{$IFEND}
// TrayIcon,
  StrUtil, SyncObjs;

// =========================================================================
// WWW Server for Delphi.
//
// Based on the open source and freeware version 1.25 of Ben Zieglers WWWServer
// of the same name.
//
// The WWWServer application is a small ISAPI compliant webserver which was
// created specially for easy debugging of the WABD/kbmWABD library. Source can
// be recompiled using Delphi 5.
//
// Copyright 1997,1998 for v. 1.25 and all previous versions Ben Ziegler (Zieglersoft)
// Copyright 1999,2000 for v. 2.xx and newer Kim Bo Madsen/Optical Services - Scandinavia
// All rights reserved.
//
// The copyright includes all files in this application, not only this file.
//
// LICENSE AGREEMENT
//
// You are allowed to use this server in any project for free.
// You are NOT allowed to claim that you have created this server or to
// copy its code into your own server and claim that it was your idea.
//
// Since this server is open source and freeware, all later incarnations _must_ also
// be open source and freeware.
//
// All the added functionality to the original 1.25 server is offered for free
// for your convinience, and the only thing I REQUIRE is to get an e-mail about what
// project this server (or derived versions) is used for.
//
// You dont need to state the name of the creators (Ben Ziegler and Kim Bo Madsen) in
// your software, although it would be appreciated if you do.
//
// If you find bugs or alter the server (f.ex. see suggested enhancements
// further down), please DONT just send the corrected/new code out on the internet,
// but instead send it to me, so I can put it into the official version. You will
// be acredited if you do so.
//
// Ben Ziegler has to my knowledge not claimed any restrictions which
// should violate this license agreement.
//
// DISCLAIMER
// By using this server or parts theirof you are accepting the full
// responsibility of the use. You are understanding that the authors (Ben Ziegler or
// Kim Bo Madsen) cant be made responsible in any way for any problems occuring
// using this server.
// You also recognize the authors (Ben Ziegler and Kim Bo Madsen) as the creators
// of this server and agrees not to claim otherwize!
//
// Please forward corrected versions (source code ONLY!), comments,
// and emails saying you are using it for this or that project to:
//            kbm@optical.dk
//
// New versions can be found at the kbmwabd community at
//    http://www.egroups.com/group/kbmwabd
//
// Please direct all support questions through the kbmwabd community.
// Support questions send directly to me will in general _not_ be answered.
//
// Suggestions for future enhancements:
//
//      - Help file,
//      - Support for more server variables.
//
// History:
// 1.25  Original version programmed and released as open source and freeware by Ben Ziegler.
//
// 2.00  Added several new things like more server variables, and termination
//       and initialization calls to ISAPI dlls and more.
//
// 2.07  Can now compilate under Delphi 5. Fixed a few small bugs.          KBM 25. Oct. 2000
//
// 2.08  Fixed RunDLL to handle PathInfo part eg. xxx.dll/pathinfo?somedata KBM 26. Dec. 2000
//       Fixed returning / not \ in path names to DLL's.
//
// 2.10  Fixed SCRIPT_NAME variable. Bug reported by                        KBM 28. Dec. 2000
//       Mauricio de Britto Longo (mlongo@ctupdate.com). Rewrote the DLL
//       loading and killing procedures. Removed dependencies of Bens Libs.
//       Added support for redirection.
//
// 2.11  Fixed bug truncating binary data from ISAPI application.           KBM 24. Jan. 2001
//
// 2.12  Added possibility to specify port number as a parameter to the commandline.
//       Syntax: wwwserver.exe [PortNo]
//       Portno is default 80 if not specified.                             KBM 26. Jan. 2001
//       Added support for Content-Length server variable, and
//       variable HSE_REQ_SEND_RESPONSE_HEADER by Ilkka Hyvrinen (ilkka.hyvarinen@fulcrum.fi)
//
// 2.13  Fixed minor bug when response was empty.                           KBM 8. May 2001
//       Removed TTrayIcon installation as a requirement by Miha Remec (mail@miharemec.com)
//       Fixed exception when logfile not existing.
//       Added support for POST method by Ilkka Hyvrinen (ilkka.hyvarinen@fulcrum.fi)
//
// 2.14  Improved ISAPI compatibility.                                      KBM 19. June 2001
//       Now correctly returning cbTotalBytes, cbAvailable, querystring and data.
//       Also correctly handling ReadClient ECB method.
//       Added support for QUERY_STRING and CONTENT_TYPE server variable.
//
// 2.15  Added support for the following new WAP orientated mimetypes:      KBM 8. July 2001
//       .WBMP='image/vnd.wap.wbmp', .WMLC='application/vnd.wap.wmlc',
//       .WMLS='text/vnd.wap.wmlscript', .WMLSC='application/vnd.wap.wmlscriptc'
//       Improved logging of files send.
//
// 2.16  Improved security by disallowing sending exe files. Further allows only
//       relative paths.                                                    KBM 23. July 2001

type
  TDLLStorage = class
  private
    FList:TStringList;
    FSyncObj:TCriticalSection;
  protected
    procedure Lock;
    procedure Unlock;
    procedure KillDLL(h:THandle);
  public
    constructor Create;
    destructor Destroy; override;
    function LoadDLL(DLLName:string):THandle;
    procedure UnloadAll;
  end;

  TMainForm = class(TForm)
    Panel1: TPanel;
    Label1: TLabel;
    Label2: TLabel;
    ServerName: TLabel;
    ServerAddress: TLabel;
    Label3: TLabel;
    NumHits: TLabel;
    Label4: TLabel;
    ErrorLab: TLabel;
    PopupMenu1: TPopupMenu;
    WWWServer1: TMenuItem;
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    Exit1: TMenuItem;
    Close1: TMenuItem;
    N1: TMenuItem;
    N2: TMenuItem;
    Shutdown1: TMenuItem;
    Tools1: TMenuItem;
    ClearLog1: TMenuItem;
    Label5: TLabel;
    PortLab: TLabel;
    CopyLogtoClipboard1: TMenuItem;
    ListView1: TListView;
    HitIcon: TImage;
    NormalIcon: TImage;
    miLogdatasend: TMenuItem;
    TrayIcon1: TTrayIcon;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure WWWServer1Click(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure Exit1Click(Sender: TObject);
    procedure Close1Click(Sender: TObject);
    procedure ClearLog1Click(Sender: TObject);
    procedure CopyLogtoClipboard1Click(Sender: TObject);
    procedure ServerAddressDblClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure ListView1DblClick(Sender: TObject);
    procedure miLogdatasendClick(Sender: TObject);
  protected
//    TrayIcon1: TTrayIcon;
    procedure  OnQueryEndSession(var Msg: TMessage); message WM_QUERYENDSESSION;
  public
    { Public declarations }
    LisSock    : TSocket;
    LisThread  : TListenThread;
    Hits       : integer;
    Port       : integer;
    procedure  Start_Server;
    procedure  Stop_Server;
    procedure  UpdateHits;
    procedure  Log(Hit: integer; const RemoteHost, Msg, Data, Hdrs, ErrMsg: string);
    procedure  LogToFile(Hit: integer; const RemoteHost, Msg, Data, Hdrs, ErrMsg: string);
  end;

var
  MainForm: TMainForm;
  DLLStorage:TDLLStorage;

implementation

uses HitDetail;

{$R *.DFM}

// DLL storage.
constructor TDLLStorage.Create;
begin
     inherited;
     FList:=TStringList.Create;
     FSyncObj:=TCriticalSection.Create;
end;

destructor TDLLStorage.Destroy;
begin
     UnloadAll;
     FList.Free;
     FSyncObj.Free;
end;

procedure TDLLStorage.Lock;
begin
     FSyncObj.Enter;
end;

procedure TDLLStorage.Unlock;
begin
     FSyncObj.Leave;
end;

function TDLLStorage.LoadDLL(DLLName:string):THandle;
var
   i:integer;
   h:THandle;
begin
     Lock;
     DLLName:=UpperCase(DLLName);
     try
        // Try to find existing instance.
        for i:=FList.Count-1 downto 0 do
            if FList.Strings[i]=DLLName then
            begin

                 // Check to see if it is actually still loaded. If not, reload.
               {$IFDEF VER230}
                 h:=GetModuleHandle(PWideChar(DLLName));
               {$ELSE}
                 h:=GetModuleHandle(PChar(DLLName));
               {$ENDIF}
                 if h=0 then
                 begin
                      FList.Delete(i);
                      continue;
                 end;

                 // Ok, module in memory, return handle.
                 FList.Objects[i]:=TObject(h);
                 Result:=h;
                 exit;
            end;

        // No valid libaries in memory found, load library.
      {$IFDEF VER230}
        h:=LoadLibrary(PWideChar(DLLName));
      {$ELSE}
        h:=LoadLibrary(PChar(DLLName));
      {$ENDIF}                    
        Assert(h<>0, 'Unable to load DLL: '+DLLName);
        FList.AddObject(DLLName,TObject(h));
        Result:=h;
     finally
        Unlock;
     end;
end;

procedure TDLLStorage.KillDLL(h:THandle);
type
   THttpTerminateExtension = function(dwFlags:DWORD):boolean; stdcall;
var
   Termfn:THttpTerminateExtension;
 {$IF CompilerVersion>=23}
   buffer:array [0..64] of widechar;
 {$ELSE}
   buffer:array [0..64] of ansichar;
 {$IFEND}
begin
     Termfn:=GetProcAddress(h,'TerminateExtension');
     MainForm.Log(MainForm.Hits, '','Terminating ISAPI handle:', inttostr(h),'','');
     if Assigned(Termfn) then
     begin
          MainForm.Log(MainForm.Hits, '','Asking ISAPI for shutdown on handle:', inttostr(h),'','');
          Termfn(0);
     end;

     // To be on the safe side, check if library died by itself. If not kill it.
   {$IF CompilerVersion>=23 }
     if GetModuleFileName(h,buffer,63)>0 then
   {$ELSE}
     if GetModuleFileName(h,buffer,63)>0 then
   {$IFEND}
        FreeLibrary(h);
end;

procedure TDLLStorage.UnloadAll;
var
   i:integer;
   h:THandle;
begin
     Lock;
     try
        for i:=0 to FList.Count-1 do
        begin
             // Check if library in memory still, then free it.
           {$IF CompilerVersion>=23}
             h:=GetModuleHandle(PWideChar(FList.Strings[i]));
           {$ELSE}
             h:=GetModuleHandle(PAnsiChar(FList.Strings[i]));
           {$IFEND}
             if h>0 then KillDLL(h)
             else MainForm.Log(MainForm.Hits, '','Library unloaded prematurely: ',FList.Strings[i],'','');
        end;
        FList.Clear;
     finally
        Unlock;
     end;
end;

// Main form stuff.

procedure TMainForm.Start_Server;
const
   CR       = #13#10;
var
   wd       : TWSAdata;
   rc       : integer;
   optval   : integer;
   sa       : TSockAddr;
   buf      : array[0..MAX_PATH] of ansichar;
   buf2     : PAnsiChar;
   in_addr  : PInAddr;
   he       : PHostEnt;
begin
   // Startup WinSock
   rc := WSAStartup($0101, wd);
   Assert(rc=0, 'Unable to Start WinSock - error '+IntToStr(WSAGetLastError()));

   LisSock := socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
   Assert(LisSock<>-1, 'Socket error '+IntToStr(WSAGetLastError()));
   // ----- XE2 - Win 7
//   optval := 1;
//   rc := setsockopt(LisSock, SOL_SOCKET, SO_BROADCAST, PAnsiChar(@optval), sizeof(optval));
   // -----

   FillChar(sa, sizeof(sa), 0);
   sa.sin_family      := AF_INET;
   sa.sin_port        := htons(Port);
   sa.sin_addr.S_addr := INADDR_ANY;
   rc := bind(LisSock, sa, sizeof(sa));
   if rc<>0 then begin
      MessageDlg('FATAL ERROR: Unable to bind to Server Socket! error '
              +IntToStr(WSAGetLastError())+' '#13#10+'Shutting Down!',
         mtError, [mbOK], 0);
      Application.Terminate;
      exit;
   end;
   // Assert(rc=0, 'Unable to bind');

   rc := gethostname(buf, sizeof(buf));
   Assert(rc=0, 'Unable to GetHostName');
   he := gethostbyname(buf);
   if he=nil then begin
      MessageDlg('FATAL ERROR:  Unable to GetHostByName'#13 +
      'Be sure to start Dial-Up Networking BEFORE you start the Server',
      mtError, [mbOk], 0);
      Application.Terminate;
      exit;
   end;
   Assert(he<>nil, 'Unable to GetHostByName');
   in_addr := PInAddr(he.h_addr^);
   buf2 := inet_ntoa(in_addr^);

   ServerName.Caption := he.h_name;
   ServerAddress.Caption := buf2;
   PortLab.Caption := IntToStr(Port);

   // Listen
   rc := listen(LisSock, 10);
   Assert(rc=0, 'Unable to Listen');

   LisThread := TListenThread.Create(True);
   LisThread.LisSock := LisSock;
   LisThread.Resume;
end;


procedure TMainForm.Stop_Server;
var
   rc : integer;
begin
   // Stop the listen thread.
   LisThread.Terminate;

   // Close the Socket
   rc := closesocket(LisSock);
   Assert(rc=0, 'Unable to Close LisSocket');

   // Wait for the thread to actually stop.
   LisThread.WaitFor;
   LisThread.Free;

   // Shutdown WinSock
   rc := WSACleanup;
   Assert(rc=0, 'Unable to Shutdown WinSock');
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
// use TTrayIcon from VCL,  XE2
   with TrayIcon1 do
   begin
//        Active := True;
//        ShowDesigning := False;
        visible := true;
        BalloonHint := 'WWW server';
        OnDblClick := WWWServer1Click;
        PopupMenu := PopupMenu1;
   end;

   DLLStorage:=TDLLStorage.Create;

   // Check if argument, then use it as port number.
   Port := 80;
   if ParamCount>0 then
      try
         Port:=strtoint(ParamStr(1));
      except
         ShowMessage('Syntax: wwwserver.exe [PortNo]'+#10#13+'Using default port 80');
         Port:=80;
      end;

   Hits := -1;
   UpdateHits;
   Start_Server;
   Log(0, '', 'Startup', '', '', '');
   TrayIcon1.Icon := NormalIcon.Picture.Icon;
   Visible := False;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
   Stop_Server;
   DLLStorage.free;
   TrayIcon1.Visible:=False;
   Sleep(500);
end;

procedure TMainForm.UpdateHits;
begin
   Hits := Hits + 1;
   NumHits.Caption := Format('%1.0n', [Hits+0.0]);
end;

procedure TMainForm.WWWServer1Click(Sender: TObject);
begin
   Show;
   WindowState := wsNormal;
   BringToFront;
   SetFocus;
end;

procedure TMainForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
   CanClose := False;
   Hide;
end;

procedure TMainForm.Exit1Click(Sender: TObject);
begin
   Application.Terminate;
end;

procedure TMainForm.Close1Click(Sender: TObject);
begin
   Close;
end;

procedure TMainForm.ClearLog1Click(Sender: TObject);
begin
   ListView1.Items.Clear;
end;

procedure TMainForm.LogToFile(Hit: integer; const RemoteHost, Msg, Data, Hdrs, ErrMsg: string);
const
   LOGFILE = 'WWWLOG.TXT';
var
   fs : TFileStream;
   sl : TStringList;
   s  : string;
   ast: ansistring;
begin
  if FileExists(LOGFILE) then
    fs := TFileStream.Create(LOGFILE, fmOpenReadWrite or fmShareDenyWrite)
  else
    fs := TFileStream.Create(LOGFILE, fmCreate or fmOpenReadWrite or fmShareDenyWrite);
   fs.Seek(0, 2); // Seek EOF

   sl := TStringList.Create;
   sl.Add(IntToStr(Hit));
   sl.Add(RemoteHost);
   sl.Add(Msg);
   sl.Add(Data);
   sl.Add(Hdrs);
   sl.Add(ErrMsg);
   s := sl.CommaText;
   FindRep(s, #13#10, '\n');
   s := s + #13#10;
   sl.Free;

   ast := s;
   fs.Write(ast[1], Length(ast));
   fs.Free;
end;

procedure TMainForm.Log(Hit: integer; const RemoteHost, Msg, Data, Hdrs, ErrMsg: string);
var
   li : TListItem;
begin
   LogToFile(Hit, RemoteHost, Msg, Data, Hdrs, ErrMsg);

   li := ListView1.Items.Add;
   li.Caption := FormatDateTime('ddd dd mmm hh:nn:ss', Now);
   li.SubItems.Add(IntToStr(Hit));
   li.SubItems.Add(RemoteHost);
   li.SubItems.Add(Msg);
   li.SubItems.Add(Data);
   li.SubItems.Add(Hdrs);
   li.SubItems.Add(ErrMsg);
   li.MakeVisible(False);

   if ListView1.Items.Count > 5000 then begin
      // ListView1.Items.SaveToFile('TmpWWWLog.txt');
      ListView1.Items.Clear;
   end;

   if not Visible then
      TrayIcon1.Icon := HitIcon.Picture.Icon;
end;

procedure TMainForm.CopyLogtoClipboard1Click(Sender: TObject);
begin
   // Clipboard.SetTextBuf(PChar(ListView1.Items.Text));
end;

procedure TMainForm.ServerAddressDblClick(Sender: TObject);
begin
   ClipBoard.AsText := ServerAddress.Caption;
end;

procedure TMainForm.FormShow(Sender: TObject);
begin
   TrayIcon1.Icon := NormalIcon.Picture.Icon;
end;

procedure TMainForm.ListView1DblClick(Sender: TObject);
var
   li : TListItem;
begin
   li := ListView1.Selected;
   if li=nil then exit;

   with HitDetailForm do begin
      LV := ListView1;
      ShowModal;
   end;
end;

procedure TMainForm.OnQueryEndSession(var Msg: TMessage);
begin
   Msg.Result := 1;
   Application.Terminate;
end;


procedure TMainForm.miLogdatasendClick(Sender: TObject);
begin
     miLogdatasend.Checked:=not miLogdatasend.Checked;
end;

end.
