unit SockUtil;

interface

uses SysUtils, WinSock;

const
   CR = #13#10;

   // Basic Functions
   function  GetData(Sock: TSocket): string;
   procedure PutData(Sock: TSocket; const s: string);
   procedure PutDataNoCR(Sock: TSocket; const s: string);
   function  GetDataCR(Sock: TSocket; Echo, Password: boolean): string;

   // Server Functions
   procedure StartListenSocket(Port: integer; var LisSock: TSocket; var ServerName, ServerAddress: string);


implementation

uses Dialogs, Windows;

function GetData(Sock: TSocket): string;
var
   buf : array[0..8191] of ansichar;
   rc  : integer;
begin
     Result := '';

     // Get max 8kb data.
     rc := recv(Sock, buf, sizeof(buf), 0);
     if rc<0 then raise Exception.CreateFmt('GetData. Receive WSAError = %d', [WSAGetLastError]);
     SetString(Result,buf,rc);
end;


function GetDataCR(Sock: TSocket; Echo, Password: boolean): string;
var
   tmp : string;
   i   : integer;
begin
     Result := '';
     while Pos(#13#10, Result)=0 do
     begin
          tmp := GetData(Sock);
          Result := Result + tmp;
          if Password then
             for i := 1 to Length(tmp) do
                 if (tmp[i]<>#13) and (tmp[i]<>#10) then tmp[i] := '*';
          if Echo then PutDataNoCR(Sock, tmp);
     end;
     Result := Trim(Result);
end;


procedure PutDataNoCR(Sock: TSocket; const s: string);
var
   rc  : integer;
//XE2 requires ansistring!
   tmp : ansistring;
begin
     tmp := s;
     rc := send(Sock, tmp[1], Length(tmp), 0);
     if rc<0 then exit; //raise Exception.CreateFmt('Send WSAError = %d', [rc]);
end;

procedure PutData(Sock: TSocket; const s: string);
begin
     PutDataNoCR(Sock, s+#13#10);
end;

procedure StartListenSocket(Port: integer; var LisSock: TSocket; var ServerName, ServerAddress: string);
const
   CR       = #13#10;
var
   wd       : TWSAdata;
   rc       : integer;
   sa       : TSockAddr;
   buf      : array[0..MAX_PATH] of ansichar;
   buf2     : PansiChar;
   in_addr  : PInAddr;
   he       : PHostEnt;
begin
     // Startup WinSock
     rc := WSAStartup($0101, wd);
     Assert(rc=0, 'Unable to Start WinSock');

     LisSock := socket(AF_INET, SOCK_STREAM, 0);

     FillChar(sa, sizeof(sa), 0);
     sa.sin_family      := AF_INET;
     sa.sin_port        := htons(PORT);
     sa.sin_addr.S_addr := INADDR_ANY;
     rc := bind(LisSock, sa, sizeof(sa));
     if rc<>0 then
        raise Exception.Create('FATAL ERROR: Unable to bind to Server Socket!'#13#10'Shutting Down!');

     rc := gethostname(buf, sizeof(buf));
     Assert(rc=0, 'Unable to GetHostName');
     he := gethostbyname(buf);
     if he=nil then
        raise Exception.Create('FATAL ERROR:  Unable to GetHostByName'#13 +
              'Be sure to start Dial-Up Networking BEFORE you start the Server');

     in_addr := PInAddr(he.h_addr^);
     buf2 := inet_ntoa(in_addr^);

     ServerName := he.h_name;
     ServerAddress := buf2;

     // Listen
     rc := listen(LisSock, 10);
     Assert(rc=0, 'Unable to Listen');
end;



end.
