unit WABD_Admin;
// In the HTML pages uses the Javascript light framework W2UI v 1.5
// So the url handlers expect to receive and send data in W2UI JSON format
interface

uses
    Windows, Messages, SysUtils, Classes,
    Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs,
    superobject,
    WABD_Objects, WABD_Utils,  WABD_Common,
    WABD_Attributes, WABD_Statistics, WABD_Setup,
    WABD_Request, WABD_Response;

type
  TListOption = record
    id: string;
    text: string;
  end;
  TSetupData = record
    eAdminUser             : string;
    eAdminPassword         : string;
    eAdminPasswordRetype   : string;
    chbGarbageCollection   : Boolean;
    eGarbageInterval       : Integer;
    chbGatherStatistics    : Boolean;
    chbNetworkStats        : Boolean;
    eTimeout               : Integer;
    eAdminText             : string;
    eMaxIdenticalUser      : Integer;
	  eMaxSessions           : Integer;
	  eRouteWhen             : TListOption;   // list option
	  eRouteHow              : TListOption;   // list option
	  mRouteSites            : string;
  end;
  TRowStyle = record
    style: string;
  end;
  TStatRow = record
    recid: Integer;
    c0: string;
    c1: string;
    c2: string;
    c3: string;
    w2ui: TRowStyle;
  end;
  TStatData = array[0..77] of TStatRow;
  TStatTotals = record
    bufferSize: Integer;
    sum: string;
    average: string;
    actualSize: Integer;
    measurements: Integer;
    min: string;
    max: string;
  end;


  TWABDAdmin = class(TDataModule)
    WABDAdminSession: TWABD_Session;
    wabdHTML1: TWABD_HTMLFile;
    wabdSetup1: TWABD_Setup;
    wabdHtmlLogoff: TWABD_HTML;
    wabdHTMLAuth: TWABD_HTMLFile;
    procedure WABDAdminSessionBeforeProcessRequest(Session: TWABD_Session);
  private
    { Private declarations }
    jsonReq: ISuperObject;

    sessionsId: array of Integer;
    currStat: string;
    function cmdExtract(request: TWABD_CustomRequest): string;
    function jsonErrResp(mes: string): string;
    function jsonGridResp(recCount: Integer; records: ISuperObject;
          otherData: ISuperObject = nil): string;
    function jsonFormResp(recData: ISuperObject;
          otherData: ISuperObject = nil): string;
    procedure BuildStats(grpname:string; var statData: TStatData;
              var statTotals: TStatTotals);
    function SaveData(const sd: TSetupData; var meserr: string): Boolean;
    procedure SaveStatsToFile();
    procedure LoadStatsFromFile();
    procedure ClearAllStats();
    procedure ClearStat();
  public
    { Public declarations }
    constructor Create(AOwner:TComponent); override;


    [UrlAttribute('/sauth','POST',utInternal)]
    procedure Authorize(request: TWABD_CustomRequest;
                    response: TWABD_CustomResponse);
    [UrlAttribute('/adm','GET',utInternal)]
    procedure Admpage(request: TWABD_CustomRequest;
                    response: TWABD_CustomResponse);
    [UrlAttribute('/killa','POST',utInternal)]
    procedure KillAllSessions(request: TWABD_CustomRequest;
                    response: TWABD_CustomResponse);
    [UrlAttribute('/sessdel','POST',utInternal)]
    procedure KillSession(request: TWABD_CustomRequest;
                    response: TWABD_CustomResponse);
    [UrlAttribute('/sess','POST',utInternal)]
    procedure BuildSessionList(request: TWABD_CustomRequest;
                    response: TWABD_CustomResponse);
    [UrlAttribute('/sendtx','POST',utInternal)]
    procedure SendTextSessions(request: TWABD_CustomRequest;
                    response: TWABD_CustomResponse);
    [UrlAttribute('/setup','POST',utInternal)]
    procedure BuildSetup(request: TWABD_CustomRequest;
                    response: TWABD_CustomResponse);
    [UrlAttribute('/setups','POST',utInternal)]
    procedure SaveSetup(request: TWABD_CustomRequest;
                    response: TWABD_CustomResponse);
    [UrlAttribute('/stat','POST',utInternal)]
    procedure LoadStats(request: TWABD_CustomRequest;
                    response: TWABD_CustomResponse);
    [UrlAttribute('/statact','POST',utInternal)]
    procedure StatActions(request: TWABD_CustomRequest;
                    response: TWABD_CustomResponse);
    [UrlAttribute('/term','POST',utInternal)]
    procedure TerminateDLL(request: TWABD_CustomRequest;
                    response: TWABD_CustomResponse);
  end;

var
  WABDAdmin: TWABDAdmin;

implementation
{$R *.DFM}

var
  STAT_GROUPS: array [0..3] of string =('s1', 's2', 's3', 's4');


constructor TWABDAdmin.Create(AOwner:TComponent);
begin
     inherited;

     WABDAdminSession.urlHandlers:=Self;
     currStat:=WABD_STATGRP_RESPONSE;
end;

procedure TWABDAdmin.WABDAdminSessionBeforeProcessRequest(
  Session: TWABD_Session);
begin
  if wabdSetup1.LocalFilePath='' then
     wabdSetup1.LocalFilePath:=WABDAdminSession.SessionMgr.ModulePath;
end;

function TWABDAdmin.jsonErrResp(mes: string): string;
begin
  Result:= '{"status": "error", "message": "'+mes +'" } ';
end;

function TWABDAdmin.jsonGridResp(recCount: Integer; records: ISuperObject;
      otherData: ISuperObject): string;
begin
  Result:='{"status": "success", "total": "'
      +IntToStr(recCount)
      +'", "records": '+records.AsJSon(false, True);
  if (Assigned(otherData)) then
    Result:=Result+', "serverdata": '+ otherData.AsJSon(false, True)+ ' }'
  else
    Result:=Result+' }';

end;

function TWABDAdmin.jsonFormResp(recData: ISuperObject;
            otherData: ISuperObject): string;
begin
  Result:='{"status": "success", "record": '+recData.AsJSon(false, True);
  if (Assigned(otherData)) then
    Result:=Result+', "serverdata": '+ otherData.AsJSon(false,True)+' }'
  else
    Result:=Result+' }';
end;

function TWABDAdmin.cmdExtract(request: TWABD_CustomRequest): string;
begin
try
  jsonReq:=TSuperObject.ParseString(PWideChar(request.Query.Values['request']),false);
  Result:=jsonReq.S['cmd'];
except
  Result:='';
end;
end;

procedure TWABDAdmin.Authorize(request: TWABD_CustomRequest; response: TWABD_CustomResponse);
var
  user, pw: string;
begin
   if cmdExtract(request)<>'save' then Exit;

   user:=jsonReq.O['record'].S['f1'];
   pw:=jsonReq.O['record'].S['f2'];
   response.ContentType:='application/json';
   with wabdHTMLAuth.Session.SessionMgr.Admin do
   begin
    if (user=AdminUser) and (pw=AdminPassword) then
    begin
        wabdHTMLAuth.Session.authenticated:=True;
        wabdHTMLAuth.Session.UserName:=AdminUser;
//        wabdHTMLAuth.Session.Password:=AdminPassword;
        response.Response.Text:=jsonFormResp(TSuperObject.ParseString('{}',True), nil);
    end
    else
        response.Response.Text:=jsonErrResp('User or password not valid');
   end;

end;

procedure TWABDAdmin.Admpage(request: TWABD_CustomRequest; response: TWABD_CustomResponse);
begin
  response.Response.Text:=wabdHTML1.Object_To_HTML;
end;

procedure TWABDAdmin.KillAllSessions(request: TWABD_CustomRequest;
             response: TWABD_CustomResponse);
var
   i:integer;
   lst:TTempSessionList;
   ses:TWABD_Session;
begin

     with wabdHtml1.Session.SessionMgr do
     begin
          lst:=SessionList.LockList;
          try
             for i:=lst.Count-1 downto 0 do
             begin
                  ses:=lst.Items[i];
                  if ses.SessionID<>WABDAdminSession.SessionID then
                  begin
                       ses.LogOff;
                       CheckLogOff(ses);
                  end;
             end;
          finally
             SessionList.UnlockList;
          end;
     end;
end;

procedure TWABDAdmin.KillSession(request: TWABD_CustomRequest;
      response: TWABD_CustomResponse);
var
   i:integer;
   lst:TTempSessionList;
   ses:TWABD_Session;
   sesId: Integer;
   rec: ISuperObject;
   logpage: ISuperObject;
begin
     if cmdExtract(request)<>'delete' then Exit;


     sesId:=sessionsId[jsonReq.A['selected'].I[0]];

     response.ContentType:='application/json';
     rec:=TSuperObject.ParseString('{}',true);
     logpage:=nil;
     with wabdHtml1.Session.SessionMgr do
     begin
          lst:=SessionList.LockList;
          response.Response.Text:=jsonErrResp('Id not found');
          try
             for ses in lst do
             begin
                  if ses.SessionID=SesID then
                  begin
                       ses.LogOff;
                       if SesID<>WABDAdminSession.sessionID then CheckLogOff(ses);
                       if SesID=WABDAdminSession.sessionID then
                       begin
                                 logpage:=TSuperObject.ParseString('{}',true);
                                 logpage.S['logoff']:=wabdHtmlLogoff.Object_To_HTML;
                       end ;
                       response.Response.Text:=jsonFormResp(rec, logpage);
                       break;
                  end;
             end;
          finally
             SessionList.UnlockList;
          end;
     end;
end;

procedure TWABDAdmin.BuildSetup(request: TWABD_CustomRequest; response: TWABD_CustomResponse);
var
  ctx: TSuperRttiContext;
  js: ISuperObject;
  setupData: TSetupData;
  ff: string;
begin
     if cmdExtract(request)<>'get' then Exit;


     response.ContentType:='application/json';
     with WABDAdminSession.SessionMgr do
     begin
          setupData.eAdminUser :=Admin.AdminUser;
          setupData.eAdminPassword :=Admin.AdminPassword;
          setupData.eAdminPasswordRetype :='';   // retyped password
          setupData.chbGarbageCollection :=GarbageCollection;
          setupData.eGarbageInterval :=CheckTimeOutInterval;
          setupData.chbGatherStatistics :=GatherStatistics;
          setupData.chbNetworkStats :=NetworkStatistics;
          setupData.eTimeout :=DefaultSessionTimeout;
          setupData.eAdminText :=WebAdministrator;
          setupData.eMaxSessions :=MaxConcurrentSessions;
          setupData.eMaxIdenticalUser :=MaxIdenticalUser;
          setupData.eRouteWhen.id := IntToStr(ord(RouteWhen));
          setupData.eRouteWhen.text:='';
          setupData.eRouteHow.id :=IntToStr(ord(RouteHow));
          setupData.eRouteHow.text:='';
          setupData.mRouteSites :=RouteSites.Text;
     end;
     ctx := TSuperRttiContext.Create;
     try
        js:=ctx.AsJson<TSetupData>(setupdata);
     finally
       ctx.Free;
     end;
    response.Response.Text:=jsonFormResp(js);
end;

procedure TWABDAdmin.SaveSetup(request: TWABD_CustomRequest; response: TWABD_CustomResponse);
var
  ctx: TSuperRttiContext;
  js: ISuperObject;
  setupData: TSetupData;
  meser: string;
  mes: string;
begin
     if cmdExtract(request)<>'save' then Exit;


     response.ContentType:='application/json';

     try
       js:=TSuperObject.ParseString(PWideChar(jsonReq.S['record']),true);
       ctx := TSuperRttiContext.Create;
       try
          setupData:=ctx.AsType<TSetupData>(js);
          // save data
          SaveData(setupData, mes);
          meser:=mes;
       finally
         ctx.Free;
       end;
       if meser<>'' then
          response.Response.Text:=jsonErrResp(meser)
       else
       begin
          js:=TSuperObject.ParseString('{}',true);
          response.Response.Text:=jsonFormResp(js);
       end;
     except
       response.Response.Text:=jsonErrResp('Save error');
     end;
end;

function TWABDAdmin.SaveData(const sd: TSetupData; var meserr: string): Boolean;
begin
   meserr:='';
   with WABDAdminSession.SessionMgr do
   begin
// Setup admin username/password.
        if (length(sd.eAdminUser)>0) then
        begin
             Admin.AdminUser:=Trim(sd.eAdminUser);
        end
        else
            meserr:='Username not given ';
        if (sd.eAdminPassword<>Admin.AdminPassword) then
        begin
            if (sd.eAdminPassword = sd.eAdminPasswordRetype) then
            begin
                 Admin.AdminPassword:=Trim(sd.eAdminPassword);
            end
            else
                meserr:='Password retyped incorrectly.';
        end;

// Setup garbage collection/stat collection.
         if (sd.eGarbageInterval<10) or (sd.eGarbageInterval>9999) then
               meserr:=meserr+#10'Invalid garbage interval ';
         if (sd.eTimeout<-1) then
               meserr:=meserr+#10'Invalid session timeout ';

        CheckTimeOutInterval:=sd.eGarbageInterval;
        GarbageCollection:=sd.chbGarbageCollection;
        GatherStatistics:=sd.chbGatherStatistics;
        NetworkStatistics:=sd.chbNetworkStats;
        DefaultSessionTimeout:=sd.eTimeout;
        WebAdministrator:=trim(sd.eAdminText);
        MaxConcurrentSessions:=sd.eMaxSessions;
        MaxIdenticalUser:=sd.eMaxIdenticalUser;
        RouteWhen:=TWABD_RouteWhen(strtoint(sd.eRouteWhen.id));
        RouteHow:=TWABD_RouteHow(strtoint(sd.eRouteHow.id));
        RouteSites.Text:=sd.mRouteSites;
     end;
end;

procedure TWABDAdmin.BuildSessionList(request: TWABD_CustomRequest;
      response: TWABD_CustomResponse);
type
  TSessionData = record
    recid: Integer;
    sessionid: Integer;
    user: string;
    browser: string;
    loggedon: string;
    lastaccess: string;
    timeout: Integer;
  end;
  TSessions = array of TSessionData;
var
   i:integer;
   lst:TTempSessionList;
   sesdata: TSessions;
   ctx: TSuperRttiContext;
   js: ISuperObject;
begin
     if cmdExtract(request)<>'get' then Exit;


     response.ContentType:='application/json';

     with wabdHtml1.Session.SessionMgr do
     begin
          lst:=SessionList.LockList;
          SetLength(sessionsId,0);
          SetLength(sessionsId, lst.Count);
          SetLength(sesdata, lst.Count);
          try
             for i:=0 to lst.Count-1 do
               with (lst.Items[i]) do
               begin
                      sessionsId[i]:=SessionID;
                      sesdata[i].recid:=i;
                      sesdata[i].sessionid:=SessionID;
                      sesdata[i].user:=UserName;
                      sesdata[i].browser:=WABD_Browser[Browser]+' '+BrowserVersion ;
                      sesdata[i].loggedon:=DateTimeToStr(CreateTime) ;
                      sesdata[i].lastaccess:=DateTimeToStr(LastAccess);
                      sesdata[i].timeout:=TimeOutLength-trunc(24*60*60*(now-LastAccess));
               end;
             ctx := TSuperRttiContext.Create;
             js:=ctx.AsJson<TSessions>(sesdata);
             response.Response.Text:=jsonGridResp(lst.Count, js);
          finally
             ctx.Free;
             SessionList.UnlockList;
          end;
     end;
end;

procedure TWABDAdmin.SendTextSessions(request: TWABD_CustomRequest;
           response: TWABD_CustomResponse);
var
   lst:TTempSessionList;
   s:string;
   i:integer;
   ses: TWABD_Session;
begin
     s:=cmdExtract(request);
     s:=jsonReq.O['record'].S['eInfo'];
     with WABDAdminSession.SessionMgr do
     begin
          Info:=s;
          lst:=SessionList.LockList;
          try
               for ses in lst do
                      ses.Info:=s;
          finally
             SessionList.UnlockList;
          end;
     end;
end;



procedure TWABDAdmin.BuildStats(grpname:string; var statData: TStatData;
             var statTotals: TStatTotals);
var
   i,j:integer;
   grp:TWABD_SesStatGroup;
   avg:double;
   header:string;
   lst,grplst:TList;
const
     TITLE_STYLE = 'font-weight: bold; background-color: #CBFDC8';
begin
     for i:=0 to High(statData) do
      statData[i].recid:=i;

     with WABDAdminSession.SessionMgr do
     begin
          if not(Assigned(Stats)) then
          begin
            statData[j].c0:='';
            Exit;
          end;
          i:=Stats.IndexOf(grpname);

          lst:=Stats.LockList;
          try
              grp:=TWABD_SesStatGroup(lst.Items[i]);


              // Setup table.
              grplst:=grp.LockList;
              try
                  j:=0;

                  // Build hourly list.
                  statData[j].c0:='HOUR';
                  statData[j].w2ui.style:=TITLE_STYLE;
                  inc(j);
                  for i:=0 to 23 do
                  begin
                       statData[j].c0:=inttostr(i);
                       statData[j].c1:=inttostr(grp.HourlyCount[i]);
                       statData[j].c2:=format('%.2f',[grp.HourlyValues[i]]);
                       if grp.HourlyCount[i]>0 then
                          statData[j].c3:=format('%.2f',[grp.HourlyValues[i] / grp.HourlyCount[i]])
                       else
                           statData[j].c3:='N/A';
                       inc(j);
                  end;

                  // Build day list.
                  statData[j].c0:='DAY';
                  statData[j].w2ui.style:=TITLE_STYLE;
                  inc(j);
                  for i:=1 to 31 do
                  begin
                       statData[j].c0:=inttostr(i);
                       statData[j].c1:=inttostr(grp.DailyCount[i]);
                       statData[j].c2:=format('%.2f',[grp.DailyValues[i]]);
                       if grp.DailyCount[i]>0 then
                          statData[j].c3:=format('%.2f',[grp.DailyValues[i] / grp.DailyCount[i]])
                       else
                           statData[j].c3:='N/A';
                       inc(j);
                  end;

                  // Build monthly list.
                  statData[j].c0:='MONTH';
                  statData[j].w2ui.style:=TITLE_STYLE;
                  inc(j);
                  for i:=1 to 12 do
                  begin
                       statData[j].c0:=FormatSettings.LongMonthNames[i];
                       statData[j].c1:=inttostr(grp.MonthlyCount[i]);
                       statData[j].c2:=format('%.2f',[grp.MonthlyValues[i]]);
                       if grp.MonthlyCount[i]>0 then
                          statData[j].c3:=format('%.2f',[grp.MonthlyValues[i] / grp.MonthlyCount[i]])
                       else
                           statData[j].c3:='N/A';
                       inc(j);
                  end;

                  // Build weekday list.
                  statData[j].c0:='WEEKDAY';
                  statData[j].w2ui.style:=TITLE_STYLE;
                  inc(j);
                  for i:=1 to 7 do
                  begin
                       statData[j].c0:=FormatSettings.LongDayNames[i];
                       statData[j].c1:=inttostr(grp.DayCount[i]);
                       statData[j].c2:=format('%.2f',[grp.DayValues[i]]);
                       if grp.DayCount[i]>0 then
                          statData[j].c3:=format('%.2f',[grp.DayValues[i] / grp.DayCount[i]])
                       else
                           statData[j].c3:='N/A';
                       inc(j);
                  end;



                  // Show group values.

                  statTotals.bufferSize:=grp.BufferSize;

                  statTotals.actualSize:=grplst.Count;
                  statTotals.measurements:=grp.TotalCount;
                  if grp.TotalCount>0 then
                     avg:=grp.Sum / grp.TotalCount
                  else
                     avg:=0;
                  statTotals.average:=format('%.2f',[avg]);
                  if grp.Min>9999999 then
                    statTotals.min:='0.00'
                  else
                    statTotals.min:=format('%.2f',[grp.Min]);
                  statTotals.max:=format('%.2f',[grp.Max]);
                  statTotals.sum:=format('%.2f',[grp.Sum]);
              finally
                  grp.UnlockList;
              end;
          finally
              Stats.UnlockList;
          end;
     end;
end;


procedure TWABDAdmin.LoadStats(request: TWABD_CustomRequest; response: TWABD_CustomResponse);
var
  group: string;
  statData: TStatData;
  statTotals: TStatTotals;
  jsonRecords: ISuperObject;
  jsonTotals: ISuperObject;
  ctx: TSuperRttiContext;
  i: Integer;
begin
     if cmdExtract(request)<>'get' then Exit;

     group:=jsonReq.S['group'];
     for i:=0 to High(STAT_GROUPS) do
        if group=STAT_GROUPS[i] then Break;

     case i of
     0: currStat:=WABD_STATGRP_RESPONSE;
     1: currStat:=WABD_STATGRP_SENDSIZE;
     2: currStat:=WABD_STATGRP_RECVSIZE;
     3: currStat:=WABD_STATGRP_NETRESPONSE;
     end;


     response.ContentType:='application/json';

     BuildStats(currStat, statData, statTotals);
     if statData[0].c0='' then
     begin
        // empty array
        jsonRecords:=TSuperObject.ParseString('[]',true);
       response.Response.Text:=jsonGridResp(0, jsonRecords);
       Exit;
     end;
     ctx := TSuperRttiContext.Create;
     try
        jsonRecords:=ctx.AsJson<TStatData>(statData);
        jsonTotals:=ctx.AsJson<TStatTotals>(statTotals);
     finally
       ctx.Free;
     end;

     response.Response.Text:=jsonGridResp(Length(statData), jsonRecords, jsonTotals);
end;

procedure TWABDAdmin.SaveStatsToFile();
begin
     with WABDAdminSession do
          if SessionMgr<>nil then SessionMgr.Stats.Save(SessionMgr.ModuleName);
end;

procedure TWABDAdmin.LoadStatsFromFile();
begin
     with WABDAdminSession do
          if SessionMgr<>nil then SessionMgr.Stats.Load(SessionMgr.ModuleName);
end;

procedure TWABDAdmin.ClearStat();
var
   grp:TWABD_SesStatGroup;
   lst:TList;
   i:integer;
begin
     with WABDAdminSession.SessionMgr do
     begin
          if Stats=nil then exit;

          i:=Stats.IndexOf(currStat);
          lst:=Stats.LockList;
          try
             grp:=TWABD_SesStatGroup(lst.Items[i]);
             if grp=nil then exit;
             grp.Zero;
          finally
             Stats.UnlockList;
          end;
     end;
end;

procedure TWABDAdmin.ClearAllStats();
begin
     with WABDAdminSession.SessionMgr do
          if Stats<>nil then Stats.Zero;
end;



procedure TWABDAdmin.StatActions(request: TWABD_CustomRequest; response: TWABD_CustomResponse);
var
  op: Integer;
begin
     if cmdExtract(request)<>'save' then Exit;



     response.ContentType:='application/json';

     try
       op:=jsonReq.I['recid'];
       case op of
       91:  SaveStatsToFile;
       92:  LoadStatsFromFile;
       93:  ClearStat;
       94:  ClearAllStats;
       end;
       response.Response.Text:=jsonFormResp(TSuperObject.ParseString('{}',true));
     except
       response.Response.Text:=jsonErrResp('Stat action error');
     end;

end;

procedure TWABDAdmin.TerminateDLL(request: TWABD_CustomRequest; response: TWABD_CustomResponse);
begin
//     KillAllSessions(request, response);
//     KillSession(WABDAdminSession.SessionID);
end;


end.
