unit WABD_Admin_Component;

interface
uses Classes, SysUtils, Windows, Forms, Graphics, Controls, StdCtrls, ExtCtrls,
      Inifiles, Registry,
      WABD_Common, WABD_Setup, WABD_Utils ;

type
   TWABD_Admin = class(TComponent)
   private
      function getUser(): string;
      procedure setUser(UserName: string);
      function getPassword(): string;
      procedure setPassword(pw: string);
   protected
      FAdminUser:string;
      FAdminPassword:string;
      FLogoutHTML:string;
      LogCS      : TRTLCriticalSection;
      FLogging:boolean;
      FLogFile:string;
      FStorage    : TWABD_Storage;
      FStoragePath: string;
      FSectionName: string;
      FAutoLoad  : boolean;
      FAutoSave  : boolean;


      FLogWhat   : TWABD_Logging;
      FOnLoad    : TWABD_OnStorage;
      FOnSave    : TWABD_OnStorage;

      procedure   Loaded; override;
      function    GetLogging:boolean;
      procedure   setLogFile(filename: string);
   public
      constructor Create(AOwner:TComponent); override;
      destructor Destroy; override;
      procedure LogFmt(fmt:string; args:array of const);
      procedure Log(Text:string);
      procedure Load;
      procedure Save;
      property  AdminUser:string read getUser write setUser;
      property  AdminPassword:string read getPassword write setPassword;
   published
      property LogoutHTML:string read FLogoutHTML write FLogoutHTML;
      property Logging:boolean read GetLogging write FLogging;
      property LogFile:string read FLogFile write setLogFile;
      property AutoLog:TWABD_Logging read FLogWhat write FLogWhat;
      property Storage   : TWABD_Storage read FStorage write FStorage;
      property StoragePath:string read FStoragePath write FStoragePath;
      property SectionName:string read FSectionName write FSectionName;
      property AutoLoad:boolean read FAutoLoad write FAutoLoad;
      property AutoSave:boolean read FAutoSave write FAutoSave;
      property OnLoad:TWABD_OnStorage read FOnLoad write FOnLoad;
      property OnSave:TWABD_OnStorage read FOnSave write FOnSave;
   end;

implementation
// ************************************************************************
// TWABD_Admin
// ************************************************************************

constructor TWABD_Admin.Create(AOwner: TComponent);
begin
   inherited;
   setUser('admin');
   setPassword('password');
   FLogoutHTML:='<CENTER><P><FONT SIZE=5><B>Bye. Have a nice day.</B></FONT></P></CENTER>';
   InitializeCriticalSection(LogCS);
end;

destructor TWABD_Admin.Destroy;
begin
    // Before destruction, save info.
    if FAutoSave then Save;
    DeleteCriticalSection(LogCS);
    inherited;
end;

function TWABD_Admin.getUser(): string;
begin
  Result:= FAdminUser
end;

procedure TWABD_Admin.setUser(UserName: string);
begin
    FAdminUser:=UserName;
end;

function TWABD_Admin.getPassword(): string;
begin
  Result:=FAdminPassword;
end;

procedure TWABD_Admin.setPassword(pw: string);
begin
  FAdminPassword:=pw;
end;


procedure TWABD_Admin.setLogFile(filename: string);
begin
  // change path only at run time
  if not (csDesigning in ComponentState) then
  begin
      // relative path
      if (filename[1]='\')
      or (filename[1]='.') then
      begin
       if (filename[2]='\') then
        filename:=Copy(filename,3)
       else
        filename:=Copy(filename, 2);
        filename:=WABD_Utils.ExtractModuleDir+filename;
      end;
  end;
  FLogFile:=filename;
end;

procedure TWABD_Admin.Save;
var
   reg:TRegistry;
   ini:TIniFile;
begin
    // Before destruction, save info.
    case Storage of
        storeNone: ;
        storeIniFile:
          begin
            ini:=TIniFile.Create(FStoragePath);
            if ini<>nil then
            begin
                ini.WriteString(FSectionName,'LogFile',LogFile);
                ini.WriteString(FSectionName,'AdminUser',AdminUser);
                ini.WriteString(FSectionName,'AdminPassword',AdminPassword);
                if assigned(FOnSave) then FOnSave(self,FSectionName,'',ini);
            end;
            ini.free;
          end;
    end;
end;

procedure TWABD_Admin.Load;
var
   reg:TRegistry;
   ini:TIniFile;
begin
    case Storage of
        storeNone: ;
        storeIniFile:
          begin
            ini:=TIniFile.Create(FStoragePath);
            if ini<>nil then
            begin
                LogFile:=ini.ReadString(FSectionName,'LogFile','');
                AdminUser:=ini.ReadString(FSectionName,'AdminUser','admin');
                AdminPassword:=ini.ReadString(FSectionName,'AdminPassword','password');
                if assigned(FOnLoad) then FOnLoad(self,FSectionName,'',ini);
            end;
            ini.free;
          end;
    end;
end;

procedure TWABD_Admin.Loaded;
begin
    inherited;
    if FAutoLoad then Load;
end;

function TWABD_Admin.GetLogging:boolean;
begin
    Result:=FLogging and (LogFile<>'');
end;

procedure TWABD_Admin.LogFmt(fmt:string; args:array of const);
var
   f: TextFile;
begin
   if not Logging then exit;

   EnterCriticalSection(LogCS);
   try
      assignfile(f, fLogFile);

      if FileExists(FLogFile) then
          Append(f)
      else
          Rewrite(f);
      writeln(f,datetimetostr(Now)+' '+format(fmt,args));
      CloseFile(f);
   finally
      LeaveCriticalSection(LogCS);
   end;
end;

procedure TWABD_Admin.Log(Text:string);
begin
     LogFmt('%s',[Text]);
end;

end.
