unit WABD_Attributes;

interface
uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants,
  System.Classes,
  System.Rtti, System.TypInfo,
  WABD_Request, WABD_Response;

type
TUrlType = (utStartSession, utInternal);

UrlAttribute = class(TCustomAttribute)
  private
    FUrlDescr:  string;
    FHttpMethod: string;
    FUrlType : TUrlType;
  public
    constructor Create(const UrlDescr: string; const method: string; urlType: TUrlType = utStartSession);
    property url: string read FUrlDescr write FUrlDescr;
    property method: string read FHttpMethod write FHttpMethod;
    property urlType: TUrlType read FUrlType;
  end;


  function GetUrlAttribute(const aMethod: TRttiMethod): UrlAttribute;
  function RunHandlerForUrl(const aClass: TObject; const requestUrl: string;
        const httpMethod: string;
        request: TWABD_CustomRequest; response: TWABD_CustomResponse): integer;

  function ValidateUrlForNewSession(const aClass: TObject; const requestUrl: string): Boolean;

implementation

constructor UrlAttribute.Create(const UrlDescr: string; const method: string;
                urlType: TUrlType);
begin
  FUrlDescr:=UrlDescr;
  FHttpMethod:=UpperCase(method);
  FUrlType:=urlType;
end;


// to override bug in Delphi VCL (XE2, 10.2...):
// the direct call to TRTTIHelpers.getAttribute:
//      ma:=TRTTIHelpers.getAttribute<UrlAttribute>(cls);
// and
//      mp:=TRTTIHelpers.GetAttribute<UrlPropertyAttribute>(propertyInfo);
// doesn't work, returning empty string fields of attribute.
// Same bug if this function GetUrlAttribute return straight the 'attribute' object
function GetUrlAttribute(const aMethod: TRttiMethod): UrlAttribute;
var
  C: TRttiContext;
  rttiType: TRttiType;
  attribute: TCustomAttribute;
  modattr: UrlAttribute;
begin
  rttiType := C.GetType(aMethod);
  for attribute in rttiType.GetAttributes do
    if (attribute is UrlAttribute) then
    begin
      with UrlAttribute(attribute) do
          modattr:=UrlAttribute.Create(url,method, urlType );
      Exit(modattr);
    end;

  Result := nil;
end;

function compareUrl(const requestUrl: string; const url: string): Boolean;
begin
  Result:=False;
  if url=requestUrl then
  begin
    Result:=True;
    Exit;
  end;
  // url with a jolly char '*' as last char

  if  (url<>'') and (url[Length(url)]='*') and
     (Pos(copy(url,1,Length(url)-1), requestUrl) = 1) then
    Result:=True;
end;

function RunHandlerForUrl(const aClass: TObject;  const requestUrl: string;
        const httpMethod: string;
        request: TWABD_CustomRequest; response: TWABD_CustomResponse): integer;
var
  C: TRttiContext;
  rttiType: TRttiType;
  LMet: TRttiMethod;
  attribute: TCustomAttribute;
  ti: Pointer;
begin
  Result:=0;
  rttiType := C.GetType(aClass.ClassType);
  for LMet in rttiType.GetDeclaredMethods do
   begin
      for attribute in LMet.GetAttributes do
        if (attribute is UrlAttribute) then
        begin
            with UrlAttribute(attribute) do
            begin
                if (compareUrl(requestUrl, url)=True)
                  and ( (method='*') or (Pos(httpMethod, method)>0) ) then
                begin
                      LMet.Invoke(aClass, [TValue.From(request), TValue.From(response)]);
                      Result:=Result+1;
                end;
            end;
        end;
    end;
end;

function ValidateUrlForNewSession(const aClass: TObject; const requestUrl: string): Boolean;
var
  C: TRttiContext;
  rttiType: TRttiType;
  LMet: TRttiMethod;
  attribute: TCustomAttribute;
  ti: Pointer;
begin
  Result:=False;
  C:=TRttiContext.Create;
  rttiType := C.GetType(aClass.ClassType);
  for LMet in rttiType.GetDeclaredMethods do
   begin
      for attribute in LMet.GetAttributes do
        if (attribute is UrlAttribute) then
        begin
            with UrlAttribute(attribute) do
            begin
            if (url=requestUrl)
              and ( urlType =utStartSession ) then
                Result:=True;
                Break;
            end;
        end;
    end;
   c.Free;
end;

initialization
begin
// needed by Delphi 10.2 with VCL (not needed with FMX, why?)
{$IF CompilerVersion >= 25.0}
  TRttiContext.KeepContext;
{$IFEND}
end;


finalization
begin
{$IF CompilerVersion >= 25.0}
  TRttiContext.DropContext;
{$IFEND}
end;

end.
