unit WABD_HTMLEditor;

{$I kbmWABD.inc}

interface

uses
  DesignIntf,
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  WABD_Objects, WABD_Common,
  StdCtrls, ExtCtrls, Menus, Buttons,
  ShellAPI, ComCtrls, SynEditOptionsDialog, SynEdit, SynHighlighterJScript,
  SynHighlighterHtml, SynHighlighterCSS, SynEditHighlighter, SynHighlighterMulti,
  SynEditCodeFolding
//  {$IFDEF SYN_CodeFolding}
//  ,SynEditCodeFolding
//  {$ENDIF}
  ;

type
   tEdSyntax = (edNone, edHTML, edJavascript, edCSS);

   TWABDHTMLEditor = class(TForm)
    Panel1: TPanel;
    btnOK: TMainMenu;
    Preview1: TMenuItem;
    HTML1: TMenuItem;
    FSPanel: TPanel;
    Panel3: TPanel;
    odLoad: TOpenDialog;
    sdSave: TSaveDialog;
    ReplaceDialog1: TReplaceDialog;
    dlgFind: TFindDialog;
    Panel4: TPanel;
    Label1: TLabel;
    btnCopy: TButton;
    btnCut: TButton;
    btnPaste: TButton;
    Panel5: TPanel;
    Button1: TButton;
    btnEditHTML: TButton;
    btnLoadHTML: TButton;
    btnSaveHTML: TButton;
    mHTML: TSynEdit;
    SynEditOptionsDialog1: TSynEditOptionsDialog;
    SynMultiSyn1: TSynMultiSyn;
    pmEdit: TPopupMenu;
    mnOptions: TMenuItem;
    mnUndo: TMenuItem;
    btUndo: TButton;
    mnCopy: TMenuItem;
    mnPaste: TMenuItem;
    mnCut: TMenuItem;
    SynHTMLSyn1: TSynHTMLSyn;
    SynJScriptSyn1: TSynJScriptSyn;
    SynCssSyn1: TSynCssSyn;
    PageControl1: TPageControl;
    tsStdTemplates: TTabSheet;
    Label7: TLabel;
    btnInsertStdTemplate: TButton;
    lbStdTemplates: TListBox;
    mnSyntax: TMenuItem;
    mnHTML: TMenuItem;
    mnJscript: TMenuItem;
    mnCss: TMenuItem;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure FormDestroy(Sender: TObject);
    procedure HTML1Click(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure btnLoadHTMLClick(Sender: TObject);
    procedure btnSaveHTMLClick(Sender: TObject);
    procedure btnEditHTMLClick(Sender: TObject);
    procedure btnInsertStdTemplateClick(Sender: TObject);
    procedure btnCutClick(Sender: TObject);
    procedure btnCopyClick(Sender: TObject);
    procedure btnPasteClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure mnOptionsClick(Sender: TObject);
    procedure mnUndoClick(Sender: TObject);
    procedure btUndoClick(Sender: TObject);
    procedure mnCopyClick(Sender: TObject);
    procedure mnPasteClick(Sender: TObject);
    procedure mnCutClick(Sender: TObject);
    procedure mnHTMLClick(Sender: TObject);
    procedure mnJscriptClick(Sender: TObject);
    procedure mnCssClick(Sender: TObject);
   private
      FWABDHTML   : TWABD_Base_HTML;
      FSyntax     : tEdSyntax;
      procedure setSyntax();

      procedure   SetWABDHTML(NewHTML: TWABD_Base_HTML);
   public
      MyDesigner   : IDesigner;

      property    WABDHTML: TWABD_Base_HTML read FWABDHTML write SetWABDHTML;
      procedure   ReflectedNotify(Sender: TObject; AComponent: TComponent; Operation: TOperation);
      procedure   Initialize;
      procedure   Preview(What:integer);
      procedure   InsertText(AString:string);
   end;

var
  WABDHTMLEditor: TWABDHTMLEditor;

implementation

uses ToolsAPI;

{$R *.DFM}

function MaxInt(a, b: integer): integer;
begin
   if a < b then Result := a
      else Result := b;
end;

procedure TWABDHTMLEditor.FormCreate(Sender: TObject);
begin
  SynMultiSyn1.DefaultFilter := SynHTMLSyn1.DefaultFilter;
  FSyntax:=edHTML;
  setSyntax;
end;

procedure TWABDHTMLEditor.FormDestroy(Sender: TObject);
begin
     if WABDHTML<>nil then WABDHTML.RefNotify := nil;
end;

procedure TWABDHTMLEditor.SetWABDHTML(NewHTML: TWABD_Base_HTML);
begin
   FWABDHTML := NewHTML;
end;

procedure TWABDHTMLEditor.mnUndoClick(Sender: TObject);
begin
    btUndo.Click;
end;


procedure TWABDHTMLEditor.Initialize;
var
  leftdel, rightdel: string;
begin
   Caption := 'WABD Template Designer - ' + WABDHTML.Name;
   Assert(MyDesigner<>nil, 'Designer = nil');

   WABDHTML.RefNotify := ReflectedNotify;

   mHTML.Text:=WABDHTML.HTML.Text;

   leftdel:= wabd_variabletypestrings[WABDHTML.VariableType, 0];
   rightdel:= wabd_variabletypestrings[WABDHTML.VariableType, 1];
   lbStdTemplates.Clear;
   lbStdTemplates.Items.Add(leftdel+WABD_VAR_SESSIONID+rightdel);
   lbStdTemplates.Items.Add(leftdel+WABD_VAR_USERNAME+rightdel);
   lbStdTemplates.Items.Add(leftdel+WABD_VAR_INFO+rightdel);
   lbStdTemplates.Items.Add(leftdel+WABD_VAR_HITCOUNT+rightdel);
   lbStdTemplates.Items.Add(leftdel+WABD_VAR_VERSION+rightdel);
   lbStdTemplates.Items.Add(leftdel+WABD_VAR_SITENAME+rightdel);
   lbStdTemplates.Items.Add(leftdel+WABD_VAR_SITEID+rightdel);
   lbStdTemplates.Items.Add(leftdel+WABD_VAR_WEBADMIN+rightdel);
   lbStdTemplates.Items.Add(leftdel+WABD_VAR_MODULENAME+rightdel);
   lbStdTemplates.Items.Add(leftdel+WABD_VAR_BASEURL+rightdel);
end;


// I had to do this hack because Delphi 3 won't let me compile with
// the DsgnWnds unit.
procedure TWABDHTMLEditor.ReflectedNotify(Sender: TObject; AComponent: TComponent; Operation: TOperation);
var
   i : integer;
begin
   if (Operation <> opRemove) then exit;

   if AComponent=WABDHTML then begin
      WABDHTML.RefNotify := nil;
      WABDHTML := nil;
      Visible  := False;
   end;

end;

procedure TWABDHTMLEditor.InsertText(AString:string);
var
   p:TPoint;
   n:integer;
   s:string;
begin
     p := point( mHTML.CaretX-1, mHTML.CaretY-1);
     n:=mHTML.SelStart;
     s:=mHTML.Lines[p.y];
     mHTML.Lines[p.y]:=copy(s,1,p.x)+AString+Copy(s,p.x+1,length(s));
     mHTML.SetFocus;
     mHTML.SelStart:=n;
end;




procedure TWABDHTMLEditor.mnOptionsClick(Sender: TObject);
var
  synoptcont: TSynEditorOptionsContainer;
begin
  synoptcont := TSynEditorOptionsContainer.Create(mHTML);
  synoptcont.Assign(mHTML);
  SynEditOptionsDialog1.Execute(synoptcont);
  synoptcont.AssignTo(mHTML);
  synoptcont.Free;
end;


procedure TWABDHTMLEditor.mnPasteClick(Sender: TObject);
begin
      btnPaste.Click;
end;

procedure TWABDHTMLEditor.Preview(What:integer);
var
   f    : string;
   h    : ansistring;
   fstm : TFileStream;
   rc   : integer;

   a    : IOTAProject;
   pjName: string;
begin

try
   // get the project directory and set it as active
   a := ToolsAPI.GetActiveProject;
   pjName := a.FileName;
   ChDir(ExtractFilePath(pjName));

   case What of
     0:
        begin
             h :=  mHTML.Text;
             f := 'PreviewTemp.htm';
        end;
   end;
   fstm := TFileStream.Create(f, fmCreate or fmOpenWrite or fmShareDenyWrite);
   fstm.Write(h[1], Length(h));
   fstm.Free;

   rc := ShellExecute(Handle, 'open', PChar(f), nil, nil, SW_SHOWNORMAL);
   if (rc <= 32) then ShowMessage('Unable to execute: ' + f + #13#10 + 'Browser not installed?');
except
  on E:Exception do
     MessageDlg('HTML template preview error - '+e.message, mtError, [mbOK], 0);
end;
end;

procedure TWABDHTMLEditor.Notification(AComponent: TComponent; Operation: TOperation);
begin
   if (Operation = opRemove) and (AComponent = WABDHTML) then begin
      WABDHTML := nil;
      Hide;
   end;
end;



procedure TWABDHTMLEditor.mnCopyClick(Sender: TObject);
begin
    btnCopy.Click;
end;


procedure TWABDHTMLEditor.HTML1Click(Sender: TObject);
begin
     Preview(0);
end;


procedure TWABDHTMLEditor.mnCutClick(Sender: TObject);
begin
      btnCut.Click;
end;

procedure TWABDHTMLEditor.setSyntax();
begin
  case FSyntax of
    edHTML: begin odLoad.Filter := SynHTMLSyn1.DefaultFilter;
                  sdSave.Filter := SynHTMLSyn1.DefaultFilter;
                  mHTML.Highlighter := SynMultiSyn1;
                  mnHTML.Checked := true;
            end;
    edJavascript:
            begin odLoad.Filter := SynJScriptSyn1.DefaultFilter;
                  sdSave.Filter := SynJScriptSyn1.DefaultFilter;
                  mHTML.Highlighter := SynJScriptSyn1;
                  mnJscript.Checked := true;
            end;
    edCSS: begin
                  odLoad.Filter := SynCssSyn1.DefaultFilter;
                  sdSave.Filter := SynCssSyn1.DefaultFilter;
                  mHTML.Highlighter := SynCssSyn1;
                  mnCss.Checked := true;
            end;

  end;
end;

procedure TWABDHTMLEditor.mnHTMLClick(Sender: TObject);
begin
    FSyntax := edHTML;
    setSyntax;
end;

procedure TWABDHTMLEditor.mnJscriptClick(Sender: TObject);
begin
    FSyntax := edJavascript;
    setSyntax;
end;

procedure TWABDHTMLEditor.mnCssClick(Sender: TObject);
begin
    FSyntax := edCSS;
    setSyntax;
end;

procedure TWABDHTMLEditor.Button1Click(Sender: TObject);
begin
     WABDHTML.HTML.Text:=mHTML.Text;
     MyDesigner.Modified;
     Close;
end;


procedure TWABDHTMLEditor.btnLoadHTMLClick(Sender: TObject);
begin
     if odLoad.Execute then
        mHTML.Lines.LoadFromFile(odLoad.Filename);

end;

procedure TWABDHTMLEditor.btnSaveHTMLClick(Sender: TObject);
begin
     if sdSave.Execute then
        mHTML.Lines.SaveToFile(sdSave.Filename);
end;

procedure TWABDHTMLEditor.btUndoClick(Sender: TObject);
begin
    mHTML.Undo;
end;

procedure TWABDHTMLEditor.btnEditHTMLClick(Sender: TObject);
var
   fn:string;
begin
     fn:=mHTML.Name+'.html';
     mHTML.Lines.SaveToFile(fn);
     ShellExecute(handle, 'edit', PChar(fn), nil, nil, SW_SHOWNORMAL);
end;


procedure TWABDHTMLEditor.btnInsertStdTemplateClick(Sender: TObject);
var
   ni:integer;
begin
     ni := lbStdTemplates.ItemIndex;
     if ni=-1 then exit;
     InsertText(lbStdTemplates.Items[ni]);
end;

procedure TWABDHTMLEditor.btnCutClick(Sender: TObject);
begin
     mHTML.CutToClipboard;
end;


procedure TWABDHTMLEditor.btnCopyClick(Sender: TObject);
begin
     mHTML.CopyToClipboard;

end;

procedure TWABDHTMLEditor.btnPasteClick(Sender: TObject);
begin
     mHTML.PasteFromClipboard;
end;

end.
