unit WABD_HTMLResponse;

interface

uses SysUtils,
  WABD_Response,WABD_Cookies,WABD_Common;

type
   TWABD_CustomHTMLResponse = class(TWABD_CustomResponse)
   protected
      function FormatResponse:string; override;
   end;

implementation
uses StrUtils;

function TWABD_CustomHTMLResponse.FormatResponse:string;
var
   i:integer;
   fn:string;
   cookiestr:string;
   cookie:TWABD_Cookie;
   s:string;
const
   CR = #13#10;

   procedure stripCR(var s: string);
   begin
      // delete the last carriage return/line feed from response string
      // They are automatically added by TStringlist.text method
         if (Length(s)>0) then
         begin
           if (s[Length(s)-1] = #13) and
              (s[Length(s)] = #10) then
              s := LeftStr(s, Length(s)-2);
         end;
   end;
begin
     // Check if to redirect to another location.
     if (Status=WABD_STATUS_OK) and (Location<>'') then Status:=WABD_STATUS_REDIRECT;
     if (Status=WABD_STATUS_REDIRECT) and (Location='') then Status:=WABD_STATUS_OK;

     // Build cookie list.
     cookiestr:='';
     if Assigned(Cookies) then
        for i:=0 to Cookies.count-1 do
        begin
             cookie:=TWABD_Cookie(Cookies.Items[i]);
             if Length(Trim(cookie.Name))>0 then
                    cookiestr:=cookiestr+'Set-Cookie: '+cookie.HeaderValue+';'+CR;
        end;

     // Format HTTP.
     case Status of
        WABD_STATUS_OK:
                 begin
                    if ContentDesc<>'' then
                        fn:='Content-Disposition: attachment;filename='+ContentDesc+CR
                    else
                        fn:='';
                    if Length(BinaryResponse)>0 then
                       s:=BinaryResponse
                    else
                    begin
                       s:=Response.Text;
                       stripCR(s);
                    end;
                    Result:='HTTP/1.0 200 OK'+CR+
                            cookiestr+
                            fn+
                            'Content-Type: '+ContentType+CR+
                            'Content-Length: '+inttostr(length(s))+CR+
                            Header.Text+
                            'Content:'+CR+CR+s;
                 end;
        WABD_STATUS_REDIRECT:
                 begin
                    if ContentDesc<>'' then
                        fn:='Content-Disposition: attachment;filename="'+ContentDesc+'"'+CR
                    else
                        fn:='';
                    Result:='HTTP/1.0 302 Moved temporarily'+CR+
                            cookiestr+
                            'Location: '+Location+CR+
                            Header.Text+
                            CR+CR;
                 end;
        WABD_STATUS_AUTH:
                 begin
                    Result:='HTTP/1.0 401 Unauthorized'+CR+
                            'WWW-authenticate: Basic realm="'+ContentDesc+'"'+CR+
                            cookiestr+
                            'Content-type: text/plain'+CR+
                            'Content-Length: 14'+CR+
                            Header.Text+
                            'Content:'+CR+CR+'Access denied.';
                 end;
        WABD_STATUS_ERROR:
                 begin
                    s:=Response.Text;
                    stripCR(s);
                    Result:='HTTP/1.0 403 Forbidden'+CR+
                            cookiestr+
                            'Content-type: text/plain'+CR+
                            'Content-Length: '+inttostr(length(s))+CR+
                            Header.Text+
                            'Content:'+CR+CR+s;
                 end;
     end;
end;

end.
