unit WABD_Objects;

{$include kbmWABD.inc}

interface
uses Classes, SysUtils, Windows, Forms, Graphics, Controls, StdCtrls, ExtCtrls,
      Inifiles, Registry,
      System.Generics.Collections,

      WABD_About,   WABD_Cookies,  WABD_Common,
      WABD_Request, WABD_Response,
      WABD_Utils,
      WABD_ISAPI,
      WABD_Attributes, WABD_Statistics, WABD_Setup, WABD_Admin_Component,
      WABD_HTMLPreprocess ;

// =========================================================================
// kbmWABD
// -------
// Web Application Builder for Delphi
//
// Based on the open source and freeware version 1.25 of Ben Zieglers library
// WABD, and Kim Bo Madsen/Optical Services - Scandinavia,
//
// Copyright 1997,1998 for v. 1.25 and all previous versions Ben Ziegler (Zieglersoft)
// Copyright 1999-2003 for v. 2.xx and newer Kim Bo Madsen/Components4Developers
// Copyright 2014-2020 for v. 4.xx and newer Giandomenico De Sanctis
// All rights reserved.
//
// The copyright includes all files in this library, not only this file.
//
// LICENSE AGREEMENT
//
// You are allowed to use this component in any project for free.
// You are NOT allowed to claim that you have created this component or to
// copy its code into your own component and claim that it was your idea.
//
// Since this library is open source and freeware, all later incarnations _must_ also
// be open source and freeware.
//
// All the added functionality to the original 1.25 library is offered for free
// for your convinience, and the only thing I REQUIRE is to get an e-mail about what
// project this library (or derived versions) are used for.
//
// You dont need to state the name of the creators (Ben Ziegler and Kim Bo Madsen) in
// your software, although it would be appreciated if you do.
//
// If you find bugs or alter the component (f.ex. see suggested enhancements
// further down), please DONT just send the corrected/new code out on the internet,
// but instead send it to me, so I can put it into the official version. You will
// be acredited if you do so.
//
// Ben Ziegler has to my knowledge not claimed any restrictions which
// should violate this license agreement.
//
// DISCLAIMER
// By using this component or parts theirof you are accepting the full
// responsibility of the use. You are understanding that the authors (Ben Ziegler or
// Kim Bo Madsen) cant be made responsible in any way for any problems occuring
// using this component.
// You also recognize the authors (Ben Ziegler and Kim Bo Madsen) as the creators
// of this component and agrees not to claim otherwize!
//
//
// History:
// 4/2020
// 4.00  Totally new version


type
   // ************************************************************************
   // "Base" Level objects

   TWABD_Session = class;
   TWABD_Object = class;
//-------
   TWABD_Body = class;
   TWABD_Base_HTML = class;
   TWABD_HTML = class;
   TWABD_SessionMgr = class;

   TGarbageThread = class(TThread)
   protected
      procedure   DoGarbageCollection;
      procedure   Execute; override;
   public
      SesMgr      : TWABD_SessionMgr;
   end;

   TWABD_SequenceRec = record
       Persistent:boolean;
       Name:PAnsiChar;
       Value:longint;
   end;
   PWABD_SequenceRec = ^TWABD_SequenceRec;

// Compile for Delphi > XE2
{$IF CompilerVersion>23}
   TSessionList      = TThreadList<TWABD_Session>;
{$else}
   TSessionList = class(TThreadList)
   private
      tempList: TList<TWabd_session>;
      lockLs: TList;
   public
      function LockList: TList<TWABD_Session>; overload;
      procedure UnLockList();
      procedure DeleteLock(ind: Integer);
   end;
{$ifend}

   TTempSessionList  = TList<TWABD_Session>;

   TWABD_OnCreateGuestSession = procedure(var NewSession: TWABD_Session; TriedToBeAdmin:boolean;
          Request:TWABD_CustomRequest; validated: boolean) of object;
   TWABD_OnCreateSession = procedure(var NewSession: TWABD_Session; Request:TWABD_CustomRequest;
            var validated: boolean) of object;
   TWABD_OnDestroySession = procedure(Session: TWABD_Session) of object;

   TWABD_OnFirstSession = procedure(Session: TWABD_Session) of object;


   TWABD_RouteWhen = (rwNever, rwWhenFull, rwAllways);
   TWABD_RouteHow = (rhRandom, rhRoundRobin);
   TWABD_OnAuthenticate = procedure(RemoteHost:string; UserName:string; Password:string; var Authenticated:boolean) of object;
   TWABD_OnTerminate = function(Flags:integer):boolean of object;
   TWABD_OnValidateRequest = procedure(Request:TWABD_CustomRequest; Response:TWABD_CustomResponse; var Accept:boolean) of object;
   TWABD_OnGarbageCollection = procedure(SesMgr:TWABD_SessionMgr) of object;
   TWABD_OnException = procedure(Sender:TObject; E: Exception; var Handled: boolean) of object;
   TWABD_OnInvalidSession = procedure(Sender:TObject; var CreateNew:boolean) of object;
   TWABD_OnSystemXHttpResponse = procedure(Request:TWABD_CustomRequest; sysResponse: TWABD_SysResponse;
                     var systemResp: string) of object;

   TWABD_SessionMgr = class(TComponent)
   protected
      FAbout      : TWABDAbout;
      FCreate     : TWABD_OnCreateSession;
      FDestroy    : TWABD_OnDestroySession;
      FCreateAdmin: TWABD_OnCreateSession;
      FCreateGuest: TWABD_OnCreateGuestSession;
      FOnFirstSes : TWABD_OnFirstSession;
      FOnValidateRequest : TWABD_OnValidateRequest;
      FOnInvalidSession: TWABD_OnInvalidSession;
      FOnException: TWABD_OnException;
      FOnAuthenticate  : TWABD_OnAuthenticate;
      FOnTerminate: TWABD_OnTerminate;
      FOnSystemXHttpResponse: TWABD_OnSystemXHttpResponse;
      FOnGarbageCollection:TWABD_OnGarbageCollection;
      FOnLoad    : TWABD_OnStorage;
      FOnSave    : TWABD_OnStorage;

      FCheck      : integer;        // Interval to perform Garbage Collection (seconds)
      FGarbage    : boolean;        // True = perform Garbage collection
      FWebAdmin   : string;         // Specify string (email, phonenumber or name) for application administrator.
      FSiteName   : string;
      FRouteSites : TStringList;    // Alternative sites to route call to.
      FRouteWhen  : TWABD_RouteWhen;
      FRouteHow   : TWABD_RouteHow;
      FRouteLast  : integer;        // Points into FRouteSites for last one routed to.
      FSiteID     : integer;        // Unique ID of this site. (-1= allocates randomly)
      FGatherStats: boolean;
      FNetWorkStats: boolean;
      FStats      : TWABD_SesStat;
      FAdmin      : TWABD_Admin;
      FDefSesTimeout: integer;
      FMaxSessions: integer;
      FRandomSessionID : boolean;
      FMaxIdenticalUser: integer;
      FAutoLogoutOldIdenticalUser : boolean;
      FStorage    : TWABD_Storage;
      FStoragePath: string;
      FSectionName: string;
      FAutoLoad  : boolean;
      FAutoSave  : boolean;
      FHTMLTimeOut: TStringList;
      FInfo      : string;
      FVerDummy  : string;
      FMaxRequestSize:integer;
      FTotalSessionCount: longint;        // Number of sessions created totally until now.
      FModuleName: string;
      FModulePath: string;
      FEmptyAsRoot: Boolean;
      FBaseUrl    : string;
      FAuthType   : TWABD_AuthType;
      FUniqueList: TThreadList;     // List containing unique counters.

      GarbageThrd : TGarbageThread;
      StopEvent   : THandle;
      SesMgrCSCreate  : TRTLCriticalSection;
      SesMgrCSDestroy  : TRTLCriticalSection;
      SesMgrCSAuth : TRTLCriticalSection;
      CreateTime  : TDateTime;
      FVariables  : TStrings;
      FVariableType : TWABD_VariableType;
      FCrypted    : boolean;

      FRequest    : TWABD_CustomRequest;
      FResponse   : TWABD_CustomResponse;

      function    CreateNewSession(Request:TWABD_CustomRequest; Response:TWABD_CustomResponse):TWABD_Session;
      function    RunExistingSession(ASession:TWABD_Session; const IdStr,BodyName:string;
                                     Request:TWABD_CustomRequest; Response:TWABD_CustomResponse):TWABD_Session;
      procedure   DoDestroySession(Ses: TWABD_Session);
      function    Authenticate(Request:TWABD_CustomRequest; Response:TWABD_CustomResponse;
                        LikeToBeAdmin:boolean; var IsAdmin:boolean; var basicAuth: boolean):boolean;
      procedure   SetGatherStats(b:boolean);
      function    GetSessionCount:integer;
      procedure   Notification(AComponent: TComponent; Operation: TOperation); override;
      procedure   Loaded; override;
      procedure   SetSiteID(id:integer);
      function    GetRoute:string;
      function    CountIdenticalUser(UserName:string):integer;
      procedure   RemoveIdenticalUsers(UserName:string; MaxAllowedLeft:integer);

      procedure   SetVariables(NewVariables: TStrings);
      procedure   SetVariableByName(AName:string; AValue:string);
      function    GetVariableByName(AName:string):string;
      function    GetVersion:string;
      procedure   SetAdmin(AAdmin:TWABD_Admin);
      procedure   SetCrypted(flg_crypt: Boolean);
      procedure   CodeHiddenNames();
      procedure   SendSystemResponse(typ: TWABD_SysResponse; requestType: Integer;
            var respText: string; var respStatus: integer);
      function    AuthBasicDecode(const crypted: string; var user: string;
            var pw: string): Boolean;
   public
      SessionList : TSessionList;

      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;


      function    GetFullModuleName:string;
      function    LocateSessionByID(ASiteID:integer; ASessionID:longint):TWABD_Session;
      procedure   CheckLogOff(Ses: TWABD_Session);
      procedure   ClientRequest(Request:TWABD_CustomRequest; Response:TWABD_CustomResponse);
      function    OnTerminateCallback(Flags:longint):boolean;
      function    ProcessVariables(HTML:string):string;
      procedure   Load;
      procedure   Save;
      procedure   CreateSequence(ID:string; StartValue:longint; Persistent:boolean);
      function    DrawSequenceValue(ID:string):longint;
      procedure   DeleteSequence(ID:string);
      procedure   DeleteAllSequences;
      function    Encrypt(const sval: string): string;
      function    Decrypt(const sval: string): string;

      property    Stats:TWABD_SesStat read FStats;
      property    SessionCount:integer read GetSessionCount;
      property    TotalSessionCount:longint read FTotalSessionCount;
      property    VariableByName[s:string]:string read GetVariableByName write SetVariableByName;
      property    Request:TWABD_CustomRequest read FRequest;
      property    Response:TWABD_CustomResponse read FResponse;
      property    ModuleName:string read FModuleName;
      property    ModulePath:string read FModulePath;
   published
      property    GatherStatistics:boolean read FGatherStats write SetGatherStats;
      property    NetworkStatistics:boolean read FNetworkStats write FNetworkStats;
      property    MaxRequestSize:integer read FMaxRequestSize write FMaxRequestSize;
      property    CheckTimeoutInterval: integer read FCheck write FCheck;
      property    GarbageCollection: boolean read FGarbage write FGarbage;
      property    About: TWABDAbout read FAbout write FAbout;
      property    Variables: TStrings read FVariables write SetVariables;
      property    WebAdministrator: string read FWebAdmin write FWebAdmin;
      property    SiteName: string read FSiteName write FSiteName;
      property    SiteID:integer read FSiteID write SetSiteID;
      property    Admin:TWABD_Admin read FAdmin write SetAdmin;
      property    AuthType: TWABD_AuthType read FAuthType write FAuthType;
      property    DefaultSessionTimeout:integer read FDefSesTimeout write FDefSesTimeout;
      property    MaxConcurrentSessions:integer read FMaxSessions write FMaxSessions;
      property    MaxIdenticalUser:integer read FMaxIdenticalUser write FMaxIdenticalUser;
      property    AutoLogoutOldIdenticalUser:boolean read FAutoLogoutOldIdenticalUser write FAutoLogoutOldIdenticalUser;
      property    RandomSessionID:boolean read FRandomSessionID write FRandomSessionID;
      property    RouteSites:TStringList read FRouteSites write FRouteSites;
      property    RouteHow:TWABD_RouteHow read FRouteHow write FRouteHow;
      property    RouteWhen:TWABD_RouteWhen read FRouteWhen write FRouteWhen;
      property    Storage   : TWABD_Storage read FStorage write FStorage;
      property    StoragePath:string read FStoragePath write FStoragePath;
      property    SectionName:string read FSectionName write FSectionName;
      property    AutoLoad:boolean read FAutoLoad write FAutoLoad;
      property    AutoSave:boolean read FAutoSave write FAutoSave;
      property    HTMLTimeOut: TStringList read FHTMLTimeOut write FHTMLTimeOut;
      property    Info:string read FInfo write FInfo;
      property    Version:string read GetVersion write FVerDummy;
      property    VariableType:TWABD_VariableType read FVariableType write FVariableType;
      property    Encrypted: boolean read FCrypted write SetCrypted;
      property    EmptyUrlAsRoot: Boolean read FEmptyAsRoot write FEmptyAsRoot;
      property    BaseUrl: string read FBaseUrl write FBaseUrl;

      property    OnTerminate: TWABD_OnTerminate read FOnTerminate write FOnTerminate;
      property    OnFirstSession: TWABD_OnFirstSession read FOnFirstSes write FOnFirstSes;
      property    OnCreateSession: TWABD_OnCreateSession read FCreate write FCreate;
      property    OnCreateAdminSession: TWABD_OnCreateSession read FCreateAdmin write FCreateAdmin;
      property    OnCreateGuestSession: TWABD_OnCreateGuestSession read FCreateGuest write FCreateGuest;
      property    OnDestroySession: TWABD_OnDestroySession read FDestroy write FDestroy;
      property    OnAuthenticate: TWABD_OnAuthenticate read FOnAuthenticate write FOnAuthenticate;
      property    OnValidateRequest: TWABD_OnValidateRequest read FOnValidateRequest write FOnValidateRequest;
      property    OnInvalidSession: TWABD_OnInvalidSession read FOnInvalidSession write FOnInvalidSession;
      property    OnException: TWABD_OnException read FOnException write FOnException;
      property    OnSystemXHttpResponse: TWABD_OnSystemXHttpResponse read FOnSystemXHttpResponse write FOnSystemXHttpResponse;
      property    OnLoad:TWABD_OnStorage read FOnLoad write FOnLoad;
      property    OnSave:TWABD_OnStorage read FOnSave write FOnSave;
      property    OnGarbageCollection:TWABD_OnGarbageCollection read FOnGarbageCollection write FOnGarbageCollection;

   end;

   TWABD_OnTimeOut = procedure(SesMgr:TWABD_SessionMgr; ElapsedTime:longint; var AcceptTimeout:boolean) of object;
   TWABD_OnLogon = procedure(RemoteHost:string) of object;
   TWABD_OnLogoff = procedure of object;
   TWABD_BeforeProcessRequest = procedure(Session:TWABD_Session) of object;
   TWABD_AfterProcessRequest = procedure(Session:TWABD_Session) of object;
   TWABD_OnRequest = procedure(Session:TWABD_Session; First:boolean) of object;


   TWABD_ModuleClass = class of TDatamodule;

   TWABD_Session = class(TComponent)
   protected
      FAbout      : TWABDAbout;
      FMainBody   : TWABD_Body;
      FAuthBody   : TWABD_Body;
      FCurBody    : TWABD_Body;
      FNewBody    : TWABD_Body;     // To "Show" new body
      DidLogOff   : boolean;
      FTimeLen    : integer;
      FSessionID  : longint;
      FCreateTime : TDateTime;
      FLastAccess : TDateTime;
      FUserName   :string;
      FPassword   :string;
      FAuthenticated: Boolean;
      FQueryFields:TStrings;
      FHitCount:integer;            // Number of hits on this session.
      FVariables  : TStrings;
      FVariableType : TWABD_VariableType;
      FSessionMgr : TWABD_SessionMgr;
      FDetermineBrowser : boolean;
      FInfo       : string;
      FSemaphore  : THandle;        // Used to serialize requests for this session. (Frames could async. do requests)
      FLockCount  : integer;
      FStateless  : boolean;
      FEnableCookies:boolean;
      FVerDummy   : string;
      FProduce    : TWABD_Produce;

      FBrowser    : integer;
      FBrowserVersion : string;

      FRequest    : TWABD_CustomRequest;
      FResponse   : TWABD_CustomResponse;

      FModules    : TList;

      FUrlHandlers: TObject;

      FLogon      : TWABD_OnLogon;
      FLogoff     : TWABD_OnLogoff;
      FTimeOut    : TWABD_OnTimeOut;
      FGarbageCollection : TWABD_OnGarbageCollection;
      FExcept     : TWABD_OnException;
      FBeforeProcessRequest : TWABD_BeforeProcessRequest;
      FAfterProcessRequest : TWABD_AfterProcessRequest;
      FOnRequest  : TWABD_OnRequest;

      procedure   SetVariables(NewVariables: TStrings);
      procedure   SetQueryFields(NewQueryFields:TStrings);
      procedure   SetCookies(NewCookies:TWABD_Cookies);
      procedure   SetVariableByName(AName:string; AValue:string);
      procedure   SetCookieByName(AName:string; AValue:string);
      procedure   SetQueryFieldByName(AName:string; AValue:string);
      function    GetVariableByName(AName:string):string;
      function    GetCookieByName(AName:string):string;
      function    GetQueryFieldByName(AName:string):string;
      function    GetCookies:TWABD_Cookies;

      function    ProcessRequest(BodyName:string; Request:TWABD_CustomRequest): string;
      procedure   Notification(AComponent: TComponent; Operation: TOperation); override;
      procedure   Loaded; override;
      function    Authenticate:boolean;
      function    GetSesUserName:string;
      function    GetSesPassword:string;
      function    GetVersion:string;
      procedure   SetMainBody(ABody:TWABD_Body);
      procedure   SetAuthBody(ABody:TWABD_Body);
   public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;

      procedure   LogOff;
      procedure   ProcessVariables();
      function    FindComponent(AName:string):TComponent;
      procedure   SendFile(ContentType,FileName,AsFileName:string);
      procedure   LockSession;
      procedure   UnlockSession;
      procedure   Redirect(URL:string);
      procedure   UnlinkModules;
      procedure   UnlinkModule(ModuleClass:TWABD_ModuleClass);
      function    ValidateUrlForNewSession(const url: string): Boolean;
      function    Module(ModuleClass:TWABD_ModuleClass):TDataModule;
      function    GetSessionIDString(const objName: string): string;

      property    SessionID:longint read FSessionID;
      property    CurBody: TWABD_Body read FCurBody;
      property    NewBody: TWABD_Body read FNewBody write FNewBody;
      property    SessionMgr: TWABD_SessionMgr read FSessionMgr;
      property    UserName:string read GetSesUserName write FUserName;
      property    Password:string read GetSesPassword write FPassword;
      property    LastAccess:TDateTime read FLastAccess;
      property    CreateTime:TDateTime read FCreateTime;
      property    HitCount:integer read FHitCount;
      property    VariableByName[s:string]:string read GetVariableByName write SetVariableByName;
      property    CookieByName[s:string]:string read GetCookieByName write SetCookieByName;
      property    QueryFieldByName[s:string]:string read GetQueryFieldByName write SetQueryFieldByName;
      property    Cookies:TWABD_Cookies read GetCookies write SetCookies;
      property    QueryFields:TStrings read FQueryFields write SetQueryFields;
      property    Response:TWABD_CustomResponse read FResponse;
      property    Request:TWABD_CustomRequest read FRequest;
      property    Browser:integer read FBrowser;
      property    BrowserVersion:string read FBrowserVersion;
      property    urlHandlers: TObject read FUrlHandlers write FUrlHandlers;
      property    authenticated: Boolean read FAuthenticated write FAuthenticated;
   published
      property    MainBody: TWABD_Body read FMainBody write SetMainBody;
      property    AuthBody: TWABD_Body read FAuthBody write SetAuthBody;
      property    Variables: TStrings read FVariables write SetVariables;
      property    TimeOutLength: integer read FTimeLen write FTimeLen;
      property    About: TWABDAbout read FAbout write FAbout;
      property    DoDetermineBrowser: boolean read FDetermineBrowser write FDetermineBrowser;
      property    Info:string read FInfo write FInfo;
      property    Stateless:boolean read FStateless write FStateless;
      property    EnableCookies:boolean read FEnableCookies write FEnableCookies;
      property    Version:string read GetVersion write FVerDummy;
      property    Produce:TWABD_Produce read FProduce write FProduce;
      property    VariableType:TWABD_VariableType read FVariableType write FVariableType;

      property    OnFirstLogon: TWABD_OnLogon read FLogon write FLogon;
      property    OnLogoff: TWABD_OnLogoff read FLogoff write FLogoff;
      property    OnTimeOut: TWABD_OnTimeOut read FTimeOut write FTimeOut;
      property    OnGarbageCollection: TWABD_OnGarbageCollection read FGarbageCollection write FGarbageCollection;
      property    OnException: TWABD_OnException read FExcept write FExcept;
      property    BeforeProcessRequest : TWABD_BeforeProcessRequest read FBeforeProcessRequest write FBeforeProcessRequest;
      property    AfterProcessRequest : TWABD_AfterProcessRequest read FAfterProcessRequest write FAfterProcessRequest;
      property    OnRequest : TWABD_OnRequest read FOnRequest write FOnRequest;
   end;


   TWABD_ObjectClass = class of TWABD_Object;
   TWABD_Object = class(TComponent)
   protected
      FAbout      : TWABDAbout;
      FParent     : TWABD_object;
      FParentName : string;      // tmp to hold name before all components are loaded
      FVisible    : boolean;
      FPathInfo   : string;
      InLoaded    : boolean;
      FDependingOn: TWABD_Object;

      FOnChange   : TNotifyEvent;

      procedure   SetVisible(b: boolean);
      function    GetVisible:boolean;
      procedure   Changed;
      procedure   Loaded; override;
      procedure   ReadParentName(Reader: TReader);
      procedure   WriteParentName(Writer: TWriter);
      procedure   SetName(const Value: TComponentName); override;
      procedure   Notification(AComponent: TComponent; Operation: TOperation); override;
      function    GetSession: TWABD_Session;
      procedure   SetDependingOn(AObject:TWABD_Object);

      property    PathInfo:string read FPathInfo write FPathInfo;
   public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
      function    HTML_META(options: string): string; virtual;
      function    Object_To_HTML: string; virtual; abstract;
      procedure   SaveHTMLToFile(AFile:string);

      property    OnChange: TNotifyEvent read FOnChange write FOnChange;
      property    Session:TWABD_Session read GetSession;
   published
      property    Visible: boolean read GetVisible write SetVisible;
      property    About: TWABDAbout read FAbout write FAbout;
      property    DependingOn: TWABD_Object read FDependingOn write SetDependingOn;
   end;

   TWABD_ReflectNotify = procedure(Sender: TObject; AComponent: TComponent; Operation: TOperation) of object;


   // ************************************************************************
   // "Top" Level objects


   TWABDEditHTML = class(TPersistent)
   public
      ParHTML     : TWABD_Base_HTML;
   end;


   TWABD_Body = class(TWABD_Object)
   protected
      FOnCreate   : TNotifyEvent;
      FOnShow     : TNotifyEvent;
      FWSession   : TWABD_Session;
      FClientProcessTime: double;

      FEnableTemplates: boolean;
      FVariables:TStrings;
      FVariableType:TWABD_VariableType;

      FIsReload   : boolean;
      FUseSessionCookie:boolean;
      FSubmitCount  : longint;
      FHandleRequests: boolean;

      procedure   SetProperties(Request:TWABD_CustomRequest);


      procedure   DefineProperties(Filer: TFiler); override;
      procedure   Loaded; override;
      procedure   BuildTemplates(AProduce:TWABD_Produce); virtual;
      procedure   DoShow; virtual;

      procedure   SetVariables(NewVariables:TStrings);
      procedure   SetVariableByName(AName:string; AValue:string);
      function    GetVariableByName(AName:string):string;

      property    EnableTemplates:boolean read FEnableTemplates write FEnableTemplates;
      property    Variables:TStrings read FVariables write SetVariables;
      property    VariableType:TWABD_VariableType read FVariableType write FVariableType;
      property    VariableByName[s:string]:string read GetVariableByName write SetVariableByName;

   public
      RefNotify   : TWABD_ReflectNotify;     // No one should use this but it's Component Editor

      procedure   CreateSessionCookie;
      property    Session: TWABD_Session read FWSession write FWSession;
      property    ClientProcessTime:double read FClientProcessTime ;
      property    IsReload:boolean read FIsReload;

      constructor Create(AOwner:TComponent); override;
      destructor  Destroy; override;
  published
      property    OnCreate: TNotifyEvent read FOnCreate write FOnCreate;
      property    OnShow: TNotifyEvent read FOnShow write FOnShow;
      property    UseSessionCookie:boolean read FUseSessionCookie write FUseSessionCookie;
      property    HandleRequests:boolean read FHandleRequests write FHandleRequests;
   end;



   TWABD_Base_HTML = class(TWABD_Body)
   protected
      FSysResp  : TWABD_SysResponse;

      FHTML     : TStrings;
      procedure   Notification(AComponent: TComponent; Operation: TOperation); override;
      procedure   BuildTemplates(AProduce:TWABD_Produce); override;

   public
      function    Object_To_HTML:string; override;
      constructor Create(AOwner:TComponent); override;
      destructor  Destroy; override;
      procedure   Show;
      property    VariableType;
      property    HTML:TStrings read FHTML write FHTML stored true;
   published
      property    sysResponse: TWABD_SysResponse read FSysResp write FSysResp;
   end;

   TWABD_HTML = class(TWABD_Base_HTML)
   protected
      FEdHTML   : TWABDEditHTML;
      procedure   SetEdHTML(NewEdHTML: TWABDEditHTML);
   public
      constructor Create(AOwner:TComponent); override;
      destructor  Destroy; override;
      function    Object_To_HTML: string; override;
      property    VariableByName;
   published
      property    HTML;
      property    EnableTemplates;
      property    EditHTML: TWABDEditHTML read FEdHTML write SetEdHTML;
      property    Variables;
      property    VariableType;
   end;

   TWABD_HTMLFile = class(TWABD_Base_HTML)
   protected
      FFileName                   : TFileName;
      FSetup                      : TWABD_Setup;
      FLoadedWhen                 : TDateTime;
      FSecsBeforeReload           : integer;
      FCached                     : boolean;
      procedure   Notification(AComponent: TComponent; Operation: TOperation); override;
      procedure   SetSetup(ASetup:TWABD_Setup);
   public
      constructor Create(AOwner:TComponent); override;
      destructor  Destroy; override;
      function    Object_To_HTML: string; override;
      procedure   Reload;
      property    LoadedWhen:TDateTime read FLoadedWhen;
      property    VariableByName;
   published
      property    EnableTemplates;
      property    FileName:TFileName read FFileName write FFileName;
      property    SecsBeforeReload:integer read FSecsBeforeReload write FSecsBeforeReload;
      property    Cached:boolean read FCached write FCached;
      property    Setup:TWABD_Setup read FSetup write SetSetup;
      property    Variables;
      property    VariableType;
   end;



{********************************************************************************}
implementation

uses
   strUtils,
   System.DateUtils,
   superobject,
   WABD_Crypt;

{******************************************************************************}
{***************************** TSessionList ***********************************}
{******************************************************************************}

// Compile for Delphi <= XE2
{$IF CompilerVersion<=23}
function TSessionList.LockList(): TList<TWABD_Session>;
var
  i: Integer;
  pp: Pointer;
begin
  if Assigned(tempList) then
    tempList.Clear
  else
    tempList:=TList<TWABD_Session>.Create;
  lockls:=inherited LockList();
  for i:=0 to lockls.Count-1 do
  begin
    pp:=lockLs.Items[i];
    tempList.Add(TWABD_Session(pp));
  end;
  Result:=tempList;
end;

procedure TSessionList.UnLockList;
begin
  if Assigned(tempList) then
  begin
    FreeAndNil(tempList);
  end;
  inherited UnLockList;
end;

procedure TSessionList.DeleteLock(ind: integer);
begin
  if Assigned(lockLs) then
    lockLs.Delete(ind);
end;

{$ifend}

{******************************************************************************}
{***************************** TWABD_SessionMgr *******************************}
{******************************************************************************}

constructor TWABD_SessionMgr.Create(AOwner: TComponent);
var
   s:string;
begin
   inherited;
   CreateTime  := Now;
   SessionList := TSessionList.Create;
   FUniqueList := TThreadList.create;
   FRouteSites := TStringList.create;
   FRouteWhen  := rwNever;
   FGarbage    := True;
   FCheck      := 60;                     // Run the check every 60 seconds
   FVariables  := TStringList.create;
   FVariableType:=wabdVarTypeStandard;
   FHTMLTimeOut := TStringList.create;
   FWebAdmin   := 'the WEB administrator';
   FSiteName   := 'kbmWABD';
   FDefSesTimeout:=600;                   // 10 minutes.
   FMaxSessions:=-1;
   FRouteLast  :=-1;
   FMaxIdenticalUser:=-1;
   FMaxRequestSize:=-1;                   // Request size not limited.
   FRandomSessionID:=false;
   FTotalSessionCount:=0;
//gds
   FCrypted := false;
   FEmptyAsRoot := True;
   FBaseUrl := '/';
   FOnSystemXHttpResponse:=nil;
   FStorage:=storeNone;
   FSectionName:='SessionMgr';
   FAuthType:=atBasic;

   // Determine modulename.
   s:=GetFullModuleName;
   FModuleName:=ExtractFileName(s);
   FModulePath:=ExtractFilePath(s);

   // Register call backs.
   Register_WABD_Callback(ClientRequest);  // The interface between ISAPI and WABD on requests.
   Register_WABD_Term_Callback(OnTerminateCallback);  // The interface between ISAPI and WABD on termination.
   InitializeCriticalSection(SesMgrCSCreate);
   InitializeCriticalSection(SesMgrCSDestroy);
   InitializeCriticalSection(SesMgrCSAuth);
   GatherStatistics:=false;

   // Add an image number sequence.
   CreateSequence(WABD_IMAGE_SEQUENCE,0,false);

   StopEvent   := CreateEvent(nil, False, False, nil);
   GarbageThrd := TGarbageThread.Create(True);
   GarbageThrd.SesMgr := self;
   GarbageThrd.FreeOnTerminate := False;
   GarbageThrd.Resume;
end;

destructor TWABD_SessionMgr.Destroy;
var
   i:integer;
   lst:TTempSessionList;
begin
   // Before destruction, save info.
   if FAutoSave then Save;

   SetEvent(StopEvent);
   GarbageThrd.WaitFor;
   CloseHandle(StopEvent);

   // Free stats.
   if FStats<>nil then FStats.Free;

   // Free all sessions.
   lst:=SessionList.LockList;

   try
      for i:=0 to lst.count-1 do
          DoDestroySession(lst.Items[i]);
      lst.clear;
   finally
      SessionList.UnLockList;
   end;

   // Delete all global sequences.
   DeleteAllSequences;

   // Free remaining variables.
   FRouteSites.Free;
   SessionList.Free;
   GarbageThrd.Free;
   FVariables.free;
   FHTMLTimeOut.free;
   FUniqueList.free;
   DeleteCriticalSection(SesMgrCSCreate);
   DeleteCriticalSection(SesMgrCSDestroy);
   DeleteCriticalSection(SesMgrCSAuth);
   inherited;
end;

function TWABD_SessionMgr.GetVersion:string;
begin
     Result:=WABD_VERSION_STR;
end;

function TWABD_SessionMgr.GetFullModuleName:string;
var
   Buffer: string;
   BufferSize: DWORD;
begin
     BufferSize:=512;
     SetLength(Buffer,BufferSize);

     BufferSize:=GetModuleFileName(HInstance,PWideChar(Buffer),BufferSize-1);

     if BufferSize<>0 then
     begin
          SetLength(Buffer,BufferSize);
          Result:=Buffer;
     end
     else
         Result:='';
end;

procedure TWABD_SessionMgr.Save;
var
   jsonOb: ISuperObject;
   ini:TIniFile;
   i,j:integer;
   s:string;
   lst:TList;
begin
    case Storage of
        storeNone: ;
        storeIniFile:
          begin
            ini:=TIniFile.Create(FStoragePath);
            if ini<>nil then
            begin
                ini.WriteInteger(FSectionName,'MaxSessions',FMaxSessions);
                ini.WriteInteger(FSectionName,'MaxIdenticalUser',FMaxIdenticalUser);
                ini.WriteBool(FSectionName,'GatherStatistics',FGatherStats);
                ini.WriteBool(FSectionName,'NetworkStatistics',FNetworkStats);
                ini.WriteInteger(FSectionName,'CheckTimeOutInterval',FCheck);
                ini.WriteBool(FSectionName,'GarbageCollection',FGarbage);
                ini.WriteString(FSectionName,'WebAdmin',FWebAdmin);
                ini.WriteString(FSectionName,'SiteName',FSiteName);
                ini.WriteInteger(FSectionName,'SiteID',FSiteID);
                ini.WriteString(FSectionName,'RouteSites',FRouteSites.CommaText);
                ini.WriteInteger(FSectionName,'RouteWhen',ord(FRouteWhen));
                ini.WriteInteger(FSectionName,'RouteHow',ord(FRouteHow));
                ini.WriteInteger(FSectionName,'DefaultSessionTimeout',FDefSesTimeout);
                if assigned(FOnSave) then FOnSave(self,FSectionName,'',ini);
            end;

            // Save persistent sequences.
            s:=FSectionName+' Sequences';
            j:=0;
            lst:=FUniqueList.LockList;
            try
               for i:=0 to lst.count-1 do
               begin
                    with PWABD_SequenceRec(lst.Items[i])^ do
                    begin
                         if Persistent then
                         begin
                              ini.WriteString(s,inttostr(j),Name+':'+inttostr(Value));
                              inc(j);
                         end;
                    end;
               end;
            finally
               FUniqueList.UnlockList;
               ini.WriteInteger(s,'Sequences',j);
            end;
            ini.free;
          end;
        storeJsonFile:
          begin
            jsonOb:=tsuperobject.ParseString(PWideChar('{}'),true);
            jsonOb.O[FSectionName]:=tsuperobject.ParseString('{}',true);
            begin
                jsonOb.O[FSectionName].I['MaxSessions']:=FMaxSessions;
                jsonOb.O[FSectionName].I['MaxIdenticalUser']:=FMaxIdenticalUser;
                jsonOb.O[FSectionName].B['GatherStatistics']:=FGatherStats;
                jsonOb.O[FSectionName].B['NetworkStatistics']:=FNetworkStats;
                jsonOb.O[FSectionName].I['CheckTimeOutInterval']:=FCheck;
                jsonOb.O[FSectionName].B['GarbageCollection']:=FGarbage;
                jsonOb.O[FSectionName].S['WebAdmin']:=FWebAdmin;
                jsonOb.O[FSectionName].S['SiteName']:=FSiteName;
                jsonOb.O[FSectionName].I['SiteID']:=FSiteID;
                jsonOb.O[FSectionName].S['RouteSites']:=FRouteSites.CommaText;
                jsonOb.O[FSectionName].I['RouteWhen']:=ord(FRouteWhen);
                jsonOb.O[FSectionName].I['RouteHow']:=ord(FRouteHow);
                jsonOb.O[FSectionName].I['DefaultSessionTimeout']:=FDefSesTimeout;
                if assigned(FOnSave) then FOnSave(self,FSectionName, jsonOb.AsString,nil);
            end;

            // Save persistent sequences.
            s:=FSectionName+'_Sequences';
            jsonOb.O[s]:=tsuperobject.ParseString('{}',true);
            j:=0;
            lst:=FUniqueList.LockList;
            try
               for i:=0 to lst.count-1 do
               begin
                    with PWABD_SequenceRec(lst.Items[i])^ do
                    begin
                         if Persistent then
                         begin
                              jsonOb.O[s].S[inttostr(j)]:=inttostr(Value);
                              inc(j);
                         end;
                    end;
               end;
            finally
               FUniqueList.UnlockList;
                jsonOb.O[s].I['Sequences']:= j;
            end;
            jsonOb.SaveTo(FStoragePath,false,true);

          end;
    end;
end;

procedure TWABD_SessionMgr.Load;
var
   jsonOb: ISuperObject;
   ini:TIniFile;
   s,s1:string;
   i,j,k:integer;
begin
    case Storage of
        storeNone: ;
        storeJsonFile:
          begin
          try
            jsonOb:=tsuperobject.ParseFile(FStoragePath,true);
            begin
                FMaxSessions:=jsonOb.O[FSectionName].I['MaxSessions'];
                FMaxIdenticalUser:=jsonOb.O[FSectionName].I['MaxIdenticalUser'];
                GatherStatistics:=jsonOb.O[FSectionName].B['GatherStatistics'];
                FNetworkStats:=jsonOb.O[FSectionName].B['NetworkStatistics'];
                FCheck:=jsonOb.O[FSectionName].I['CheckTimeOutInterval'];
                FGarbage:=jsonOb.O[FSectionName].B['GarbageCollection'];
                FWebAdmin:=jsonOb.O[FSectionName].S['WebAdmin'];
                FSiteName:=jsonOb.O[FSectionName].S['SiteName'];
                FSiteID:=jsonOb.O[FSectionName].I['SiteID'];
                FRouteSites.CommaText:=jsonOb.O[FSectionName].S['RouteSites'];
                FRouteWhen:=TWABD_RouteWhen(jsonOb.O[FSectionName].I['RouteWhen']);
                FRouteHow:=TWABD_RouteHow(jsonOb.O[FSectionName].I['RouteHow']);
                FDefSesTimeout:=jsonOb.O[FSectionName].I['DefaultSessionTimeout'];
                if assigned(FOnLoad) then FOnLoad(self,FSectionName,  jsonOb.AsString,nil);
            end;
          except
            raise Exception.Create('Sessionmgr: Error in loading JSON parameters file');
          end;

            // Load persistent sequences.
            try
              s:=FSectionName+'_Sequences';
              begin
                   j:=jsonOb.O[s].I['Sequences'];
                   for i:=0 to j-1 do
                   begin
                        s1:=jsonOb.O[s].S[inttostr(i)];
                        k:=pos(':',s1);
                        CreateSequence(copy(s1,1,k-1),strtoint(copy(s1,k+1,length(s1))),true);
                   end;
              end;
            except
              //nop
            end;
          end;
        storeIniFile:
          begin
            ini:=TIniFile.Create(FStoragePath);
            if ini<>nil then
            begin
                FMaxSessions:=ini.ReadInteger(FSectionName,'MaxSessions',-1);
                FMaxIdenticalUser:=ini.ReadInteger(FSectionName,'MaxIdenticalUser',-1);
                GatherStatistics:=ini.ReadBool(FSectionName,'GatherStatistics',true);
                FNetworkStats:=ini.ReadBool(FSectionName,'NetworkStatistics',false);
                FCheck:=ini.ReadInteger(FSectionName,'CheckTimeOutInterval',60);
                FGarbage:=ini.ReadBool(FSectionName,'GarbageCollection',true);
                FWebAdmin:=ini.ReadString(FSectionName,'WebAdmin','the Web administrator');
                FSiteName:=ini.ReadString(FSectionName,'SiteName','-');
                FSiteID:=ini.ReadInteger(FSectionName,'SiteID',0);
                FRouteSites.CommaText:=ini.ReadString(FSectionName,'RouteSites','');
                FRouteWhen:=TWABD_RouteWhen(ini.ReadInteger(FSectionName,'RouteWhen',0));
                FRouteHow:=TWABD_RouteHow(ini.ReadInteger(FSectionName,'RouteHow',0));
                FDefSesTimeout:=ini.ReadInteger(FSectionName,'DefaultSessionTimeout',600);
                if assigned(FOnLoad) then FOnLoad(self,FSectionName,'',ini);
            end;

            // Load persistent sequences.
            s:=FSectionName+' Sequences';
            j:=ini.ReadInteger(s,'Sequences',0);
            for i:=0 to j-1 do
            begin
                 s1:=ini.ReadString(s,inttostr(i),'');
                 k:=pos(':',s);
                 if k<=0 then continue;
                 CreateSequence(copy(s,1,k-1),strtoint(copy(s,k+1,length(s))),true);
            end;
            ini.free;
          end;
    end;
end;

procedure TWABD_SessionMgr.Loaded;
begin
    inherited;
    if SiteID<0 then SiteID:=Random(255); // Max. 256 unique sites.
    if FAutoLoad then Load;
end;

procedure TWABD_SessionMgr.SetAdmin(AAdmin:TWABD_Admin);
begin
     FAdmin:=AAdmin;
     if FAdmin<>nil then FAdmin.FreeNotification(self);
end;

procedure TWABD_SessionMgr.Notification(AComponent: TComponent; Operation: TOperation);
begin
     inherited;
     if (Operation=opRemove) and (AComponent = FAdmin) then FAdmin:=nil;
end;

procedure TWABD_SessionMgr.SetSiteID(id:integer);
begin
     if (id<-1) or (id>=256) then
        raise ERangeError.Create('SiteID must be between -1 to 255.');
     FSiteID:=id;
end;

// Called by ISAPI when webserver want to shutdown the DLL.
function TWABD_SessionMgr.OnTerminateCallback(Flags:longint):boolean;
begin
     if Assigned(FOnTerminate) then Result:=FOnTerminate(Flags)
     else Result:=true;
     free;
end;

// Create a unique global sequence.
procedure TWABD_SessionMgr.CreateSequence(ID:string; StartValue:longint; Persistent:boolean);
var
   p:PWABD_SequenceRec;
   i:integer;
   lst:TList;
begin
     // Check if its there allready, dont recreate it but adjust startvalue if > old value.
     lst:=FUniqueList.LockList;
     try
        for i:=0 to lst.count-1 do
        begin
             with PWABD_SequenceRec(lst.Items[i])^ do
             begin
                  if Name=ID then
                  begin
                       if StartValue>Value then Value:=StartValue;
                       exit;
                  end;
             end;
        end;
     finally
        FUniqueList.UnlockList;
     end;

     // Add new member to the list.
     GetMem(p,sizeof(TWABD_SequenceRec));
     p^.Persistent:=Persistent;
     p^.Name:=StrNew(PAnsiChar(ID));
     p^.Value:=StartValue;
     FUniqueList.Add(p);
end;

// Draw a sequence number from a specified global sequence.
// Return -1 if sequence not found.
function TWABD_SessionMgr.DrawSequenceValue(ID:string):longint;
var
   i:integer;
   lst:TList;
begin
     Result:=-1;
     lst:=FUniqueList.LockList;
     try
        for i:=0 to lst.count-1 do
        begin
             with PWABD_SequenceRec(lst.Items[i])^ do
             begin
                  if Name=ID then
                  begin
                       Result:=Value;
                       inc(Value);
                  end;
             end;
        end;
     finally
        FUniqueList.UnlockList;
     end;
end;

// Remove global sequence.
procedure TWABD_SessionMgr.DeleteSequence(ID:string);
var
   i:integer;
   lst:TList;
begin
     lst:=FUniqueList.LockList;
     try
        for i:=0 to lst.count-1 do
        begin
             with PWABD_SequenceRec(lst.Items[i])^ do
             begin
                  if Name=ID then
                  begin
                       StrDispose(PAnsiChar(Name));
                       FreeMem(lst.Items[i]);
                       lst.Delete(i);
                       break;
                  end;
             end;
        end;
     finally
        FUniqueList.UnlockList;
     end;
end;

// Remove all global sequences.
procedure TWABD_SessionMgr.DeleteAllSequences;
var
   i:integer;
   lst:TList;
begin
     lst:=FUniqueList.LockList;
     try
        for i:=lst.count-1 downto 0 do
        begin
             with PWABD_SequenceRec(lst.Items[i])^ do
             begin
                  StrDispose(PAnsiChar(Name));
                  FreeMem(lst.Items[i]);
                  lst.Delete(i);
             end;
        end;
     finally
        FUniqueList.UnlockList;
     end;
end;

// Setup gathering of statistics.
procedure TWABD_SessionMgr.SetGatherStats(b:boolean);
begin
     if b=FGatherStats then exit;

     if b then
     begin
          if FStats=nil then
          begin
               FStats:=TWABD_SesStat.create;
               FStats.AddGroup(WABD_STATGRP_RESPONSE,wabdStatGroupTurnAround,70);
               FStats.AddGroup(WABD_STATGRP_NETRESPONSE,wabdStatGroupTurnAround,100);
               FStats.AddGroup(WABD_STATGRP_SENDSIZE,wabdStatGroupValue,50);
               FStats.AddGroup(WABD_STATGRP_RECVSIZE,wabdStatGroupValue,50);
          end;
     end;
     FGatherStats:=b;
end;

procedure TWABD_SessionMgr.DoDestroySession(Ses: TWABD_Session);
begin
   if not Assigned(OnDestroySession) then
      raise Exception.CreateFmt('DestroySession not defined for %s', [Name]);

   EnterCriticalSection(SesMgrCSDestroy);
   try
      OnDestroySession(Ses);
   finally
      LeaveCriticalSection(SesMgrCSDestroy);
   end;
end;

//gds crypt
procedure TWABD_SessionMgr.CodeHiddenNames;
//   WABD_SES_ID_STR                 // Hidden Text field to hold session information
//   WABD_CLIENTPROCESSTIME_STR      // Hidden Text field send by client of how long time from request to answer.
var
  ran: Integer;
begin
  ran := Random(300);
  WABD_SES_ID_STR := Format('%s%d',[wabd_coded_name_prefix, ran+Random(20)] );
  WABD_CLIENTPROCESSTIME_STR := Format('%s%d',[wabd_coded_name_prefix, ran+100+Random(20)] );

end;

procedure TWABD_SessionMgr.SetCrypted(flg_crypt: Boolean);
var
  i: Integer;
  s: TBytes;
begin
  FCrypted := flg_crypt;
  if flg_crypt = True then
  begin
      Randomize;
      SetLength(s, 10);
      for i:=0 to 9 do
      begin
        s[i] := Ord(Random(128));
      end;
      WABD_CryptSeed(s);
      CodeHiddenNames;
  end;
end;


procedure TWABD_SessionMgr.SetVariables(NewVariables: TStrings);
begin
     FVariables.Assign(NewVariables);
end;

procedure TWABD_SessionMgr.SetVariableByName(AName,AValue:string);
var
   i:integer;
   s:string;
begin
     i:=FVariables.IndexOfName(AName);
     s:=AName+'='+AValue;
     if i>=0 then FVariables.strings[i]:=s
     else FVariables.add(s);
end;

function TWABD_SessionMgr.GetVariableByName(AName:string):string;
begin
     Result:=FVariables.Values[AName];
end;

procedure TWABD_SessionMgr.CheckLogOff(Ses: TWABD_Session);
begin
   if Ses.DidLogOff then begin
      if Assigned(Ses.OnLogOff) then Ses.OnLogOff;
      SessionList.Remove(Ses);
      DoDestroySession(Ses);
   end;
end;

// Get sessioncount.
function TWABD_SessionMgr.GetSessionCount:integer;
var
   lst:TTempSessionList;
begin
     lst:=SessionList.LockList;
     try
        Result:=lst.count;
     finally
        SessionList.UnlockList;
     end;
end;

// Get next route according to setup.
function TWABD_SessionMgr.GetRoute:string;
var
   n:integer;
begin
     Result:='';
     n:=FRouteSites.Count;
     if (FRouteWhen=rwNever) or (n<=0) then exit;

     // OK, we can route. Determine to which host.
     case FRouteHow of
          rhRandom: FRouteLast:=Random(n);
          rhRoundRobin:
             begin
                  inc(FRouteLast);
                  if FRouteLast>=n then FRouteLast:=0;
             end;
     end;
     Result:=FRouteSites.Strings[FRouteLast];
end;

// Substitute all parameters.
function TWABD_SessionMgr.ProcessVariables(HTML:string):string;
begin
     Result:=Process_Variables(HTML,FVariables,FVariableType);
end;

function TWABD_SessionMgr.AuthBasicDecode(const crypted: string;
          var user: string; var pw: string): boolean;
var
  s, u, p: string;
  i: Integer;
begin
    s:=WABD_DecodeBase64(crypted);
    i:=pos(':',s);
    u:=copy(s,1,i-1);
    p:=copy(s,i+1,length(s));
    user:=u;
    pw:=p;
    Result:=True;
end;


// Check if user is authenticated by basic method.
function TWABD_SessionMgr.Authenticate(Request:TWABD_CustomRequest; Response:TWABD_CustomResponse;
         LikeToBeAdmin:boolean; var IsAdmin:boolean; var basicAuth: boolean):boolean;
var
   s:string;
   r:string;
   u, p: string;
   i:integer;
   auth:boolean;

   
begin
     IsAdmin:=false;
     basicAuth:=False;

     // If application dont want to authenticate, allow normal users to access site.
     // But potential administrators always have to identify themselfs.
     if not (LikeToBeAdmin) and not(Assigned(FOnAuthenticate)) then
     begin
          if Request.UserName='' then Request.UserName:=Request.RemoteUser;
          Result:=true;
          exit;
     end;

     if (LikeToBeAdmin) and (FAuthType<>atBasic) then
     begin
      IsAdmin:=True;
      Result:=True;
      Exit;
     end;

     // Check if authentication info is given.
     s:=Request.Auth;
     if s<>'' then
     begin
          // Is it basic?
          i:=pos(' ',s);
          if i>0 then
          begin
               r:=UpperCase(copy(s,1,i-1));
               if (r='BASIC')  then
               begin
                    AuthBasicDecode(copy(s,i+1,length(s)), u, p );
                    Request.UserName:=u;
                    Request.Password:=p;

                    r:=Request.RemoteHost;
                    auth:=false;

                    // If user wants to be admin.
                    if LikeToBeAdmin then
                    begin
                         if assigned(FAdmin) and (Admin.AutoLog>=logLevel1) then
                            Admin.LogFmt('REQ:User %s (%s) like to be administrator',[Request.UserName,r]);
                         if Assigned(FAdmin) and assigned(FCreateAdmin)
                            and (lowercase(FAdmin.AdminUser)=lowercase(Request.UserName))
                            and (lowercase(FAdmin.AdminPassword)=lowercase(Request.Password))
                            then
                            begin
                                 basicAuth:=True;
                                 auth:=true;
                                 IsAdmin:=true;
                            end;
                    end
                    else
                         if Assigned(FOnAuthenticate) then
                            FOnAuthenticate(r,Request.UserName,Request.Password,auth);
                    Result:=auth;

                    if assigned(FAdmin) and (FAdmin.AutoLog>=logLevel1) then
                    begin
                        if Result then
                            FAdmin.LogFmt('REQ:User %s (%s) authenticated',[Request.UserName,r])
                        else
                            FAdmin.LogFmt('REQ:User %s (%s) NOT authenticated',[Request.UserName,r]);
                    end;

                    if Result then exit;
               end;
          end;
     end;


     // Force authentication dialog.
     Response.ContentDesc:=SiteName;
     Response.Status:=WABD_STATUS_AUTH;

     Result:=false;
end;

// Count sessions with this userid.
function TWABD_SessionMgr.CountIdenticalUser(UserName:string):integer;
var
   i:integer;
   lst:TTempSessionList;
   ses: TWABD_Session;
begin
     Result:=0;
     UserName:=lowercase(trim(UserName));

     // Look for the username in the list.
     lst:=SessionList.LockList;
     try
       for ses in lst do
            if lowercase(trim(ses.UserName)) = UserName then inc(Result);
     finally
        SessionList.UnLockList;
     end;
end;

// Remove sessions for identical usernames until only MaxAllowedLeft is left.
procedure TWABD_SessionMgr.RemoveIdenticalUsers(UserName:string; MaxAllowedLeft:integer);
var
   i:integer;
   lst:TTempSessionList;
   ses:TWABD_Session;
begin
     UserName:=lowercase(trim(UserName));

     // Look for the username in the list.
     lst:=SessionList.LockList;
     try
          for ses in lst do
            if lowercase(trim(ses.UserName)) = UserName then
            begin
                 if MaxAllowedLeft<=0 then
                 begin
                      DoDestroySession(ses);
                      lst.Remove(ses);
                 end
                 else
                     dec(MaxAllowedLeft);
            end;
     finally
        SessionList.UnLockList;
     end;
end;

function TWABD_SessionMgr.CreateNewSession(Request:TWABD_CustomRequest; Response:TWABD_CustomResponse):TWABD_Session;
   procedure DoRoute(UserName:string);
   var
      Site:string;
   begin
        Site:=GetRoute;
        if Site<>'' then
        begin
             if assigned(FAdmin) and (FAdmin.AutoLog>=logLevel1) then
                Admin.LogFmt('REQ:Redirecting request for user %s to %s',[UserName,Site]);
             Response.Response.Clear;
             Response.Status:=WABD_STATUS_REDIRECT;
             Response.Location:=Site;
             exit;
        end
   end;

   function GetRandomSessionID:longint;
   var
      i:integer;
      l:longint;
      lst:TTempSessionList;
      ses: TWABD_Session;
   begin
        lst:=SessionList.LockList;
        try
           while true do
           begin
                // Draw number,
                l:=random(1 shl 23);

                // Check if number already used, then draw another.
                for ses in lst do
                    if ses.SessionID = l then break;
                // No conflict, use the one we have drawn.
                Result:=l;
                exit;
           end;
        finally
           SessionList.UnlockList;
        end;
   end;

   // Check limits. Doesnt apply to the administrator.
   function checkLimits(): Boolean;
   var
      c:integer;
      respStatus: Integer;
      respText: string;
   begin
      // Check if not accepting sessions here. Either reroute or give error.
      Result:=False;
      if FMaxSessions=0 then
      begin
           // Check if to route.
           if FRouteWhen=rwWhenFull then
              DoRoute(Request.UserName)
           else
           begin
            SendSystemResponse(srNoSes, Request.RequestType, respText, respStatus);
            Response.Response.Text:=respText;
            Response.Status:=respStatus;
           end;
           Result:=true;
           Exit;

      end;

      // Check if full. Either reroute or give error.
      if FMaxSessions>0 then
      begin
           c:=SessionCount;
           if c>=FMaxSessions then
           begin
                // Check if to route.
                if FRouteWhen=rwWhenFull then
                begin
                   DoRoute(Request.UserName);
                end else
                begin
                    SendSystemResponse(srFullSes, Request.RequestType, respText, respStatus);
                    Response.Response.Text:=respText;
                    Response.Status:=respStatus;
                end;
                Result:=true;
                Exit;
           end;
      end;

      // Check if to remove old users.
      if (FMaxIdenticalUser>0) and FAutoLogoutOldIdenticalUser then
        RemoveIdenticalUsers(Request.UserName,FMaxIdenticalUser-1);

      // Check if to count identical users.
      if (FMaxIdenticalUser>0) then
      begin
           if (CountIdenticalUser(Request.UserName) >= FMaxIdenticalUser) then
           begin
                SendSystemResponse(srFullUser, Request.RequestType, respText, respStatus);
                Response.Response.Text:=respText;
                Response.Status:=respStatus;
                Result:=true;
                Exit;
           end;
      end;
   end;
var
   NewSes : TWABD_Session;
   LikeToBeAdmin,IsAdmin,Authenticated:boolean;
   stamp:TDateTime;
   c:integer;
   s:string;
   NewSesID:longint;

   validated: Boolean;
   basicAuth: Boolean;
label
   L_Exit;
begin
   if not Assigned(OnCreateSession) then
      raise Exception.CreateFmt('CreateSession not defined for %s', [Name]);

   stamp:=now;

   NewSes := nil;
   Request.UserName:='';
   Request.Password:='';
   Result:=nil;

   Response.Status:=WABD_STATUS_OK;

   if assigned(FAdmin) and (FAdmin.AutoLog>=logLevel2) then
      FAdmin.LogFmt('REQ:Creating new session for user %s',[Request.UserName]);

   // Check if to do sessionmanager based (basic) authentication).
   LikeToBeAdmin:=(Request.Query.Count>0) and (lowercase(Request.Query.Strings[0])='admin');
   EnterCriticalSection(SesMgrCSAuth);
   try
      Authenticated:=Authenticate(Request,Response,LikeToBeAdmin,IsAdmin, basicAuth);
   finally
      LeaveCriticalSection(SesMgrCSAuth);
   end;

   // Check if to reroute allways for everyone except administrator.
   if (FRouteWhen=rwAllways) and (not (IsAdmin or LikeToBeAdmin)) then
   begin
        DoRoute(Request.UserName);
        exit;
   end;

   // Check if not authenticated and no guest session defined, exit.
   if (not Authenticated) and ((Request.UserName='')
         or (not Assigned(OnCreateGuestSession))) then goto L_exit;

   // Check limits. Doesnt apply to the administrator.
   if not IsAdmin then
   begin
        if checkLimits()=true then Exit;
   end;

   EnterCriticalSection(SesMgrCSCreate);
   try
      NewSes:=nil;
      // Check if to show another form if not authenticated.
      if not Authenticated then
      begin
           // Remove reauthentication request.
           Response.Status:=WABD_STATUS_OK;
           Response.ContentDesc:='';
           OnCreateGuestSession(NewSes,LikeToBeAdmin,Request, validated);
      end
      // Check if logged in as admin.
      else if IsAdmin and Assigned(OnCreateAdminSession) then
         OnCreateAdminSession(NewSes,Request, validated)
      // Else an ordinary user.
      else
         OnCreateSession(NewSes,Request, validated);

      if (validated) then
      begin
        Inc(FTotalSessionCount);
        if (IsAdmin) and (basicAuth=True) then
          NewSes.authenticated:=True;


        // If to use randomized session ID, get one.
        if FRandomSessionID then
           NewSesID:=GetRandomSessionID
        else
           NewSesID:=FTotalSessionCount;
      end;
   finally
      LeaveCriticalSection(SesMgrCSCreate);
   end;

   if (validated) then
   begin
       if NewSes=nil then
          raise Exception.Create('CreateSession Failed');
   end
   else
   begin
     if Response.Location<>'' then
         Response.Status:=WABD_STATUS_REDIRECT
     else
     begin
         Response.Status:=WABD_STATUS_ERROR;
         Response.Response.Text:='Internal error 81';
     end;
    if assigned(FAdmin)  then
    begin
        FAdmin.LogFmt('REQ: New session NOT validated from request path %s ',[Request.RequestPath]);
    end;
    Exit;
   end;

   NewSes.LockSession;
   try
       NewSes.FRequest:=Request;
       NewSes.FResponse:=Response;
       SessionList.Add(NewSes);

       // Set properties for new session.
       NewSes.FInfo:=FInfo;
       NewSes.FSessionID := (SiteID shl 24) + NewSesID;
       NewSes.FLastAccess := Now;
       NewSes.FCreateTime := Now;
       NewSes.FSessionMgr := self;

       // Get original username and password into session variables.
       NewSes.FUsername:=NewSes.GetSesUserName;
//       NewSes.FPassword:=NewSes.GetSesPassword;

       if NewSes.FTimeLen<0 then NewSes.FTimeLen:=DefaultSessionTimeout;

       Response.Status:=WABD_STATUS_OK;
       Result:=NewSes;

       if Assigned(FOnFirstSes) and (FTotalSessionCount=1) then FOnFirstSes(NewSes);

       if Assigned(NewSes.BeforeProcessRequest) then NewSes.BeforeProcessRequest(NewSes);
//       Response.Response.Text:=NewSes.ProcessVariables(NewSes.ProcessRequest('',Request));
       Response.Response.Text:=NewSes.ProcessRequest('',Request);
       if Assigned(NewSes.AfterProcessRequest) then NewSes.AfterProcessRequest(NewSes);

       // Check if stateless sessiontype. Then clearout session.
       if NewSes.Stateless then NewSes.LogOff;
   finally
       NewSes.UnlockSession;
   end;

   CheckLogOff(NewSes);

L_Exit:
   if GatherStatistics then
   begin
        if NewSes<>nil then
           s:='Site:'+inttostr(NewSes.SessionID shr 24) + ' Session ID:'+inttostr(NewSes.SessionID and $FFFFFF)+' (New)'
        else
            s:='No session';
        Stats.AddPoint(WABD_STATGRP_RESPONSE,Request.UserName,s,now-stamp);
        Stats.AddPoint(WABD_STATGRP_SENDSIZE,Request.UserName,s,length(Response.Response.Text));
        Stats.AddPoint(WABD_STATGRP_RECVSIZE,Request.UserName,s,Request.Size);
   end;
end;

function TWABD_SessionMgr.LocateSessionByID(ASiteID:integer; ASessionID:longint):TWABD_Session;
var
   sList    : TTempSessionList;
   i        : integer;
   ts       : TWABD_Session;
   l        : longint;
begin
   // We have an existing session
   l:=(ASiteID shl 24) + ASessionID;
   Result:=nil;
   sList:=SessionList.LockList;
   try
      for ts in sList do
      begin
           if ts.SessionID = l then
           begin
                Result:=ts;
                break;
           end;
      end;
   finally
      SessionList.UnlockList;
   end;
end;

function TWABD_SessionMgr.RunExistingSession(ASession:TWABD_Session; const IdStr,BodyName:string;
                               Request:TWABD_CustomRequest; Response:TWABD_CustomResponse):TWABD_Session;
var
   stamp    : TDateTime;
   s        : string;
begin
   // start of session processing
   stamp:=now();

   Result:=ASession;
   Result.FLastAccess:=Now();

   if assigned(FAdmin) and (Admin.AutoLog>=logLevel2) then
      Admin.LogFmt('REQ:Request from site %d, session %d. This site is %d',
          [Result.SessionID shr 24, Result.SessionID and $FFFFFF,SiteID]);

   Result.LockSession;
   try
      Result.FRequest:=Request;
      Result.FResponse:=Response;

      Response.Status:=WABD_STATUS_OK;
      if Assigned(Result.BeforeProcessRequest) then
          Result.BeforeProcessRequest(Result);

//      Response.Response.Text := Result.ProcessVariables(Result.ProcessRequest(BodyName,Request));
      Response.Response.Text := Result.ProcessRequest(BodyName,Request);

      if Assigned(Result.AfterProcessRequest) then
          Result.AfterProcessRequest(Result);

      // Check if stateless sessiontype. Then clearout session.
      if Result.Stateless then Result.LogOff;
   finally
      Result.UnlockSession;
   end;

   if GatherStatistics then
   begin
        s:='Site:'+inttostr(Result.SessionID shr 24) + ' Session ID:'+inttostr(Result.SessionID and $FFFFFF)+' Net address:'+Request.RemoteAddr;
        Stats.AddPoint(WABD_STATGRP_RESPONSE,Result.UserName,s, SecondSpan(now() , stamp));
        Stats.AddPoint(WABD_STATGRP_SENDSIZE,Result.UserName,s,length(Response.Response.Text));
        Stats.AddPoint(WABD_STATGRP_RECVSIZE,Result.UserName,s,Request.Size);
        if NetworkStatistics and (Result.CurBody.FClientProcessTime>0) then
           Stats.AddPoint(WABD_STATGRP_NETRESPONSE,
                          Result.UserName,
                          s+' Input:'+Request.Query.Text,
                          SecondSpan(Now() , FloatToDateTime(Result.FCurBody.FClientProcessTime)) );
   end;

   CheckLogOff(Result);
end;

function TWABD_SessionMgr.Encrypt(const sval: string): string;
begin
  Result:=WABD_Encrypt(sval);
end;

function TWABD_SessionMgr.Decrypt(const sval: string): string;
begin
  Result:=WABD_Decrypt(sval);
end;

procedure TWABD_SessionMgr.SendSystemResponse(typ: TWABD_SysResponse;
      requestType: Integer; var respText: string; var respStatus: integer);
var
  i: Integer;
  htmlMes: TWABD_Base_HTML;

begin
  htmlMes:=nil;
  for i:=0 to  self.Owner.ComponentCount-1 do
  begin
    if (Owner.Components[i] is TWABD_Base_HTML) 
       and (TWABD_Base_HTML(Owner.Components[i]).sysResponse=typ) then
       begin
        htmlMes:=TWABD_Base_HTML(Owner.Components[i]);
        Break;
       end;
  end;
  
  case typ of
  srExpired: begin
       if requestType=WABD_REQUESTTYPE_HTML then
       begin
          if htmlMes<>nil then
            respText:=htmlMes.Object_To_HTML
          else
            respText:=FHTMLTimeOut.Text;
          respStatus:=WABD_STATUS_OK;
       end else
       begin
          if Assigned(FOnSystemXHttpResponse) then
              FOnSystemXHttpResponse(Request, srExpired, respText)
          else
              respText:='SESSION EXPIRED';
          respStatus:=WABD_STATUS_ERROR;
       end;
   end;
   srNoHandler: begin
      if Assigned(FOnSystemXHttpResponse) then
          FOnSystemXHttpResponse(Request, srNoHandler, respText)
      else
          respText:='NO HANDLER FOR URL';
      respStatus:=WABD_STATUS_ERROR;

   end;
   srNoSes: begin
   
          if htmlMes<>nil then
            respText:=htmlMes.Object_To_HTML
          else
             respText:='<BODY><P><H1>Welcome</H1></P><P>'
                +SiteName+'('+inttostr(SiteID)
                +') is currently not accepting new sessions.</P><P>Please try later.</P></BODY>';
          respStatus:=WABD_STATUS_OK;
   end;
   srFullSes: begin
          if htmlMes<>nil then
            respText:=htmlMes.Object_To_HTML
          else
              respText:='<BODY><P><H1>Welcome</H1></P><P>'+SiteName+'('+inttostr(SiteID)
              +') is currently fully booked. </P><P>Please try later.</P></BODY>';
          respStatus:=WABD_STATUS_OK;
   end;
   srFullUser: begin
          if htmlMes<>nil then
            respText:=htmlMes.Object_To_HTML
          else
             respText:='<BODY><P><H1>Welcome</H1></P><P>'+SiteName
             +'('+inttostr(SiteID)+') only allow you to be logged in '+inttostr(FMaxIdenticalUser)
             +' times with the same username.</P>'
             + '<P>If you cant do a proper logout at this time then please wait another 10-20 minutes,'
             + ' and try to logon again.</P></BODY>';
          respStatus:=WABD_STATUS_OK;
   end;
   srHttp403: begin
          if htmlMes<>nil then
            respText:=htmlMes.Object_To_HTML
          else
             respText:='<BODY><P><H1>Server error/H1></P><P></BODY>';
          respStatus:=WABD_STATUS_ERROR;
   
   end;
  end;

end;

// Called by interface whenever a request is made from a client.
procedure TWABD_SessionMgr.ClientRequest(Request:TWABD_CustomRequest; Response:TWABD_CustomResponse);
var
   s           : string;
   IdStr       : string;
   BodyName    : string;
   ses         : TWABD_Session;
   ok          : boolean;
   cookie      : TWABD_Cookie;
   ele         : TCollectionItem;
   handled     : boolean;
   sid,sesid   : longint;
   l           : longint;

   i           : integer;
   rs          : string;
   status      : Integer;
begin
   ses:=nil;
   FRequest:=Request;
   FResponse:=Response;
   try
      try
         Request.DetermineRequestType;
         VariableByName[WABD_VAR_SERVERAPP]:=Request.ScriptName;
         VariableByName[WABD_VAR_SERVERNAME]:=Request.ServerName;


// se attivata crypt
 //   decritta in Request.Query variabili in lista crypt
        if FCrypted = true then
         begin
           with Request.Query do
           begin
              if Values[WABD_SES_ID_STR]<>'' then
                    Values[WABD_SES_ID_STR] := WABD_Decrypt(Values[WABD_SES_ID_STR]);
              if Values[WABD_CLIENTPROCESSTIME_STR]<>'' then
                    Values[WABD_CLIENTPROCESSTIME_STR] := WABD_Decrypt(Values[WABD_CLIENTPROCESSTIME_STR]);
           end;
//   decritta in Request.Cookies variabili in lista crypt
           with Request.Cookies do
           begin
              cookie := GetCookieByName(WABD_SES_ID_STR);
              if (cookie<>nil) then
              begin
                  cookie.Value := WABD_Decrypt(cookie.Value);
              end;
              cookie := GetCookieByName(WABD_CLIENTPROCESSTIME_STR);
              if (cookie<>nil) then
              begin
                  cookie.Value := WABD_Decrypt(cookie.Value);
              end;
           end;

         end;


         // Check if to validate contenttype.
         ok:=true;
         if Assigned(FOnValidateRequest) then
            FOnValidateRequest(Request,Response,ok);
         if not ok then
         begin
              if Response.Response.Count=0 then raise Exception.CreateFmt('Content type ''%s'' not accepted.',[Request.ContentType])
              else exit;
         end;

         // Process request.
         if assigned(Admin) and (Admin.AutoLog=logAll) then
                  Admin.LogFmt('HEADER:%s, QUERY:%s',[Request.Headers.Text,Request.Query.Text]);

         // Try to find session string in cookies or form values.
         IdStr:='';
         cookie:=Request.Cookies.GetCookieByName(WABD_SES_ID_STR);

         if (cookie<>nil) then
             IdStr:=cookie.Value
         else
         begin
             s:=Request.Query.Values[WABD_SES_ID_STR];
             if s<>'' then IdStr:=s;
         end;

         // Prepare response.
         Response.ContentType:='text/html';
         Response.ContentDesc:='';
         Response.Location:='';
         Response.Cookies.Assign(Request.Cookies);
         for ele in Response.Cookies do
         begin
            cookie:=TWABD_Cookie(ele);
            // set the path to root path
            cookie.Path:='/';
            // mark for deletion old system cookies
            // they have a standard prefix.
            // Only session ID cookie and server data are
            // retained.
            // All other application cookies are retained, too.
            if (cookie.Name<>WABD_SES_ID_STR)  and
              (cookie.Name<>WABD_CLIENTPROCESSTIME_STR) and
              (  (Pos(WABD_CODED_NAME_PREFIX, cookie.Name)=1)
                or  (Pos(WABD_NAME_PREFIX, cookie.Name)=1) )
              then
              begin
                cookie.Expires:=StrToDateTime('01/01/1970');
                cookie.Value:='cancellato';
              end;
         end;




         if FCrypted = true then
         begin
//   cripta in Response.Cookies variabili in lista crypt
           with Response.Cookies do
           begin
              cookie := GetCookieByName(WABD_SES_ID_STR);
              if (cookie<>nil) then
              begin
                  cookie.Value := WABD_Encrypt(cookie.Value);
              end;
              cookie := GetCookieByName(WABD_CLIENTPROCESSTIME_STR);
              if (cookie<>nil) then
              begin
                  cookie.Value := WABD_Encrypt(cookie.Value);
              end;
           end;

         end;


         // if wants that empty request path is equal to root Url '/',
         // redirect to site root
         if FEmptyAsRoot then
         begin
             if Request.RequestPath='' then
             begin
                 Response.Status:=WABD_STATUS_REDIRECT;
                 Response.Location:=Request.ScriptName+'/';
                 exit;
             end;
         end;



         // Determine if new or old session.
         SplitSessionID(IdStr,IdStr,BodyName);

         // Find session.
         if IdStr<>'' then
         begin
              l:=strtoint(IdStr);
              sid:=l shr 24;
              sesid:=l and $FFFFFF;
              Ses:=LocateSessionByID(sid,sesid);

              // Check if session found.
              if Ses=nil then
              begin
                   // Check if to create new session automatically. If yes, let it go through.
                   ok:=false;
                   if Assigned(FOnInvalidSession) then
                      FOnInvalidSession(self,ok);

                   if not ok then
                   begin
                      SendSystemResponse(srExpired, Request.RequestType, rs, status);
                      response.Response.Text:=rs;
                      response.Status:=status;
                      Exit;
                   end;
              end;
         end
         else
             Ses:=nil;

         if Ses=nil then
            Ses := CreateNewSession(Request,Response)
         else
            Ses := RunExistingSession(Ses, IdStr, BodyName, Request,Response);

         if FNetWorkStats then
         begin
            rs:= format('CLPRT:%f',[Double(Now)]);
            if FCrypted then
              rs:=WABD_Encrypt(rs);
            Response.Cookies.Add(WABD_CLIENTPROCESSTIME_STR, rs );
         end;


         // Format result to match HTTP protocol.
//         Response.Response.Text:=ProcessVariables(Response.Response.Text);
      except
         on e: Exception do begin
            Response.Status:=WABD_STATUS_OK;
            Response.Response.Clear;
            handled:=false;
            if Assigned(FOnException) then
               FOnException(self,e,handled);
            if not handled then
            begin
                      Response.Response.Add('<!DOCTYPE HTML><HTML><BODY><tt><font size=5><u>E R R O R</u></font><br>'+
                         '<br>Please report this message and what happened to '+WebAdministrator+'.<br>'+
                         '<br>  Site........:<b>'+SiteName+'('+inttostr(SiteID)+')</b>'+
                         '<br>  Application.:<b>'+Request.ScriptName+'</b>');
            // gds crypt
                      if not(FCrypted) then
                         Response.Response.Add('<br>  Error.......:<b>'+e.Message+'</b>');
                      if not(FCrypted) and  (Ses<>nil) then
                      begin
                           Response.Response.Add('<br>  Session site:<b>'+inttostr(Ses.SessionID shr 24)+'</b>');
                           Response.Response.Add('<br>  Session ID..:<b>'+inttostr(Ses.SessionID and $FFFFFF)+'</b>');
                           if Ses.NewBody<>nil then Response.Response.Add('<br>  Body........:<b>'+Ses.NewBody.Name+'</b>');
                      end;
                      Response.Response.Add('<br><br><hr><i>'+DateTimeToStr(Now)+'</i><br><hr></tt></BODY></HTML>');
                      Response.ContentType:='text/html';
            end;
         end;
      end;
   finally
      FRequest:=nil;
      FResponse:=nil;
   end;
end;

procedure TGarbageThread.DoGarbageCollection;
var
   sList : TTempSessionList;
   i     : integer;
   s     : TWABD_Session;
   Elap  : integer;
   AcceptTimeOut: boolean;
   s1    : string;
begin
   if SesMgr.GarbageCollection = False then exit;

   // Clean up stat. records.
   if SesMgr.FStats<>nil then SesMgr.FStats.Clean;

   if Assigned(SesMgr.OnGarbageCollection) then SesMgr.OnGarbageCollection(SesMgr);

   sList := SesMgr.SessionList.LockList;
   try
      for i := sList.Count-1 downto 0 do begin
         s := sList.Items[i];

         // If the session want to get a garbagecollection signal whether or not it is due to timeout.
         if Assigned(s.OnGarbageCollection) then s.OnGarbageCollection(SesMgr);

         // Calculate elapsed time since last access.
         Elap := SecondsBetween(Now(), s.FLastAccess);
         if Elap > s.TimeOutLength then begin

            // Ask politely if to accept timeout.
            AcceptTimeOut:=true;
            if Assigned(s.OnTimeOut) then s.OnTimeOut(SesMgr,Elap,AcceptTimeOut);
            if not AcceptTimeOut then
               s1:='NOT ACCEPTED'
            else
               s1:='ACCEPTED';

            // If session accepts timeout, time it out.
            if AcceptTimeOut then
            begin
               SesMgr.DoDestroySession(s);
              // Compile for Delphi <= XE2
              {$IF CompilerVersion<=23}
                  SesMgr.SessionList.DeleteLock(i);
              {$else}
                  sList.Delete(i);
              {$ifend}
            end;
         end;
      end;
   finally
      SesMgr.SessionList.UnlockList;
   end;
end;

procedure TGarbageThread.Execute;
var
   rc : integer;
begin
   repeat
      rc := WaitForSingleObject(SesMgr.StopEvent, SesMgr.FCheck * 1000);
      if rc = WAIT_TIMEOUT	then DoGarbageCollection;
   until rc <> WAIT_TIMEOUT;
end;


{******************************************************************************}
{***************************** TWABD_Session **********************************}
{******************************************************************************}

constructor TWABD_Session.Create(AOwner: TComponent);
begin
   inherited;
   TimeOutLength := 600;
   FVariables := TStringList.create;
   FVariableType:=wabdVarTypeStandard;
   FQueryFields:=TStringList.Create;
   FHitCount := 0;
   FStateless := false;
   FEnableCookies := true;
   FSemaphore:=CreateSemaphore(nil,1,1,nil);
   FProduce:=prodHTML;
   FUrlHandlers:=nil;
   FAuthenticated:=False;

   // Default we dont know Jack sh.. about what browser the user has.
   FBrowser:=WABD_BrowserUnknown;
   FBrowserVersion:='';

   // Add list of modules which the session object can use.
   FModules:=TList.Create;
end;

destructor TWABD_Session.Destroy;
begin
   FVariables.Free;
   FQueryFields.Free;
   CloseHandle(FSemaphore);

   // Remove modules.
   UnlinkModules;
   inherited;
end;

function TWABD_Session.GetVersion:string;
begin
     Result:=WABD_VERSION_STR;
end;

// Lock this session to synchronize access to it.
procedure TWABD_Session.LockSession;
var
   n:DWORD;
begin
     inc(FLockCount);
     n:=WaitForSingleObject(FSemaphore,WABD_SEMAPHORE_TIMEOUT);
     if (n=WAIT_TIMEOUT) or (n=WAIT_FAILED) then
     begin
          dec(FLockCount);
     end;
end;

procedure TWABD_Session.UnlockSession;
begin
     if FLockCount<=0 then
     begin
          FLockCount:=0;
          exit;
     end;
     dec(FLockCount);
     ReleaseSemaphore(FSemaphore,1,nil);
end;

procedure TWABD_Session.Redirect(URL:string);
begin
     Response.Response.Clear;
     Response.Status:=WABD_STATUS_REDIRECT;
     Response.Location:=URL;
end;

// Get username for this session.
function TWABD_Session.GetSesUserName:string;
begin
     Result:=FUserName;
     if (Result='') and (FRequest<>nil) then
     begin
          Result:=FRequest.UserName;
          if (Result='') then Result:=FRequest.RemoteUser;
     end;
end;

// Get password for this session.
function TWABD_Session.GetSesPassword:string;
begin
     Result:=FPassword;
     if (Result='') and (FRequest<>nil) then
     begin
          Result:=FRequest.Password;
     end;
end;

// Check if to initiate Form authentication of user.
function TWABD_Session.Authenticate:boolean;
begin
     Result:=true;

     // Dont authenticate.
     if (not Assigned(FAuthBody))
         // just authenticated with HTTP Basic method
        or (FAuthenticated)
     then exit;

     // Show authentication form.
     FCurBody:=AuthBody;
     Result:=false;
end;

procedure TWABD_Session.SendFile(ContentType,FileName,AsFileName:string);
var
   s:string;
   hf:integer;
   buffer:string;
   cbuffer:array [0..2047] of ansichar;
   n:integer;
begin

   // Open file and figure out its size.
   hf:=FileOpen(FileName, fmOpenRead+fmShareDenyWrite);
   try
      if hf<0 then RaiseLastOSError;
      FileSeek(hf,0,2);
      FileSeek(hf,0,0);

      // Loop until whole file has been written to webstream.
      s:='';
      while true do begin
         n:=FileRead(hf,cbuffer,2048);
         if n<0 then RaiseLastOSError;
         if n>0 then
         begin
              SetString(buffer,cbuffer,n);
              s:=s+buffer;
         end;
         if n<2048 then break;
      end;
   finally
      // Done with it, close file.
      FileClose(hf);
   end;

   // Override normal response.
   Response.ContentType:=ContentType;
   Response.ContentDesc:=AsFileName;
   Response.BinaryResponse:=s;
end;

procedure TWABD_Session.SetVariables(NewVariables: TStrings);
begin
     FVariables.Assign(NewVariables);
end;

procedure TWABD_Session.SetVariableByName(AName,AValue:string);
var
   i:integer;
   s:string;
begin
     i:=FVariables.IndexOfName(AName);
     s:=AName+'='+AValue;
     if i>=0 then FVariables.strings[i]:=s
     else FVariables.add(s);
end;

function TWABD_Session.GetVariableByName(AName:string):string;
begin
     Result:=FVariables.Values[AName];
end;

procedure TWABD_Session.SetCookies(NewCookies: TWABD_Cookies);
begin
     FResponse.Cookies.Assign(NewCookies);
end;

function TWABD_Session.GetCookies:TWABD_Cookies;
begin
     Result:=FResponse.Cookies;
end;

procedure TWABD_Session.SetCookieByName(AName,AValue:string);
var
   c:TWABD_Cookie;
begin
     c:=FResponse.Cookies.GetCookieByName(AName);
     if (c<>nil) then
        c.Value:=AValue
     else
         FResponse.Cookies.Add(AName,AValue);
end;

function TWABD_Session.GetCookieByName(AName:string):string;
var
   c:TWABD_Cookie;
begin
     c:=FResponse.Cookies.GetCookieByName(AName);
     if (c<>nil) then
        Result:=c.Value
     else
        Result:='';
end;

procedure TWABD_Session.SetQueryFields(NewQueryFields:TStrings);
begin
     FQueryFields.Assign(NewQueryFields);
end;

procedure TWABD_Session.SetQueryFieldByName(AName,AValue:string);
var
   i:integer;
   s:string;
begin
     i:=FQueryFields.IndexOfName(AName);
     s:=AName+'='+AValue;
     if i>=0 then FQueryFields.strings[i]:=s
     else FQueryFields.add(s);
end;

function TWABD_Session.GetQueryFieldByName(AName:string):string;
begin
     Result:=FQueryFields.Values[AName];
end;

// Update variables list with the ones from session manager
procedure TWABD_Session.ProcessVariables();
var
  i: Integer;
begin
     for i:=0 to SessionMgr.Variables.Count-1 do
     begin
         VariableByName[SessionMgr.Variables.Names[i]]:=SessionMgr.Variables.ValueFromIndex[i];
     end;
end;

procedure TWABD_Session.SetMainBody(ABody:TWABD_Body);
begin
     FMainBody:=ABody;
     if FMainBody<>nil then FMainBody.FreeNotification(self);
end;

procedure TWABD_Session.SetAuthBody(ABody:TWABD_Body);
begin
     FAuthBody:=ABody;
     if FAuthBody<>nil then FAuthBody.FreeNotification(self);
end;

procedure TWABD_Session.Notification(AComponent: TComponent; Operation: TOperation);
begin
   inherited;
   if (Operation = opRemove) then
   begin
        if (AComponent = MainBody) then FMainBody := nil
        else if (AComponent = AuthBody) then FAuthBody := nil;
   end;
end;

// Remove all activated modules.
procedure TWABD_Session.UnlinkModules;
var
   i:integer;
   m:TDataModule;
begin
     for i:=FModules.Count-1 downto 0 do
     begin
          m:=TDataModule(FModules.Items[i]);
          m.Free;
     end;
     FModules.Clear;
end;

// Remove specific activated module.
procedure TWABD_Session.UnlinkModule(ModuleClass:TWABD_ModuleClass);
var
   i:integer;
   m:TDataModule;
begin
     // Look for module.
     for i:=0 to FModules.Count-1 do
     begin
          m:=TDataModule(FModules.Items[i]);
          if m is ModuleClass then
          begin
               m.Free;
               FModules.delete(i);
               exit;
          end;
     end;
end;

// Find existing module and return it.
// If none found, add new instance of it.
function TWABD_Session.Module(ModuleClass:TWABD_ModuleClass):TDataModule;
var
   i:integer;
begin
     // Check if session module.
     if ModuleClass=Owner.ClassType then
     begin
          Result:=TDataModule(Owner);
          exit;
     end;

     // Look for module.
     for i:=0 to FModules.Count-1 do
     begin
          if TDataModule(FModules.Items[i]) is ModuleClass then
          begin
               Result:=TDataModule(FModules.Items[i]);
               exit;
          end;
     end;

     // None found, add one.
     Result:=ModuleClass.Create(self);
     FModules.Add(Result);

     // Find all bodys on that module and link them to self session.
     for i:=0 to Result.ComponentCount-1 do
     begin
          if Result.Components[i] is TWABD_Body then
             TWABD_Body(Result.Components[i]).Session:=self;
     end;
end;


function TWABD_Session.ProcessRequest(BodyName:string; Request:TWABD_CustomRequest): string;
var
   b: TWABD_Body;
   o: TObject;
   h: boolean;
   nm: Integer;
   resp: string;
   respSt: Integer;
   cookie: TWABD_Cookie;

   function processEx(): Boolean;
   begin
     Result:=False;
     h:=false;
     if assigned(FExcept) then FExcept(self,Exception(ExceptObject),h);
     if (not h) or (FNewBody=nil) then
     begin
       Exit;
     end;
     if FNewBody<>FCurBody then
     begin
          FCurBody:=FNewBody;
          Response.Response.Text:=FCurBody.Object_To_HTML;
     end;
     Result:=True;
   end;


   procedure prepareResp();
   var
    nm: Integer;
   begin
       assert(assigned(FUrlHandlers),'Session has no url handlers container');
       nm:=RunHandlerForUrl(FUrlHandlers, Request.RequestPath, Request.RequestMethod,
              Request, Response );

       // no handler found for url, then ...
       if nm=0 then
       begin
          // default Hmtl producer
          if Request.RequestType=WABD_REQUESTTYPE_HTML then
            Response.Response.Text:=FCurBody.Object_To_HTML
          else
          begin
            // or XHtml system response from session manager
            SessionMgr.SendSystemResponse(srNoHandler, request.RequestType,
                       resp, respSt );
            Response.Response.Text:=resp;
            Response.Status:=respSt;
          end;

       end;
   end;

begin
   inc(FHitCount);

   FRequest:=Request;

   if FDetermineBrowser then
   begin
        Request.DetermineBrowser;
        FBrowser:=Request.Browser;
        FBrowserVersion:=Request.BrowserVersion;
   end;

   if BodyName='' then   // first access
   begin

      // This is a new logon
      if not Assigned(FMainBody) then
         raise Exception.CreateFmt('MainBody not defined for Session: %s', [Name]);

      FMainBody.FWSession := self;
      FNewBody := MainBody;

      if Assigned(OnFirstLogon) then OnFirstLogon(FRequest.RemoteHost); // NewBody might change
      if Assigned(FOnRequest) then FOnRequest(Self,true);
      FCurBody := FNewBody;

      // Check if to do session based authentication (form authentication).
      Authenticate;

      if FCurBody<>nil then
      begin
           FCurBody.FIsReload:=false;
           FCurBody.FWSession := self;

           // If an exception occurs, let program intercept.
           try
              prepareResp;
              Result:=Response.Response.Text;
              if FCurBody.FUseSessionCookie then FCurBody.CreateSessionCookie;
           except
              if not(processEx) then raise
              else
                Result:=Response.Response.Text;
           end;
      end;

   end else begin
      // If request procedure, call it.
      if Assigned(FOnRequest) then FOnRequest(Self,false);

      try
        //path broker
         prepareResp;

         FCurBody.SetProperties(Request);
         Result:=Response.Response.Text ;
         if FMainBody.FUseSessionCookie then FMainBody.CreateSessionCookie;
      except
        if not(processEx) then raise
        else
           Result:=Response.Response.Text;
      end;
   end;

   FLastAccess := Now;
end;

procedure TWABD_Session.LogOff;
begin
   DidLogOff := True;
end;

procedure TWABD_Session.Loaded;
var
   c : TComponent;
   i : integer;
begin
   inherited;
   Assert(Owner<>nil, Format('TWABD_Session %s has no owner', [Name]));
   for i := 0 to Owner.ComponentCount-1 do begin
      c := Owner.Components[i];
      if c is TWABD_Body then (c as TWABD_Body).Session := self  ;
   end;
end;

function TWABD_Session.FindComponent(AName:string):TComponent;
  function FindComponentRecursive(ARoot:TComponent; AName:string):TComponent;
  var
     i:integer;
  begin
       for i:=0 to ARoot.ComponentCount-1 do
       begin
            Result:=ARoot.Components[i];
            if LowerCase(Result.Name)=AName then exit;
            if (Result.ComponentCount<>0) then
            begin
                 Result:=FindComponentRecursive(Result,AName);
                 if Result<>nil then exit;
            end;
       end;
       Result:=nil;
  end;
var
   i:integer;
begin
     Result:=nil;
     AName:=LowerCase(AName);
     if AName<>'' then
     begin
          Result:=FindComponentRecursive(Owner,AName);
          if Result=nil then
          begin
               for i:=0 to FModules.Count-1 do
               begin
                    Result:=FindComponentRecursive(TDataModule(FModules[i]),AName);
                    if Result<>nil then exit;
               end;
          end;
     end;
end;

function TWABD_Session.GetSessionIDString(const objName: string): string;
var
  dts: string;
begin
  DateTimeToString(dts,'yyyymmddhhnnss',Now());
  Result := Format(WABD_SES_ID_STR_FORMAT, [FSessionID, objName, dts]);
  if (assigned(FSessionMgr)) and  (FSessionMgr.Encrypted) then
      result := WABD_Encrypt(result);

end;

function TWABD_Session.ValidateUrlForNewSession(const url: string): Boolean;
begin
  Result:=False;
  // empty url is always invalid
  if (url='')  then Exit;
  // other urls have to be validated by handlers attributes
  if Assigned(FUrlHandlers) then
      Result:=WABD_Attributes.ValidateUrlForNewSession(FUrlHandlers, url);
end;


{******************************************************************************}
{***************************** TWABD_Object ***********************************}
{******************************************************************************}

constructor TWABD_Object.Create(AOwner: TComponent);
begin
   inherited;
   FVisible := True;
   InLoaded := False;
   FParent := nil;
end;

destructor TWABD_Object.Destroy;
begin
   inherited;
end;

procedure TWABD_Object.SetDependingOn(AObject:TWABD_Object);
begin
     FDependingOn:=AObject;
     if FDependingOn<>nil then FDependingOn.FreeNotification(self);
end;

procedure TWABD_Object.Notification(AComponent: TComponent; Operation: TOperation);
begin
   inherited;
   if (Operation = opRemove) then
   begin
        if AComponent=FDependingOn then FDependingOn:=nil;
   end;
end;


procedure TWABD_Object.SetVisible(b: boolean);
begin
   if b<>FVisible then begin
      FVisible := b;
      Changed;
   end;
end;

function TWABD_Object.GetVisible:boolean;
begin
    Result:=FVisible;
    if FDependingOn<>nil then Result:=FDependingOn.Visible;
end;

procedure TWABD_Object.SetName(const Value: TComponentName);
var
   OldName, NewName  : string;
   ParBody           : TWABD_Body;
begin
   OldName := Name;
   NewName := Value;
   inherited;
   Changed;

end;

procedure TWABD_Object.Changed;
begin
   if Assigned(OnChange) then OnChange(Self);
end;


function TWABD_Object.GetSession: TWABD_Session;
var
    oc:TWABD_Object;
begin
     Result:=nil;
          if (oc is TWABD_Body) then
          begin
               Result:=TWABD_Body(oc).Session;
               exit;
          end;
end;


procedure TWABD_Object.ReadParentName(Reader: TReader);
begin
   Reader.ReadListBegin;
   FParentName := Reader.ReadString;
   Reader.ReadListEnd;
end;

procedure TWABD_Object.WriteParentName(Writer: TWriter);
begin
   Writer.WriteListBegin;
//   if Parent<>nil then
//      Writer.WriteString(Parent.Name)
//   else
//      Writer.WriteString('');
   Writer.WriteListEnd;
end;

procedure TWABD_Object.Loaded;
begin
   inherited;
   InLoaded := True;
   Assert(Owner<>nil, 'Owner = nil');
   InLoaded := False;
end;

procedure TWABD_Object.SaveHTMLToFile(AFile:string);
var
   f:TextFile;
   s:string;
begin
     s:=Object_To_HTML;
     AssignFile(f,AFile);
     try
        Rewrite(f);
        Write(f,s);
     finally
        CloseFile(f);
     end;
end;

function TWABD_Object.HTML_META(options: string): string;
begin
     result := '<META NAME="Generator" CONTENT="'+WABD_VERSION_STR
     + '" http-equiv="Content-Type" content="text/html; charset=UTF-8" '
     + options
     + ' >';
end;


{******************************************************************************}
{**** TWABD_Body. Ancestor of  TWABD_HTML and TWABD_HTMLFile ******************}
{******************************************************************************}

constructor TWABD_Body.Create(AOwner:TComponent);
begin
     inherited;


     FVariables  := TStringList.create;
     FVariableType:=wabdVarTypeStandard;

     FClientProcessTime:=-1;

     FUseSessionCookie:=false;
     FEnableTemplates:=false;
     FHandleRequests:=true;
end;

destructor TWABD_Body.Destroy;
begin
     FVariables.Free;

     inherited;
end;

procedure TWABD_Body.Loaded;
var
   i:integer;
   s:string;
   c:TComponent;
begin
     inherited;


     // Check if onCreate handler.
     if Assigned(FOnCreate) then FOnCreate(self);
end;

procedure TWABD_Body.DefineProperties(Filer: TFiler);
begin
   inherited;
end;


procedure TWABD_Body.DoShow;
begin
     if Assigned(OnShow) then OnShow(Self);
end;

procedure TWABD_Body.BuildTemplates(AProduce:TWABD_Produce);
var
  i: Integer;
begin
     if Session=nil then exit;

     VariableByName[WABD_VAR_SESSIONID]:=inttostr(Session.SessionID);
     VariableByName[WABD_VAR_USERNAME]:=Session.UserName;
     VariableByName[WABD_VAR_INFO]:=Session.Info;
     VariableByName[WABD_VAR_HITCOUNT]:=inttostr(Session.HitCount);
     VariableByName[WABD_VAR_VERSION]:=Session.Version;
     VariableByName[WABD_VAR_SITENAME]:=Session.SessionMgr.SiteName;
     VariableByName[WABD_VAR_SITEID]:=inttostr(Session.SessionMgr.SiteID);
     VariableByName[WABD_VAR_WEBADMIN]:=Session.SessionMgr.WebAdministrator;
     VariableByName[WABD_VAR_MODULENAME]:=Session.SessionMgr.ModuleName;
     VariableByName[WABD_VAR_BASEURL]:=Session.SessionMgr.BaseUrl;
     VariableByName[WABD_VAR_BODYNAME]:=Name;

     Session.ProcessVariables;
     // update variables list with the ones from session object
     for i:=0 to Session.Variables.Count-1 do
     begin
        VariableByName[Session.Variables.Names[i]]:=Session.Variables.ValueFromIndex[i];
     end;

end;

procedure TWABD_Body.CreateSessionCookie;
var
   s:string;
   c:TWABD_Cookie;
begin
     if (Session<>nil) and (Session.Response<>nil) then
     begin
          s:=Session.GetSessionIDString(Name);
          c:=Session.Response.Cookies.GetCookieByName(WABD_SES_ID_STR);
          if c=nil then
          begin
             c:=Session.Response.Cookies.Add(WABD_SES_ID_STR,s);
          end
          else
              c.Value:=s;
     end;
end;

procedure TWABD_Body.SetVariables(NewVariables: TStrings);
begin
     FVariables.Assign(NewVariables);
end;

procedure TWABD_Body.SetVariableByName(AName,AValue:string);
var
   i:integer;
   s:string;
begin
     i:=FVariables.IndexOfName(AName);
     s:=AName+'='+AValue;
     if i>=0 then FVariables.strings[i]:=s
     else FVariables.add(s);
end;

function TWABD_Body.GetVariableByName(AName:string):string;
begin
     Result:=FVariables.Values[AName];
end;



procedure TWABD_Body.SetProperties(Request:TWABD_CustomRequest);
var
   i,j      : integer;
   OneLine  : string;
   NameVal  : string;
   Value    : string;
   c        : TWABD_Object;
   processed: boolean;

   cookie   : TWABD_Cookie;
begin

   cookie:=nil;
   cookie:=Request.Cookies.GetCookieByName(WABD_CLIENTPROCESSTIME_STR);
   if Assigned(cookie) then
   begin
           Value:=Copy(cookie.Value, Pos(':',cookie.Value)+1);
           FClientProcessTime:=StrToFloat(value);
   end;

end;



{******************************************************************************}
{***************************** TWABD_Base_HTML ********************************}
{******************************************************************************}

constructor TWABD_Base_HTML.Create(AOwner:TComponent);
begin
     inherited;

     FHandleRequests:=false;
     FUseSessionCookie:=false;
     FSysResp:=srNone;

     FHTML:=TStringList.Create;
     FHTML.Add('<!DOCTYPE HTML>');
     FHTML.Add('<HTML>');
     FHTML.Add(inherited HTML_META(''));
     FHTML.Add('<HEAD>');
     FHTML.Add('<TITLE>ENTER THE TITLE HERE</TITLE>');
     FHTML.Add('<BODY>');
     FHTML.Add('...ENTER THE HTML HERE...');
     FHTML.Add('</BODY>');
     FHTML.Add('</HTML>');

end;

destructor TWABD_Base_HTML.Destroy;
begin
     inherited;
end;


function  TWABD_Base_HTML.Object_To_HTML:string;
begin
     DoShow;
     if FEnableTemplates then
        BuildTemplates(prodHTML);

     Result:=FHTML.Text;

     Result:=Process_Variables(Result,FVariables,FVariableType);
end;


procedure TWABD_Base_HTML.Show;
begin
   if not Assigned(FWSession) then
      raise Exception.CreateFmt('TWABD_Base_HTML "%s" does not have a WABD_Session!', [Name]);
   FWSession.NewBody := Self;
end;


procedure TWABD_Base_HTML.Notification(AComponent: TComponent; Operation: TOperation);
begin
     inherited;
end;

procedure TWABD_Base_HTML.BuildTemplates(AProduce:TWABD_Produce);
begin
     // Build event handler template.
     inherited BuildTemplates(AProduce); //BuildTemplates del Body
end;



// ************************************************************************
// TWABD_HTML
// ************************************************************************
constructor TWABD_HTML.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
     FEdHTML := TWABDEditHTML.Create;
     FEdHTML.ParHTML := self;

end;

destructor TWABD_HTML.Destroy;
begin
   FEdHTML.free;
   inherited Destroy;
end;

procedure TWABD_HTML.SetEdHTML(NewEdHTML: TWABDEditHTML);
begin
   // Do nothing
end;

function TWABD_HTML.Object_To_HTML:string;
begin
     Result:=inherited Object_To_HTML;
end;


{******************************************************************************}
{***************************** TWABD_HTMLFile *********************************}
{******************************************************************************}

constructor TWABD_HTMLFile.Create(AOwner:TComponent);
begin
     inherited;
     FLoadedWhen:=0;
     FSecsBeforeReload:=0;
     FCached:=false;
end;

destructor TWABD_HTMLFile.Destroy;
begin
     inherited;
end;


function TWABD_HTMLFile.Object_To_HTML:string;
begin
     // Check if cached and to be reloaded or not loaded yet then load.
     if (FLoadedWhen = 0) or
        (  (FCached=true) and (FSecsBeforeReload>0)
            and
            (SecondsBetween(Now, FLoadedWhen) > FSecsBeforeReload)   )
     then
        Reload;

     Result:=inherited Object_To_HTML;
end;


procedure TWABD_HTMLFile.Reload;
var
   fn:string;
begin
     if FSetup=nil then
        fn:=FFileName
     else
         fn:=FSetup.GetLocalFilePath+FFileName;
     if fn='' then exit;
     FHTML.LoadFromFile(fn);
     if FSecsBeforeReload>0 then
        FLoadedWhen:=now;
end;

procedure TWABD_HTMLFile.SetSetup(ASetup:TWABD_Setup);
begin
     FSetup:=ASetup;
     if FSetup<>nil then FSetup.FreeNotification(self);
end;

procedure TWABD_HTMLFile.Notification(AComponent: TComponent; Operation: TOperation);
begin
    inherited;
    if (Operation=opRemove) then
    begin
         if AComponent=FSetup then FSetup:=nil;
    end;
end;


{****************************************************************************}
initialization
  Randomize;


end.
