unit WABD_Reg;

{$include kbmWABD.inc}

interface

uses
    Classes,
    WABD_Objects, WABD_Setup, WABD_Admin_Component,
    DesignEditors,  toolsApi,
    DesignIntf ;


type

   TWABDAboutProperty = class(TClassProperty)
   public
      procedure Edit; override;
      function  GetAttributes: TPropertyAttributes; override;
      function  GetValue: string; override;
   end;


   TWABD_HTML_Comp_Editor = class(TComponentEditor)
      procedure   ExecuteVerb(Index: Integer); override;
      function    GetVerb(Index: Integer): string; override;
      function    GetVerbCount: Integer; override;
   public
      procedure   DesignHTML;
   end;


   TString_List_Editor = class(TPropertyEditor)
      procedure   Edit; override;
      function    GetAttributes: TPropertyAttributes; override;
      function    GetValue: string; override;
   end;


   TWABDEditHTMLProperty = class(TPropertyEditor)
   public
      procedure Edit; override;
      function  GetAttributes: TPropertyAttributes; override;
      function  GetValue: string; override;
      function  GetName: string; override;
   end;


   TWABDFileNameProperty = class(TStringProperty)
   public
      procedure Edit; override;
      function  GetAttributes: TPropertyAttributes; override;
      function  GetValue: string; override;
   end;



//-------------------------------------
   procedure Register;

   procedure RegisterWABDComponent(WABDClass: TWABD_ObjectClass);
   procedure RegisterWABDComponents(WABDClasses: array of TWABD_ObjectClass);
   function  GetWABDClassListCount: integer;
   function  GetWABDClassListItem(i: integer): TWABD_ObjectClass;


var
   WABDClassList         : TList;

implementation

uses
     Dialogs,
     Forms,
     SysUtils,
     WABD_About,
     WABD_HTMLEditor,
     WABD_StrListEditor
     ;

var
   HTMLEditor            : TWABDHTMLEditor;

// ************************************************************************
// About box property editor.
// ************************************************************************

procedure TWABDAboutProperty.Edit;
var
   f : TWABDAboutForm;

begin
   f := TWABDAboutForm.Create(Application);
   try
      f.ShowModal;
      Modified;
   finally
      f.Free;
   end;
end;

function TWABDAboutProperty.GetValue: string;
begin
   Result := '(Click Me)';
end;

function TWABDAboutProperty.GetAttributes: TPropertyAttributes;
begin
   Result := [paDialog];
end;


// ************************************************************************
// HTML property editor.
// ************************************************************************

procedure TWABD_HTML_Comp_Editor.DesignHTML;
begin
   if HTMLEditor=nil then HTMLEditor := TWABDHTMLEditor.Create(nil);
   HTMLEditor.WABDHTML := Component as TWABD_Base_HTML;
   HTMLEditor.MyDesigner := Designer;

   HTMLEditor.Initialize;
   HTMLEditor.Show;
end;

procedure TWABD_HTML_Comp_Editor.ExecuteVerb(Index: Integer);
begin
   case Index of
      0: DesignHTML;
   end;
end;

function TWABD_HTML_Comp_Editor.GetVerb(Index: Integer): string;
begin
   case Index of
      0: Result := '&Design template...';
   end;
end;

function TWABD_HTML_Comp_Editor.GetVerbCount: Integer;
begin
   Result := 1;
end;


// ************************************************************************
// TStrings property editor.
// ************************************************************************

procedure TString_List_Editor.Edit;
var
   d : TStringList_Editor;
   S: TStrings;
   n: Integer;
   comp: TPersistent;
   syn: tEdSyntax;
begin
     comp := GetComponent(0);
     syn  := edNone;
     if (comp is TWABD_Base_HTML)
        then
          syn := edHTML;
// --------
     d := TStringList_Editor.Create(nil, syn);
     try

        S := TStrings(Pointer(GetOrdValue));
        D.Caption := 'Edit - '+ GetName;
        D.Memo.Lines.Text := S.Text;
        D.ShowModal;
        for n := 0 to PropCount-1 do begin
          S := TStrings(Pointer(GetOrdValueAt(n)));
          S.Text := D.Memo.Lines.Text;
        end;
        try
          Modified;
        except
        end;
     finally
        d.Release;
     end;
end;

function TString_List_Editor.GetAttributes: TPropertyAttributes;
begin
   Result := [paDialog];
end;

function TString_List_Editor.GetValue: string;
begin
   Result := '(Strings)';
end;



// ************************************************************************
// HTML property editor.
// ************************************************************************

function TWABDEditHTMLProperty.GetAttributes: TPropertyAttributes;
begin
   Result := [paDialog];
end;

function TWABDEditHTMLProperty.GetValue: string;
begin
   Result := '(Click Me)';
end;

function TWABDEditHTMLProperty.GetName: string;
begin
   Result := 'Edit template';
end;

procedure TWABDEditHTMLProperty.Edit;
begin
   if HTMLEditor=nil then HTMLEditor := TWABDHTMLEditor.Create(nil);
   HTMLEditor.WABDHTML := TWABDEditHTML(GetOrdValue).ParHTML;
   HTMLEditor.MyDesigner := Designer;

   HTMLEditor.Initialize;
   HTMLEditor.Show;
end;


// ************************************************************************
// Filename property editor.
// ************************************************************************

procedure TWABDFileNameProperty.Edit;
var
  fo: TOpenDialog;
begin
  fo := TOpenDialog.Create(nil);
  fo.Filename := GetValue;
  fo.Options := [ofPathMustExist,ofFileMustExist];
  try
    if fo.Execute then SetValue(fo.Filename);
  finally
    fo.Free;
  end;
end;

function  TWABDFileNameProperty.GetAttributes: TPropertyAttributes;
begin
   Result := [paDialog];
end;

function  TWABDFileNameProperty.GetValue: string;
begin
   Result := GetStrValue;
end;


// ************************************************************************
// RegisterWABDComponent
// ************************************************************************

procedure RegisterWABDComponents(WABDClasses: array of TWABD_ObjectClass);
var
   i : integer;
begin
   for i := Low(WABDClasses) to High(WABDClasses) do begin
      RegisterWABDComponent(WABDClasses[i]);
   end;
end;

procedure RegisterWABDComponent(WABDClass: TWABD_ObjectClass);
begin
   if WABDClassList.IndexOf(WABDClass) = -1 then
      WABDClassList.Add(WABDClass);
   RegisterClasses([WABDClass]);
end;

function GetWABDClassListCount: integer;
begin
   Result := WABDClassList.Count;
end;

function GetWABDClassListItem(i: integer): TWABD_ObjectClass;
begin
   if (i < 0) or (i >= GetWABDClassListCount) then
      Result := nil
   else
      Result := WABDClassList.Items[i];
end;


// ************************************************************************
// Registration
// ************************************************************************

procedure Register;
begin
   RegisterComponentEditor(TWABD_HTML, TWABD_HTML_Comp_Editor);

   RegisterPropertyEditor(TypeInfo(TStrings), nil, '', TString_List_Editor);
   RegisterPropertyEditor(TypeInfo(TWABDAbout), nil, '', TWABDAboutProperty);
   RegisterPropertyEditor(TypeInfo(TWABDEditHTML), nil, '', TWABDEditHTMLProperty);


   RegisterComponents('kbmWABD', [
      TWABD_SessionMgr,
      TWABD_Session,
      TWABD_Admin,
      TWABD_Setup,
      TWABD_HTML,
      TWABD_HTMLFile
      ]);

end;
{****************************************************************************}



{****************************************************************************}
initialization
   Randomize;
   WABDClassList := TList.Create;
   IsMultiThread := true;
   RegisterClasses([
      TWABD_Admin,
      TWABD_Setup
      ]);

finalization
   WABDClassList.Free;
   HTMLEditor.Free;

end.
