unit WABD_Request;

interface

uses classes,sysutils,WABD_Cookies,WABD_Common;


type
   TWABD_ListAdder = procedure(AList:TObject; Value:string) of object;

   TWABD_CustomRequest = class
   protected
      FLocalFilePath:string;
      //gds4
      FRequestPath: string;
      FRequestMethod: string;
      //------
      FHeaders : TStringList;
      FQuery   : TStringList;
      FRequest : TStringList;
      FParams  : TStringList;
      FCookies : TWABD_Cookies;
      FSize    : integer;
      FScriptName : string;
      FServerName : string;
      FContentType:string;
      FAuth    : string;
      FRemoteHost : string;
      FRemoteAddr : string;
      FRemoteUser : string;
      FUserName : string;
      FPassword : string;
      FBrowser    : integer;
      //gds4
      //FBrowserVersion : double;
      FBrowserVersion : string;
      FRequestType : integer;

      procedure Parse; virtual; abstract;

   public
      constructor Create;
      destructor Destroy; override;

      procedure DetermineBrowser; virtual;
      procedure DetermineRequestType; virtual;
      
      property LocalFilePath:string read FLocalFilePath write FLocalFilePath;
      property Headers:TStringList read FHeaders;
      property Query:TStringList read FQuery;
      property Request:TStringList read FRequest;
      property Params:TStringList read FParams;
      property Cookies:TWABD_Cookies read FCookies;
      property ContentType:string read FContentType;
      property Size:integer read FSize;
      property ScriptName:string read FScriptName;
      property ServerName:string read FServerName;
      property Auth:string read FAuth;
      property RemoteUser:string read FRemoteUser;
      property RemoteAddr:string read FRemoteAddr;
      property RemoteHost:string read FRemoteHost;
      property UserName:string read FUserName write FUserName;
      property Password:string read FPassword write FPassword;
      property Browser:integer read FBrowser;
      property BrowserVersion:string read FBrowserVersion;
      property RequestType:integer read FRequestType;
      //gds4
      property RequestPath: string  read FRequestPath ;
      property RequestMethod: string read FRequestMethod;

   end;

var
   WABD_DefaultRequestLocalFilePath:string;


implementation

constructor TWABD_CustomRequest.Create;
begin
     inherited;
     FHeaders:=TStringList.Create;
     FQuery:=TStringList.Create;
     FRequest:=TStringList.Create;
     FParams:=TStringList.Create;
     FCookies:=TWABD_Cookies.Create(TWABD_Cookie);
     FLocalFilePath:=WABD_DefaultRequestLocalFilePath;

     // Default we dont know  what browser the user has.
     FBrowser:=WABD_BrowserUnknown;
     //FBrowserVersion:=-1.0;
     FBrowserVersion:='';
end;

destructor TWABD_CustomRequest.Destroy;
begin
     FHeaders.Free;
     FQuery.Free;
     FParams.Free;
     FRequest.Free;
     FCookies.Free;
     inherited;
end;

// Figure out browser from headerinfo.
procedure TWABD_CustomRequest.DetermineBrowser;
var
   i:integer;
   s,v:string;
   p1,p2,p3:integer;


   procedure findBrowser();
   begin
     // from the most easy to recognize to the most difficult
     p1:=pos(WABD_IExplorer_Name,s);
     if p1>0 then
     begin
      FBrowser:=WABD_BrowserIExplorer;
      Exit;
     end;
      p1:=pos(WABD_IExplorer11_Name,s);
      if p1>0 then
      begin
        FBrowser:=WABD_BrowserIExplorer11;
        Exit;
      end;
      p1:=pos(WABD_Edge_Name,s);
      if p1>0 then
      begin
        FBrowser:=WABD_BrowserEdge;
        Exit;
      end;
      p1:=pos(WABD_Mozilla_Name,s);
      if p1>0 then
      begin
          FBrowser:=WABD_BrowserMozilla;
          Exit;
      end;
      p1:=pos(WABD_Chrome_Name,s);
      if p1>0 then
      begin
          FBrowser:=WABD_BrowserChrome;
          Exit;
      end;
     end;


begin
     s:=UpperCase(FHeaders.Text);

     FBrowser:=WABD_BrowserUnknown;

     findBrowser;

     // Not found browsertype?
     if p1<=1 then exit;

     // Look for version number start.
     p2:=-1;
     for i:=p1 to length(s) do
         if (s[i]<'A') or (s[i]>'Z') then
         begin
              p2:=i+1;
              break;
         end;
     if p2<=0 then exit;

     // Look for version number end. V.V
     p3:=length(s)+1;
     for i:=p2 to length(s) do
         if ((s[i]<'0') or (s[i]>'9')) and (s[i]<>'.') then
         begin
              p3:=i;
              break;
         end;
     v:=copy(s,p2,p3-p2);
     //gds4
     FBrowserVersion:=v;
end;

// Figure what type of request.
procedure TWABD_CustomRequest.DetermineRequestType;
var
   s:string;
begin

   FRequestType:=WABD_REQUESTTYPE_HTML;
   s := UpperCase(FHeaders.Values['X-REQUESTED-WITH']);
   // another format
   if s='' then
       s := UpperCase(FHeaders.Values['HTTP_X_REQUESTED_WITH']);
   if pos('XMLHTTPREQUEST', s)>0 then
      FRequestType := WABD_REQUESTTYPE_XMLHTTP;
end;

initialization
   WABD_DefaultRequestLocalFilePath:='.\';

end.


