unit WABD_Setup;

interface
uses Classes, SysUtils, Windows, Forms, Graphics, Controls, StdCtrls, ExtCtrls,
      Inifiles, Registry,
      WABD_Request,  WABD_Utils, WABD_Common;
type
   TWABD_OnStorage = procedure(Sender:TComponent; Section:String; JsonData: string; Ini:TIniFile) of object;
   TWABD_Setup = class(TComponent)
   protected
      FLock : TRTLCriticalSection;

      FLocalRootPath   : string;
      FAutoSetGlobalRootPath : boolean;
      FLocalImagePath  : string;
      FImagePath       : string;
      FLocalFilePath   : string;
      FFilePath        : string;

      FStorage    : TWABD_Storage;
      FStoragePath: string;
      FSectionName: string;
      FAutoLoad  : boolean;
      FAutoSave  : boolean;
      FExpandFromRootPath : boolean;
      FExpandFromGlobalRootPath: boolean;
      FOnLoad    : TWABD_OnStorage;
      FOnSave    : TWABD_OnStorage;
      procedure   Loaded; override;
      procedure   SetLocalRootPath(APath:string);
      function    GetLocalRootPath:string;
   public
      constructor   Create(AOwner:TComponent); override;
      destructor    Destroy; override;
      procedure     Load;
      procedure     Save;
      function      GetImagePath:string;
      function      GetLocalImagePath:string;
      function      GetFilePath:string;
      function      GetLocalFilePath:string;
      procedure     Lock;
      procedure     Unlock;
   published
      property      LocalImagePath : string read FLocalImagePath write FLocalImagePath;
      property      ImagePath : string read FImagePath write FImagePath;
      property      LocalFilePath : string read FLocalFilePath write FLocalFilePath;
      property      FilePath : string read FFilePath write FFilePath;
      property      Storage   : TWABD_Storage read FStorage write FStorage;
      property      StoragePath:string read FStoragePath write FStoragePath;
      property      SectionName:string read FSectionName write FSectionName;
      property      AutoLoad:boolean read FAutoLoad write FAutoLoad;
      property      AutoSave:boolean read FAutoSave write FAutoSave;
      property      OnLoad:TWABD_OnStorage read FOnLoad write FOnLoad;
      property      OnSave:TWABD_OnStorage read FOnSave write FOnSave;
      property      ExpandFromRootPath:boolean read FExpandFromRootPath write FExpandFromRootPath;
      property      ExpandFromGlobalRootPath:boolean read FExpandFromGlobalRootPath write FExpandFromGlobalRootPath;
      property      LocalRootPath:string read GetLocalRootPath write SetLocalRootPath;
      property      AutoSetGlobalRootPath:boolean read FAutoSetGlobalRootPath write FAutoSetGlobalRootPath;
   end;

implementation
uses
   superobject;
// ************************************************************************
// TWABD_Setup
// ************************************************************************
constructor TWABD_Setup.Create(AOwner:TComponent);
begin
    inherited;
    InitializeCriticalSection(FLock);
    FAutoSetGlobalRootPath:=false;
    FExpandFromRootPath:=false;
    FExpandFromGlobalRootPath:=false;
    FSectionName:='Setup';
end;

destructor TWABD_Setup.Destroy;
begin
    // Before destruction, save info.
    if FAutoSave then Save;
    DeleteCriticalSection(FLock);
    inherited;
end;

procedure TWABD_Setup.Save;
var
   jsonOb: ISuperObject;
   ini:TIniFile;
   p, fn: string;
begin
     Lock;
     try
        // Before destruction, save info.
        case Storage of
            storeNone: ;
            storeJsonFile:
              begin
                jsonOb:=TSuperObject.ParseString('{}',true);
                jsonOb.O[FSectionName]:=TSuperObject.ParseString('{}',true);
                begin
                    jsonOb.O[FSectionName].S['ImagePath']:=ImagePath;
                    jsonOb.O[FSectionName].S['LocalImagePath']:=LocalImagePath;
                    jsonOb.O[FSectionName].S['FilePath']:=FilePath;
                    jsonOb.O[FSectionName].S['LocalFilePath']:=LocalFilePath;
                    jsonOb.O[FSectionName].S['LocalRootPath']:=LocalRootPath;
                    if assigned(FOnSave) then FOnSave(self,FSectionName, jsonOb.AsString,nil);
                end;
                p := ExtractFilePath(FStoragePath);
                fn := ExtractFileName(FStoragePath);
                jsonOb.SaveTo(FormatLocalPath(p)+fn);

              end;
            storeIniFile:
              begin
                p := ExtractFilePath(FStoragePath);
                fn := ExtractFileName(FStoragePath);
                ini:=TIniFile.Create(FormatLocalPath(p)+fn);
                if ini<>nil then
                begin
                    ini.WriteString(FSectionName,'ImagePath',ImagePath);
                    ini.WriteString(FSectionName,'LocalImagePath',LocalFilePath);
                    ini.WriteString(FSectionName,'FilePath',ImagePath);
                    ini.WriteString(FSectionName,'LocalFilePath',LocalFilePath);
                    ini.WriteString(FSectionName,'LocalRootPath',LocalRootPath);
                    if assigned(FOnSave) then FOnSave(self,FSectionName,'',ini);
                end;
                ini.free;
              end;
        end;
     finally
        Unlock;
     end;
end;

procedure TWABD_Setup.Load;
var
   jsonOb: ISuperObject;
   ini:TIniFile;

   p, fn: string;
begin
     Lock;
     try
        case Storage of
            storeNone: ;
            storeJsonFile:
              begin
//gds  use DLL directory as root, if drive is missing in path
                p := ExtractFilePath(FStoragePath);
                fn := ExtractFileName(FStoragePath);
                jsonOb:=TSuperObject.ParseFile(FormatLocalPath(p)+fn, True);
                begin
                    ImagePath:=jsonOb.O[FSectionName].S['ImagePath'];
                    LocalImagePath:=jsonOb.O[FSectionName].S['LocalImagePath'];
                    FilePath:=jsonOb.O[FSectionName].S['FilePath'];
                    LocalFilePath:=jsonOb.O[FSectionName].S['LocalFilePath'];
                    LocalRootPath:=jsonOb.O[FSectionName].S['LocalRootPath'];
                    if assigned(FOnLoad) then FOnLoad(self,FSectionName,jsonOb.AsString,nil);
                end;

              end;
            storeIniFile:
              begin
//gds  use DLL directory as root, if drive is missing in path
                p := ExtractFilePath(FStoragePath);
                fn := ExtractFileName(FStoragePath);
                ini:=TIniFile.Create(FormatLocalPath(p)+fn);
                if ini<>nil then
                begin
                    ImagePath:=ini.ReadString(FSectionName,'ImagePath','');
                    LocalImagePath:=ini.ReadString(FSectionName,'LocalImagePath','');
                    FilePath:=ini.ReadString(FSectionName,'FilePath','');
                    LocalFilePath:=ini.ReadString(FSectionName,'LocalFilePath','');
                    LocalRootPath:=ini.ReadString(FSectionName,'LocalRootPath','');
                    if assigned(FOnLoad) then FOnLoad(self,FSectionName,'',ini);
                end;
                ini.free;
              end;
        end;
     finally
      Unlock;
     end;
end;

procedure TWABD_Setup.Loaded;
begin
    inherited;
    if FAutoLoad then Load;
end;

function TWABD_Setup.GetImagePath:string;
begin
     Lock;
     try
        Result:=FormatPath(FImagePath);
     finally
        Unlock;
     end;
end;

function TWABD_Setup.GetLocalImagePath:string;
var
   s:string;
begin
     Lock;
     try
        if FExpandFromRootPath then
        begin
             if FExpandFromGlobalRootPath then
                s:=WABD_DefaultRequestLocalFilePath
             else
                 s:=LocalRootPath;
             Result:=FormatLocalPath(ExpandFileName(s+'\'+FLocalImagePath))
        end
        else
            Result:=FormatLocalPath(FLocalImagePath);
     finally
        Unlock;
     end;
end;

function TWABD_Setup.GetFilePath:string;
begin
     Lock;
     try
        Result:=FormatPath(FFilePath);
     finally
        Unlock;
     end;
end;

function TWABD_Setup.GetLocalFilePath:string;
var
   s:string;
begin
     Lock;
     if (csDesigning in ComponentState) then
            result:='.\'    // project local directory
     else
         try
            if FExpandFromRootPath then
            begin
                 if FExpandFromGlobalRootPath then
                    s:=WABD_DefaultRequestLocalFilePath
                 else
                     s:=LocalRootPath;
                 Result:=FormatLocalPath(ExpandFileName(s+'\'+FLocalFilePath))
            end
            else
                 Result:=FormatLocalPath(FLocalFilePath);
         finally
            Unlock;
         end;
end;

procedure TWABD_Setup.Lock;
begin
     EnterCriticalSection(FLock);
end;

procedure TWABD_Setup.Unlock;
begin
     LeaveCriticalSection(FLock);
end;

procedure TWABD_Setup.SetLocalRootPath(APath:string);
begin
     Lock;
     try
        if FAutoSetGlobalRootPath then
           WABD_DefaultRequestLocalFilePath:=APath;
        FLocalRootPath:=APath;
     finally
        Unlock;
     end;
end;

function  TWABD_Setup.GetLocalRootPath:string;
begin
     Lock;
     try
        if FAutoSetGlobalRootPath then
           Result:=WABD_DefaultRequestLocalFilePath
        else
           Result:=FLocalRootPath;
     finally
        Unlock;
     end;
end;

end.
