unit WABD_Statistics;

interface
uses Classes, SysUtils, Windows, Forms, Graphics, Controls, StdCtrls,
     ExtCtrls,
      Inifiles, Registry;
type

   TWABD_SesStatRec = record
       Stamp:TDateTime;
       Value:double;
       User:PAnsiChar; //string[20];
       Info:PAnsiChar; //string[100];
   end;
   PWABD_SesStatRec = ^TWABD_SesStatRec;

   TWABD_SesSubStatRec = record
       Count:longint;
       Value:double;
       Min:double;
       Max:double;
   end;
   PWABD_SesSubStatRec = ^TWABD_SesSubStatRec;

   TWABD_SesSubStat = class(TStringList)
   public
       destructor Destroy; override;
       procedure AddPoint(ID:string; Value:double);
       procedure ClearPoints;
   end;

   TWABD_SesStatGroups = (wabdStatGroupTurnAround,wabdStatGroupCount,wabdStatGroupValue);

   TWABD_SesStatGroup = class(TThreadList)
   protected
         FName:string;
         FSum:double;
         FCount:integer;
         FMin:double;
         FMax:double;
         FBufferSize:integer;
         FGroupType:TWABD_SesStatGroups;
   public
         HourlyValues:array [0..23] of double;
         HourlyCount:array [0..23] of integer;
         DailyValues:array [1..31] of double;
         DailyCount:array [1..31] of integer;
         MonthlyValues:array [1..12] of double;
         MonthlyCount:array [1..12] of integer;
         DayValues:array [1..7] of double;
         DayCount:array [1..7] of integer;
         SubStat:TWABD_SesSubStat;

         constructor Create;
         destructor Destroy; override;
         procedure Clean;
         procedure Zero;
         property Name:string read FName;
         property Sum:double read FSum;
         property TotalCount:integer read FCount;
         property Min:double read FMin;
         property Max:double read FMax;
         property BufferSize:integer read FBufferSize;
         property GroupType:TWABD_SesStatGroups read FGroupType;
   end;
   TWABD_SesStat = class(TThreadList)
   protected
   public
         destructor Destroy; override;
         procedure AddGroup(GrpName:string; GrpType:TWABD_SesStatGroups; BufSize:integer);
         procedure AddPoint(GrpName:string; User,Info:string; Value:double);
         procedure Clean;
         procedure Save(dllname:string);
         procedure Load(dllname:string);
         procedure Zero;
         function IndexOf(GrpName:string):integer;
   end;

implementation
const

MSECS=60*60*24*1000;                      // Millisecs per 24 hours.

destructor TWABD_SesSubStat.Destroy;
begin
     ClearPoints;
     inherited;
end;

// Add sub stat point.
procedure TWABD_SesSubStat.AddPoint(ID:string; Value:double);
var
   i:integer;
   p:PWABD_SesSubStatRec;
begin
     i:=IndexOf(ID);
     if i<0 then
     begin
          GetMem(p,SizeOf(TWABD_SesSubStatRec));
          p^.Count:=0;
          p^.Value:=0.0;
          p^.Min:=9999999;
          p^.Max:=0;
          AddObject(ID,TObject(p));
     end
     else
         p:=PWABD_SesSubStatRec(Objects[i]);
     inc(p^.Count);
     p^.Value:=p^.Value+Value;
     if Value<p^.Min then p^.Min:=Value;
     if Value>p^.Max then p^.Max:=Value;
end;

// Clear all substat points.
procedure TWABD_SesSubStat.ClearPoints;
var
   i:integer;
begin
     for i:=0 to Count-1 do
         FreeMem(Pointer(Objects[i]));
end;

constructor TWABD_SesStatGroup.Create;
begin
     inherited;
     SubStat:=TWABD_SesSubStat.create;
end;

destructor TWABD_SesStatGroup.Destroy;
var
   i:integer;
   prec:PWABD_SesStatRec;
   lst:TList;
begin
     lst:=LockList;
     try
        with lst do
        begin
            for i:=0 to lst.count-1 do
            begin
                 prec:=PWABD_SesStatRec(Items[i]);
                 StrDispose(prec^.User);
                 StrDispose(prec^.Info);
                 FreeMem(Items[i]);
            end;
            Clear;
        end;
     finally
        UnlockList;
     end;

     SubStat.free;
     inherited;
end;
// Make sure only FBuffersize records is in the list.
procedure TWABD_SesStatGroup.Clean;
var
   i,n:integer;
   prec:PWABD_SesStatRec;
   lst:TList;
begin
     lst:=LockList;
     try
        with lst do
        begin
             n:=count - FBufferSize;
             for i:=0 to n-1 do
             begin
                  prec:=PWABD_SesStatRec(Items[0]);
                  StrDispose(prec^.User);
                  StrDispose(prec^.Info);
                  FreeMem(Items[0]);
                  Delete(0);
             end;
        end;
     finally
        UnlockList;
     end;
end;

// Clear stat. values for group.
procedure TWABD_SesStatGroup.Zero;
var
   i:integer;
begin
     FSum:=0;
     FCount:=0;
     FMin:=999E39;
     FMax:=0;

     for i:=0 to 23 do
     begin
          HourlyValues[i]:=0;
          HourlyCount[i]:=0;
     end;
     for i:=1 to 31 do
     begin
          DailyValues[i]:=0;
          DailyCount[i]:=0;
     end;
     for i:=1 to 12 do
     begin
          MonthlyValues[i]:=0;
          MonthlyCount[i]:=0;
     end;
     for i:=1 to 7 do
     begin
          DayValues[i]:=0;
          DayCount[i]:=0;
     end;
     SubStat.ClearPoints;
end;

// Add stat. group.
procedure TWABD_SesStat.AddGroup(GrpName:string; GrpType:TWABD_SesStatGroups; BufSize:integer);
var
   grp:TWABD_SesStatGroup;
begin
     grp:=TWABD_SesStatGroup.Create;
     grp.FName:=GrpName;
     grp.FBufferSize:=BufSize;
     grp.FGroupType:=GrpType;
     grp.Zero;
     Add(grp);
end;

// Find group by name.
function TWABD_SesStat.IndexOf(GrpName:string):integer;
var
   i:integer;
   lst:TList;
begin
     Result:=-1;
     lst:=LockList;
     try
        with lst do
        begin
            for i:=0 to count-1 do
                with TWABD_SesStatGroup(Items[i]) do
                    if Name=GrpName then
                    begin
                        Result:=i;
                        break;
                    end;
        end;
     finally
        UnlockList;
     end;
end;

// Add meassure to a stat. group.
procedure TWABD_SesStat.AddPoint(GrpName:string; User,Info:string; Value:double);
var
   i:integer;
   grp:TWABD_SesStatGroup;
   prec:PWABD_SesStatRec;
   h,mm,s,ms,d,m,y,dw:word;
   stamp:TDateTime;
   lst:TList;
{$ifdef LEVEL23} // XE2
   userAS: ansistring;
   infoAS: ansistring;
{$endif}

begin
     i:=IndexOf(GrpName);
     if (i<0) then exit;

     lst:=LockList;
     try
         with lst do
         begin
             grp:=TWABD_SesStatGroup(Items[i]);

             // Convert value to correct format.
            // if grp.FGroupType=wabdStatGroupTurnAround then Value:=trunc(MSECS*Value);

             // Get datetime for record.
             stamp:=Now;
             DecodeDate(stamp,y,m,d);
             DecodeTime(stamp,h,mm,s,ms);
             dw:=DayOfWeek(stamp);

             // Allocate room for new stat. record.
             GetMem(prec,sizeof(TWABD_SesStatRec));
             prec^.Stamp:=Now;
             prec^.Value:=Value;
{$ifdef LEVEL23} // XE2
             userAS:=user;
             infoAS:=info;
             prec^.User:=StrNew(PAnsiChar(UserAS));
             prec^.Info:=StrNew(PAnsiChar(InfoAS));
{$else}
             prec^.User:=StrNew(PAnsiChar(User));
             prec^.Info:=StrNew(PAnsiChar(Info));
{$endif}
             grp.Add(prec);

             // Calculate group sums.
             inc(grp.FCount);
             grp.FSum:=grp.FSum+Value;
             if Value<grp.FMin then grp.FMin:=Value;
             if Value>grp.FMax then grp.FMax:=Value;

             // Calculate hourly sums.
             grp.HourlyValues[h]:=grp.HourlyValues[h]+Value;
             inc(grp.HourlyCount[h]);

             // Calculate daily sums.
             grp.DailyValues[d]:=grp.DailyValues[d]+Value;
             inc(grp.DailyCount[d]);

             // Calculate monthly sums.
             grp.MonthlyValues[m]:=grp.MonthlyValues[m]+Value;
             inc(grp.MonthlyCount[m]);

             // Calculate day sums.
             grp.DayValues[dw]:=grp.DayValues[dw]+Value;
             inc(grp.DayCount[dw]);
         end;
    finally
        UnlockList;
    end;
end;

procedure TWABD_SesStat.Clean;
var
   i:integer;
   lst:TList;
begin
     lst:=LockList;
     try
        with lst do
        begin
             for i:=0 to count-1 do
                 with TWABD_SesStatGroup(Items[i]) do Clean;
        end;
     finally
        UnlockList;
     end;
end;

// Save stat info.
procedure TWABD_SesStat.Save(dllname:string);
var
   g:integer;
   grp:TWABD_SesStatGroup;
   i:integer;
   s:string;
   INI:TIniFile;
   os:ansichar;
   lst:TList;
begin
     // Replace
     s:=ChangeFileExt(dllname,'.STA');
     if s=dllname then s:=s+'.STA';          // Dont want to overwrite the dll by accident.
     if ExtractFilePath(s)='' then s:='.\'+s;
// Delphi version >= XE2
{$IF CompilerVersion >= 23.0}
     os:=AnsiChar(FormatSettings.Decimalseparator);
     FormatSettings.Decimalseparator:='.';
{$else}
     os:=AnsiChar(Decimalseparator);
     Decimalseparator:='.';
{$ifend}

     INI:=TInifile.Create(s);

     lst:=LockList;
     try
        with lst do
        begin
             // For all groups.
             for g:=0 to count-1 do
             begin
                  grp:=TWABD_SesStatGroup(Items[g]);

                  // Write group global info.
                  s:=grp.Name;
                  INI.writestring(s,'Sum',floattostr(grp.FSum));
                  INI.writeinteger(s,'Count',grp.FCount);
                  INI.writestring(s,'Min',floattostr(grp.FMin));
                  INI.writestring(s,'Min',floattostr(grp.FMax));

                  // Write hourly info.
                  s:=grp.Name+' Hourly';
                  for i:=0 to 23 do
                  begin
                       INI.WriteString(s,'Val'+inttostr(i),floattostr(grp.HourlyValues[i]));
                       INI.Writeinteger(s,'Cnt'+inttostr(i),grp.HourlyCount[i]);
                  end;

                  // Write daily info.
                  s:=grp.Name+' Daily';
                  for i:=1 to 31 do
                  begin
                       INI.WriteString(s,'Val'+inttostr(i),floattostr(grp.DailyValues[i]));
                       INI.Writeinteger(s,'Cnt'+inttostr(i),grp.DailyCount[i]);
                  end;

                  // Write monthly info.
                  s:=grp.Name+' Monthly';
                  for i:=1 to 12 do
                  begin
                       INI.WriteString(s,'Val'+inttostr(i),floattostr(grp.MonthlyValues[i]));
                       INI.Writeinteger(s,'Cnt'+inttostr(i),grp.MonthlyCount[i]);
                  end;

                  // Write day info.
                  s:=grp.Name+' Weekday';
                  for i:=1 to 7 do
                  begin
                       INI.WriteString(s,'Val'+inttostr(i),floattostr(grp.DayValues[i]));
                       INI.Writeinteger(s,'Cnt'+inttostr(i),grp.DayCount[i]);
                  end;
             end;
        end;
     finally
        UnlockList;
     end;

     INI.free;
// Delphi version >= XE2
{$IF CompilerVersion >= 23.0}
     FormatSettings.Decimalseparator:=char(os);
{$else}
     Decimalseparator:=char(os);
{$ifend}

 end;

// Load stat info.
procedure TWABD_SesStat.Load(dllname:string);
var
   g:integer;
   grp:TWABD_SesStatGroup;
   i,n:integer;
   s:string;
   INI:TIniFile;
   os:ansichar;
   lst:TList;
begin
     // Replace
     s:=ChangeFileExt(dllname,'.STA');
     if s=dllname then s:=s+'.STA';          // Dont want to overwrite the dll by accident.
     if ExtractFilePath(s)='' then s:='.\'+s;

{$IF CompilerVersion >= 23.0}
      // XE2
     os:=AnsiChar(FormatSettings.Decimalseparator);
     FormatSettings.DecimalSeparator:='.';
{$else}
     os:=AnsiChar(Decimalseparator);
     Decimalseparator:='.';
{$ifend}
     INI:=TInifile.Create(s);

     lst:=LockList;
     try
        with lst do
        begin
             // For all groups.
             for g:=0 to count-1 do
             begin
                  grp:=TWABD_SesStatGroup(Items[g]);

                  // Write group global info.
                  s:=grp.Name;
                  n:=INI.readinteger(s,'Count',-1);
                  if n<0 then continue;
                  grp.FCount:=n;
                  grp.FSum:=strtofloat(INI.readstring(s,'Sum','0.0'));
                  grp.FMin:=strtofloat(INI.readstring(s,'Min','0.0'));
                  grp.FMax:=strtofloat(INI.readstring(s,'Min','0.0'));

                  // read hourly info.
                  s:=grp.Name+' Hourly';
                  for i:=0 to 23 do
                  begin
                       grp.HourlyValues[i]:=strtofloat(INI.readString(s,'Val'+inttostr(i),'0.0'));
                       grp.HourlyCount[i]:=INI.readinteger(s,'Cnt'+inttostr(i),0);
                  end;

                  // read daily info.
                  s:=grp.Name+' Daily';
                  for i:=1 to 31 do
                  begin
                       grp.DailyValues[i]:=strtofloat(INI.readString(s,'Val'+inttostr(i),'0.0'));
                       grp.DailyCount[i]:=INI.readinteger(s,'Cnt'+inttostr(i),0);
                  end;

                  // read monthly info.
                  s:=grp.Name+' Monthly';
                  for i:=1 to 12 do
                  begin
                       grp.MonthlyValues[i]:=strtofloat(INI.readString(s,'Val'+inttostr(i),'0.0'));
                       grp.MonthlyCount[i]:=INI.readinteger(s,'Cnt'+inttostr(i),0);
                  end;

                  // read day info.
                  s:=grp.Name+' Weekday';
                  for i:=1 to 7 do
                  begin
                       grp.DayValues[i]:=strtofloat(INI.readString(s,'Val'+inttostr(i),'0.0'));
                       grp.DayCount[i]:=INI.readinteger(s,'Cnt'+inttostr(i),0);
                  end;
             end;
        end;
     finally
        UnlockList;
     end;

     INI.free;
// Delphi version >= XE2
{$IF CompilerVersion >= 23.0}
     FormatSettings.Decimalseparator:=char(os);
{$else}
     Decimalseparator:=char(os);
{$ifend}

end;

procedure TWABD_SesStat.Zero;
var
   g:integer;
   grp:TWABD_SesStatGroup;
   lst:TList;
begin
     lst:=LockList;
     try
        with lst do
        begin
             // For all groups.
             for g:=0 to count-1 do
             begin
                  grp:=TWABD_SesStatGroup(Items[g]);
                  grp.Zero;
             end;
        end;
     finally
        UnlockList;
     end;
end;

// Clear all stat.groups and deallocate memory.
destructor TWABD_SesStat.Destroy;
var
   i:integer;
   lst:TList;
begin
     lst:=LockList;
     try
        with lst do
        begin
             for i:=0 to count-1 do
                 TWABD_SesStatGroup(Items[i]).free;
             Clear;
        end;
     finally
        UnlockList;
     end;

     inherited;
end;

end.
