unit WABD_StrListEditor;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, ComCtrls, StdCtrls, ExtCtrls, SynEdit, SynMemo,
  SynHighlighterJScript, SynHighlighterHtml, SynHighlighterCSS,
  SynEditHighlighter, SynHighlighterMulti, SynEditOptionsDialog, Menus,
  SynEditSearch, SynEditMiscClasses;
type
  tEdSyntax = (edNone, edHTML, edJavascript, edCSS);
type
  TStringList_Editor = class(TForm)
    Panel1: TPanel;
    btnLoad: TButton;
    btnSave: TButton;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    memo: TSynMemo;
    pmEdit: TPopupMenu;
    mnCopy: TMenuItem;
    mnPaste: TMenuItem;
    mnCut: TMenuItem;
    mnUndo: TMenuItem;
    mnOptions: TMenuItem;
    SynEditOptionsDialog1: TSynEditOptionsDialog;
    SynMultiSyn1: TSynMultiSyn;
    SynCssSyn1: TSynCssSyn;
    SynJScriptSyn1: TSynJScriptSyn;
    mnSyntax: TMenuItem;
    mnHTML: TMenuItem;
    mnJscript: TMenuItem;
    mnCss: TMenuItem;
    SynHTMLSyn1: TSynHTMLSyn;
    panelFunctions: TPanel;
    cbFunctions: TComboBox;
    SynEditSearch1: TSynEditSearch;
    Label1: TLabel;
    mnSearch: TMenuItem;
    procedure btnLoadClick(Sender: TObject);
    procedure btnSaveClick(Sender: TObject);
    procedure mnOptionsClick(Sender: TObject);
    procedure mnUndoClick(Sender: TObject);
    procedure mnCopyClick(Sender: TObject);
    procedure mnPasteClick(Sender: TObject);
    procedure mnCutClick(Sender: TObject);
    procedure mnHTMLClick(Sender: TObject);
    procedure mnJscriptClick(Sender: TObject);
    procedure mnCssClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure cbFunctionsChange(Sender: TObject);
    procedure mnSearchClick(Sender: TObject);
    procedure memoChange(Sender: TObject);
  private
    { Private declarations }
    FSyntax: tEdSyntax;
    keyCount: Integer;

    procedure setSyntax();
    procedure loadJSFunctions();
  public
    { Public declarations }
    constructor Create(AOwner: TComponent; syntax: tEdSyntax);
    destructor  Destroy;
    property syntax: tEdSyntax read FSyntax write FSyntax;
  end;

  TFuncNamePos = record
    fPos: Integer;
    fLen: Integer;
  end;
  PFuncNamePos = ^TFuncNamePos;

var
  StringList_Editor: TStringList_Editor;

implementation

{$R *.DFM}

const
  FUNCTION_LEN = 8; // length of word "function"


procedure TStringList_Editor.setSyntax();
begin
  case FSyntax of
    edHTML: begin OpenDialog1.Filter := SynHTMLSyn1.DefaultFilter;
                  SaveDialog1.Filter := SynHTMLSyn1.DefaultFilter;
                  memo.Highlighter := SynMultiSyn1;
                  mnHTML.Checked := true;
                  panelFunctions.Visible := False;
            end;
    edJavascript:
            begin OpenDialog1.Filter := SynJScriptSyn1.DefaultFilter;
                  SaveDialog1.Filter := SynJScriptSyn1.DefaultFilter;
                  memo.Highlighter := SynJScriptSyn1;
                  mnJscript.Checked := true;
                  panelFunctions.Visible := true;
            end;
    edCSS: begin
                  OpenDialog1.Filter := SynCssSyn1.DefaultFilter;
                  SaveDialog1.Filter := SynCssSyn1.DefaultFilter;
                  memo.Highlighter := SynCssSyn1;
                  mnCss.Checked := true;
                  panelFunctions.Visible := False;
            end;

  end;
end;

constructor TStringList_Editor.Create(AOwner: TComponent; syntax: tEdSyntax);
begin
  inherited create(AOwner);
  FSyntax := syntax;
  setSyntax;
  keyCount := 0;
end;
destructor TStringList_Editor.Destroy;
begin
  inherited Destroy;
end;

procedure TStringList_Editor.FormShow(Sender: TObject);
begin
  if FSyntax = edJavascript then
  begin
      keyCount := Length(memo.Text);
      loadJSFunctions;
  end;
end;

procedure TStringList_Editor.btnLoadClick(Sender: TObject);
begin
    if OpenDialog1.execute then
    begin
        Memo.Lines.LoadFromFile(OpenDialog1.FileName);
        if FSyntax = edJavascript then
        begin
            keyCount := Length(memo.Text);
            loadJSFunctions;
        end;
    end;
end;

procedure TStringList_Editor.btnSaveClick(Sender: TObject);
begin
    if SaveDialog1.Execute then Memo.Lines.Savetofile(SaveDialog1.filename);
end;

procedure TStringList_Editor.cbFunctionsChange(Sender: TObject);
var
  l : PFuncNamePos;
begin
    // goto function line
    l := Pfuncnamepos (cbFunctions.Items.Objects[cbFunctions.ItemIndex]);
    memo.SelStart := l^.fPos;
    memo.SelEnd := memo.SelStart + l^.fLen;
end;

procedure TStringList_Editor.memoChange(Sender: TObject);
begin
    if (FSyntax = edJavascript) then
      // maybe the user has written the word "function" followed by the identifier
      if   (  ((length(memo.Text) - keyCount) >= (FUNCTION_LEN  + 8))  )
        or ( (keyCount - Length(memo.Text)) >= FUNCTION_LEN    ) // or the user has deleted some words
       then
      begin
          loadJSFunctions;
          keyCount := Length(memo.Text);
      end;

end;

procedure TStringList_Editor.mnCopyClick(Sender: TObject);
begin
    memo.CopyToClipboard;
end;

procedure TStringList_Editor.mnCssClick(Sender: TObject);
begin
    FSyntax := edCSS;
    setSyntax;
end;

procedure TStringList_Editor.mnCutClick(Sender: TObject);
begin
    memo.CutToClipboard;
end;

procedure TStringList_Editor.mnHTMLClick(Sender: TObject);
begin
    FSyntax := edHTML;
    setSyntax;
end;

procedure TStringList_Editor.mnJscriptClick(Sender: TObject);
begin
    FSyntax := edJavascript;
    setSyntax;
end;

procedure TStringList_Editor.mnOptionsClick(Sender: TObject);
var
  synoptcont: TSynEditorOptionsContainer;
begin
  synoptcont := TSynEditorOptionsContainer.Create(memo);
  synoptcont.Assign(memo);
  SynEditOptionsDialog1.Execute(synoptcont);
  synoptcont.AssignTo(memo);
  synoptcont.Free;
end;

procedure TStringList_Editor.mnPasteClick(Sender: TObject);
begin
    memo.PasteFromClipboard;
end;

procedure TStringList_Editor.mnSearchClick(Sender: TObject);
begin
    //memo.SearchEngine.Options;
end;

procedure TStringList_Editor.mnUndoClick(Sender: TObject);
begin
    memo.Undo;
end;

procedure TStringList_Editor.loadJSFunctions;
var
  s: string;
  l: PFuncNamePos;
begin
      cbFunctions.Items.Clear;
      cbFunctions.Sorted := True;
      cbFunctions.ItemIndex := -1;
      SynJScriptSyn1.ResetRange;
      SynJScriptSyn1.SetLine(memo.Text, 1);
      while
      not (SynJScriptSyn1.GetEol) do
      begin
        if SynJScriptSyn1.GetTokenKind = Ord(SynHighlighterJScript.tkKey) then begin
          s := SynJScriptSyn1.GetToken;
          if (UpperCase(s) = 'FUNCTION') then
          begin
              repeat
              begin
                  SynJScriptSyn1.Next;  // get function name
              end;
              until (SynJScriptSyn1.GetTokenKind = Ord(SynHighlighterJScript.tkIdentifier))
                  or  (SynJScriptSyn1.GetTokenKind = Ord(SynHighlighterJScript.tkSymbol))
                  or (SynJScriptSyn1.GetEol) ;
              if   (SynJScriptSyn1.GetTokenKind = Ord(SynHighlighterJScript.tkIdentifier)) then
              begin
                  s := SynJScriptSyn1.GetToken;
                  with cbFunctions.Items do begin
                      New(l);
                      l^.fPos := SynJScriptSyn1.GetTokenPos;
                      l^.fLen := Length(s);
                      if l^.fLen<1 then l^.fLen := 1;
                      if s='' then s := 'Noname';
                      AddObject(s, pointer(l));
                  end;
              end;
          end;
        end;
        SynJScriptSyn1.Next;
      end;
      if cbFunctions.Items.Count>0 then
          cbFunctions.ItemIndex := 0;
end;

end.
